/*************************************************
 * Abstract classes of math blocks and commands.
 ************************************************/

/**
 * Math tree node base class.
 * Some math-tree-specific extensions to Node.
 * Both MathBlock's and MathCommand's descend from it.
 */
var MathElement = P(Node, function(_, super_) {
  _.finalizeInsert = function(options, cursor) { // `cursor` param is only for
      // SupSub::contactWeld, and is deliberately only passed in by writeLatex,
      // see ea7307eb4fac77c149a11ffdf9a831df85247693
    var self = this;
    self.postOrder(function (node) { node.finalizeTree(options) });
    self.postOrder(function (node) { node.contactWeld(cursor) });

    // note: this order is important.
    // empty elements need the empty box provided by blur to
    // be present in order for their dimensions to be measured
    // correctly by 'reflow' handlers.
    self.postOrder(function (node) { node.blur(); });

    self.postOrder(function (node) { node.reflow(); });
    if (self[R].siblingCreated) self[R].siblingCreated(options, L);
    if (self[L].siblingCreated) self[L].siblingCreated(options, R);
    self.bubble(function (node) { node.reflow(); });
    /**
     * @author ThuyLTM
     * write for make the function name bold
     */
    if (self instanceof Subscript && !self.isEmpty()) {
      self.sub.ends[R].autoBoldFunctionName(options);
    } else if (self.autoBoldFunctionName) {
      self.autoBoldFunctionName(options);
    }
  };
  /**
   * @author ThuyLTM
   * write for make the function name bold
   * START
   */
  _.autoBoldFunctionName = function(opts) {
    var functionName = undefined;
    if (this.digit || this instanceof Letter || this instanceof Subscript || this.variable) {
      functionName = this._paintFunctionName(opts);
      //functionName return undefined when this is not included in the bold function name
    }
    if (this instanceof Bracket && this.ctrlSeq === "\\left(" && this.side===-1) {
      if (this[L].digit || this[L] instanceof Letter || this[L] instanceof Subscript || this[L].variable) {
        functionName = this[L]._paintFunctionName(opts);
      }
    }
    //NNANH - 4/18/2019 - S - Add - Bug#11445 check error js
    if(!(this instanceof Number)){
      this._postBoldFunctionName(opts, functionName);
    }
    //NNANH - 4/18/2019 - S - Add - Bug#11445 check error js
    /* Temporarily comment out checking the number of parameter in the function
    if (this.parent && this.parent.parent instanceof Bracket && this.parent.parent[L] !== 0) {
      //this is in the bracket
      var funcName = this.parent.parent[L];
      if (funcName instanceof Subscript) {
        funcName = funcName.getLeftOutOfSubscript();
      }
      if (!(funcName.jQ.attr("restrict-func-para"))) return;
      var expectSeparateCount = parseInt(funcName.jQ.attr("restrict-func-para")) - 1;
      var realSeparateCount = this.parent.jQ.children("span:contains(',')").length;
      if (this.parent.parent._checkParaCount(expectSeparateCount) && realSeparateCount === expectSeparateCount) {
        this.parent.parent[L]._enableCorrectFuncPara(true, opts);
      } else {
        this.parent.parent[L]._enableCorrectFuncPara(false, opts);
      }
    }
    this._checkParaCountOutOfBracket(opts); */
  };
  _._paintSpecialFunctionName = function(autoOps) {
    if (this.ctrlSeq === undefined && this.variable) {
      var greekAlphabetHex = [
        //lower Greek
        "\uED67", "\uED68", "\uED69", "\uED6A", "\uED6B", "\uED6C", "\uED6D", "\uED6E", "\uED6F", "\uED70",
        "\uED71", "\uED72", "\uED73", "\uED74", "\uED75", "\uED76", "\uED77", "\uED78", "\uED79", "\uED7A",
        "\uED7B", "\uED7C", "\uED7D", "\uED7E",
         //upper Greek
        "\uEC67", "\uEC68", "\uEC69", "\uEC6A", "\uEC6B", "\uEC6C", "\uEC6D", "\uEC6E", "\uEC6F", "\uEC70",
         "\uEC71", "\uEC72", "\uEC73", "\uEC74", "\uEC75", "\uEC76",
        "\uEC77", "\uEC78", "\uEC79", "\uEC7A", "\uEC7B", "\uEC7C", "\uEC7D", "\uEC7E"
      ];
      var greekAlphabetName = [
        //lower Greek
          "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta", "theta", "iota", "kappa",
          "lambda", "mu", "nu", "xi", "omicron", "pi", "rho", "sigma", "tau", "upsilon", "phi", "chi", "psi", "omega",
            //upper Greek
        "Alpha", "Beta", "Gamma", "Delta", "Epsilon", "Zeta", "Eta", "Theta", "Iota", "Kappa",
        "Lambda", "Mu", "Nu", "Xi", "Omicron", "Pi", "Rho", "Sigma", "Tau", "Upsilon", "Phi", "Chi", "Psi", "Omega"
      ];
      if (greekAlphabetHex.indexOf(this.variable) > -1) {
        this.ctrlSeq = "\\" + greekAlphabetName[greekAlphabetHex.indexOf(this.variable)];
      } else {
        this.ctrlSeq = this.variable;
      }
    }
    if (this.ctrlSeq
       && (!(this[L] instanceof Letter || this[L].variable) || this[L] === 0)
      // && (this[R] === 0 || (this[R] instanceof Bracket && this[R].ctrlSeq === "\\left("))
      ) {
        var str = this.ctrlSeq.trim().substring(1);
        if (autoOps.hasOwnProperty(str)) {
            this.jQ.toggleClass('bold', true);
            this.jQ.attr("func-name",this.ctrlSeq);
            this.jQ.attr("restrict-func-para",autoOps[str]);
            return this.ctrlSeq;
        }
    } else {
       this.jQ.toggleClass('bold', false);
    }
    return undefined;
  };
  _._checkParaCountOutOfBracket = function(opts) {
    //this is out of the bracket
    if ((this.digit || this instanceof Letter) && this.jQ.is(".bold")) {
      var expectSeparateCount = parseInt(this.jQ.attr("restrict-func-para")) - 1;
      var length = this.jQ.attr("func-name").length - this.jQ.attr("func-name").indexOf(this.digit||this.letter);
      var last = this;
      for (var i = 1; i <= length; i+= 1) {
        if (last instanceof Letter || last.digit) {
          if (last[R] instanceof Subscript) {
            if (last[R].isEmpty()) return;
            length+= 1;
            last = last[R].getFirstLeftSubscript();
            continue;
          }
          if (last[R]===0) break;
          last=last[R];
        } else {
          break;
        }
      }
      if (last && last[R] ===0 && last.parent.parent instanceof Subscript) {
        last = last.parent.parent.getRightOutOfSubscript();
      } else {
        last = last[R];
      }
      if (last === 0) return;
      if (!(last instanceof Bracket && last.ctrlSeq==="\\left(")) return;
      var realSeparateCount = last.children().jQ.children("span:contains(',')").length;
      if (last._checkParaCount(expectSeparateCount) && realSeparateCount === expectSeparateCount) {
        last[L]._enableCorrectFuncPara(true, opts);
      } else {
        last[L]._enableCorrectFuncPara(false, opts);
      }
    }
  };
  _._checkParaCount = function(separateCount) {
    if (this.blocks[0].isEmpty()) return false;
    if (this.blocks[0].ends[L].ctrlSeq===","||this.blocks[0].ends[R].ctrlSeq===",") return false;
    if (this === 0 && !this.blocks[0].isEmpty()) return true;
    var i = 0;
    var first = this.blocks[0].ends[L];
    while (first !== 0) {
      if (first.ctrlSeq===",") {
        if (first[L].ctrlSeq==="," || first[R].ctrlSeq===",") return false;
        if (first[L] === 0 || first[R] === 0) return false;
        i+=1;
      }
      first = first[R];
    }
    if (i===separateCount) return true;
    return false;
  };
  _._postBoldFunctionName = function(options, functionName) {
    if (this instanceof Bracket && this.ctrlSeq === "\\left(" && this.side===-1) return;
    if (((this.digit || this instanceof Letter) && this.jQ.is(".bold"))
        || (this instanceof Subscript && !this.isEmpty() && this.getFirstRightSubscript().jQ.is(".bold"))) {
        return;
    }
    // the left is the bold function name, the bold is wrong because of cross-matching
    if (this[L] !== 0 && (this[L].digit || this[L] instanceof Letter || this[L] instanceof Subscript || this[L].variable)) {
      //it stand between the bold function name and bracket
      if ((this[L].digit || this[L] instanceof Letter || this[L].variable) && this[L].jQ.is(".bold")) {
        this[L]._clearLeftBold(true);
      }
      if (this[L] instanceof Subscript && !this[L].isEmpty() && this[L].getFirstRightSubscript().jQ.is(".bold")) {
        this[L].getFirstRightSubscript()._clearLeftBold(true);
      }
    }
  };
  _._enableCorrectFuncPara = function(isCorrect, options) {
    if (this instanceof Subscript && isCorrect) {
      return this.getFirstRightSubscript()._paintFunctionName(options);
    }
    if (this instanceof Subscript && !isCorrect) {
      return this.getFirstRightSubscript()._clearLeftBold();
    }
    if ((this instanceof Letter || this.digit || this.variable) && isCorrect) {
      return this._paintFunctionName(options);
    }
    if ((this instanceof Letter || this.digit || this.variable) && !isCorrect) {
      return this._clearLeftBold();
    }
  };
  _._paintFunctionName = function(opts) {
    var autoOps = opts.autoBoldFunctionName;
    if (autoOps === undefined || autoOps.maxLength===0) return undefined;
    if (this.variable && !this.letter) {
      return this._paintSpecialFunctionName(autoOps);
    }
    if (!this._isNeedBoldFunctionName()) return undefined;
    //remove Bold
    this._clearLeftBold(true);
    this._clearRightBold(true);
    // start left matching and make them bold
    var functionName = this._leftMatching(autoOps);
    if (functionName) return functionName;
    // start right matching and make them bold
    functionName = this._rightMatching(autoOps);
    if (functionName) return functionName;
    // start cross matching and make them bold
    functionName = this._crossMatching(autoOps);
    if (functionName) return functionName;
    return undefined;
  };
  _._isNeedBoldFunctionName = function () {
    if (this.digit) {
      var loopElmt = this;
      if (loopElmt[L] !== 0 && loopElmt[L].jQ.is(".bold")
         && loopElmt[R] !== 0 && loopElmt[R].jQ.is(".bold")) {
          //number is in the bold function name
          return true;
      }
      if (loopElmt.parent && loopElmt.parent.parent instanceof Subscript) {
        //number is in the sub of the bold function name
        return true;
      }
      while (loopElmt !== 0 && loopElmt.digit) {//ignore the prefix numbers in the bold function name
        loopElmt = loopElmt[R];
      }
      if (loopElmt !== 0 && loopElmt instanceof Letter && loopElmt.jQ.is(".bold")) {
          return false;
      }
      return true;
    }
    return true;
  };
  _._crossMatching = function (autoOps) {
    var self = this;
    for (var loopCount = 1; loopCount <= autoOps.maxLength; loopCount++) {
      if (self[L] != 0 && (self[L] instanceof Letter || self[L].digit)) {
        self = self[L];
      } else if (self[L] === 0 && self.parent && self.parent.parent instanceof Subscript) {
        self = self.parent.parent.getLeftOutOfSubscript();
      } else {
        return;
      }

      for (var i = autoOps.maxLength; i >= autoOps.minLength; i--) {
        var str = "";
        var first=self;
        var last=first[L];
        var length;
        for (var length = 1; length <= i; length+= 1) {
          if (first instanceof Letter || first.digit) {
            str = str + (first.letter || first.digit);
            if (first[R] instanceof Subscript) {
              if (first[R].isEmpty()) return;
              str += "_";
              length += 1;
              first = first[R].getFirstLeftSubscript();
              continue;
            }
            if (first[R]===0) break;
            first=first[R];
          } else {
            break;
          }
      }
      var letter;
      if (length > i || !(first instanceof Letter || first.digit)) {
        letter=first[L];
      } else if (first[R]===0) {
        letter=first;
      }
      var rightLetter = letter[R];
      if (letter[R] === 0 && letter.parent.parent instanceof Subscript) {
        rightLetter = letter.parent.parent.getRightOutOfSubscript();
      }
      str = str.replace(/^\d+/, '');//remove the leading number
      if (autoOps.hasOwnProperty(str) && (last === 0 || !(last instanceof Letter|| last.variable))
         && (rightLetter === 0 || (rightLetter instanceof Bracket && rightLetter.ctrlSeq==="\\left("))
        ) {
          for (var j = 1; j <= str.length; j += 1) {
            if (letter instanceof Letter || letter.digit) {
              letter.jQ.toggleClass('bold', true);
              letter.jQ.attr("func-name",str);
              letter.jQ.attr("restrict-func-para",autoOps[str]);
              letter._addBoldFuncNameMap(str);
            }
            if (letter[L] === 0 && letter.parent && letter.parent.parent instanceof Subscript) {
              letter = letter.parent.parent.getLeftOutOfSubscript();
              j += 1;
              continue;
            }
            if (letter[L] === 0) break;
            letter = letter[L];
          }
          return str;
        }
      }
    }
    return false;
  };
  _._addBoldFuncNameMap = function(funcName) {
    if (!BOLD_FUNC_NAME_MAP[funcName]) {
      BOLD_FUNC_NAME_MAP[funcName] = [];
    }
    var blockNode = Node.getNodeOfElement($(this.jQ).children('.mq-root-block')[0]);
    BOLD_FUNC_NAME_MAP[funcName].push(blockNode);
  };
  _._removeBoldFuncNameMap = function(funcName) {
    if (!BOLD_FUNC_NAME_MAP[funcName]) return;
    var blockNode = Node.getNodeOfElement($(this.jQ).children('.mq-root-block')[0]);
    if (blockNode && BOLD_FUNC_NAME_MAP[funcName].indexOf(blockNode) > -1) {
      var elmtIndex = BOLD_FUNC_NAME_MAP[funcName].indexOf(blockNode);
      BOLD_FUNC_NAME_MAP[funcName].splice(elmtIndex, 1);
    }
  };
  _._clearLeftBold = function (isClearAttribute) {
    if ((this instanceof Letter || this.digit || this.variable)&&this.jQ.is(".bold")) {
      this.jQ.toggleClass('bold', false);
      this._removeBoldFuncNameMap(this.jQ.attr("func-name"));
      if (isClearAttribute) {
        this.jQ.removeAttr("func-name");
        this.jQ.removeAttr("restrict-func-para");
      }
      if (this instanceof Letter || this.digit) {
        this.ctrlSeq = this.letter || this.digit;
      }
    }
    if (this.variable && !this.letter) return;
    //clear Left
    var l = this[L];
    if (this[L] === 0 && this.parent && this.parent.parent instanceof Subscript) {
      l = this.parent.parent.getLeftOutOfSubscript();
    };
    for (l; (l instanceof Letter || l.digit || l.variable)&&l.jQ.is(".bold"); l = l[L]) {
      l.jQ.toggleClass('bold', false);
      l._removeBoldFuncNameMap(l.jQ.attr("func-name"));
      if (isClearAttribute) {
        l.jQ.removeAttr("func-name");
        l.jQ.removeAttr("restrict-func-para");
      }
      l.ctrlSeq = l.letter || l.digit;
      if (l[L] === 0 && l.parent && l.parent.parent instanceof Subscript) {
        l = l.parent.parent;
      }
    }
  };
  _._clearRightBold = function () {
    this.data = {};
     //clear Right
    for (var l = this[R]; (((l instanceof Letter || l.digit  || l.variable)&&l.jQ.is(".bold")) ||
                           (l instanceof Subscript && !l.isEmpty())); l = l[R]) {
      if (l instanceof Subscript) {
        l = l.sub.ends;
        continue;
      }
      l.jQ.toggleClass('bold', false);
      l._removeBoldFuncNameMap(l.jQ.attr("func-name"));
      l.jQ.removeAttr("func-name");
      l.jQ.removeAttr("restrict-func-para")
      l.ctrlSeq = l.letter || l.digit;
    }
  };
  _._leftMatching = function (autoOps) {
     for (var i = autoOps.maxLength; i >= autoOps.minLength; i--) {
      var str = "";
      var first=this;
      var last = first[R];
      var length;
      for (var length = 1; length <= i; length+= 1) {
        if (first instanceof Letter || first.digit) {
          str = (first.letter || first.digit) + str;
          if (first[L] === 0 && first.parent && first.parent.parent instanceof Subscript) {
            str = "_" + str;
            length += 1;
            first = first.parent.parent.getLeftOutOfSubscript();
            continue;
          }
          if (first[L]===0) break;
          first=first[L];
        } else {
          break;
        }
      }
      var letter;
      if (length > i || !(first instanceof Letter || first.digit)) {
        letter=first[R];
      } else if (first[L]===0) {
        letter=first;
      }
      str = str.replace(/^\d+/, '');//remove the leading number
      if (autoOps.hasOwnProperty(str)
          //&& (last === 0 || (last instanceof Bracket && last.ctrlSeq === "\\left("))
          && (letter[L] === 0 || !(letter[L] instanceof Letter || letter[L].variable))
        ) {
          for (var j = 1; j <= str.length; j += 1) {
            if (letter instanceof Letter || letter.digit) {
              letter.jQ.toggleClass('bold', true);
              letter.jQ.attr("func-name",str);
              letter.jQ.attr("restrict-func-para",autoOps[str]);
              letter._addBoldFuncNameMap(str);
            }
            letter = letter[R];
            if (letter instanceof Subscript && !letter.isEmpty()) {
              j += 1;
              letter = letter.getFirstLeftSubscript();
            }
          }
          return str;
      }
    }
    return undefined;
  };
  _._rightMatching = function (autoOps) {
    for (var i = autoOps.maxLength; i >= autoOps.minLength; i--) {
      var str = "";
      var first=this;
      var last=this[L];
      var length;
      for (var length = 1; length <= i; length+= 1) {
        if (first instanceof Letter || first.digit) {
          str = str + (first.letter || first.digit);
          if (first[R] instanceof Subscript) {
            if (first[R].isEmpty()) return;
            str += "_";
            length+= 1;
            first = first[R].getFirstLeftSubscript();
            continue;
          }
          if (first[R]===0) break;
          first=first[R];
        } else {
          break;
        }
      }
      var letter;
      if (length > i || !(first instanceof Letter || first.digit)) {
        letter=first[L];
      } else if (first[R]===0) {
        letter=first;
      }
      var rightLetter = letter[R];
      if (letter[R] === 0 && letter.parent.parent instanceof Subscript) {
        rightLetter = letter.parent.parent.getRightOutOfSubscript();
      }
      str = str.replace(/^\d+/, '');//remove the leading number
      if (autoOps.hasOwnProperty(str)
          && (last === 0 || !(last instanceof Letter|| last.variable))
          && (rightLetter === 0 || (rightLetter instanceof Bracket && rightLetter.ctrlSeq==="\\left("))
        ) {
        for (var j = 1; j <= str.length; j += 1) {
          if (letter instanceof Letter || letter.digit) {
            letter.jQ.toggleClass('bold', true);
            letter.jQ.attr("func-name",str);
            letter.jQ.attr("restrict-func-para",autoOps[str]);
            letter._addBoldFuncNameMap(str);
          }
          if (letter[L] === 0 && letter.parent && letter.parent.parent instanceof Subscript) {
            letter = letter.parent.parent.getLeftOutOfSubscript();
            j += 1;
            continue;
          }
          if (letter[L] === 0) break;
          letter = letter[L];
        }
        return str;
      }
    }
    return undefined;
  };
  _.checkFunctionNameBeforeMoving = function() {
    /**
     * @author ThuyLTM
     * check the bold function name. If not correct, clear the bold style
     * START
     */
    if ((this instanceof Letter || this.digit || this.variable) && this.jQ.is(".bold")) {
      this._clearFunctionNameIfNotCorrect();
    }
    if (this instanceof Subscript && !this.isEmpty() && this.getFirstRightSubscript() && this.getFirstRightSubscript().jQ.is(".bold")) {
      if (this.getFirstRightSubscript() instanceof Letter || this.getFirstRightSubscript().digit) {
        this.getFirstRightSubscript()._clearFunctionNameIfNotCorrect();
      }
    }
  };
  _._clearFunctionNameIfNotCorrect = function() {
    if ((this instanceof Letter || this.digit || this.variable) && this.jQ.is(".bold")) {
      var rightElmt = this[R];
      var lastElmt = this;
      while (rightElmt !== 0 && rightElmt.jQ.is(".bold")) {
        lastElmt = rightElmt;
        rightElmt = rightElmt[R];
      }
      if (rightElmt instanceof Subscript && !rightElmt.isEmpty() && rightElmt.getFirstLeftSubscript().jQ.is(".bold")) {
        rightElmt = rightElmt.getFirstLeftSubscript();
      }
      while ((rightElmt !== 0 && rightElmt.jQ.is(".bold"))) {
        lastElmt = rightElmt;
        rightElmt = rightElmt[R];
      }
      if (rightElmt === 0 && lastElmt.parent.parent instanceof Subscript) {
        rightElmt = lastElmt.parent.parent.getRightOutOfSubscript();
      }
      if (rightElmt === 0 ||
        !(rightElmt instanceof Bracket && rightElmt.ctrlSeq==="\\left(")) {
          lastElmt._clearLeftBold(true);
      }
    }
  }
  /**
   * @author ThuyLTM
   * write for make the function name bold
   * END
   */
});

/**
 * Commands and operators, like subscripts, exponents, or fractions.
 * Descendant commands are organized into blocks.
 */
var MathCommand = P(MathElement, function(_, super_) {
  _.init = function(ctrlSeq, htmlTemplate, textTemplate) {
    var cmd = this;
    super_.init.call(cmd);

    if (!cmd.ctrlSeq) cmd.ctrlSeq = ctrlSeq;
    if (htmlTemplate) cmd.htmlTemplate = htmlTemplate;
    if (textTemplate) cmd.textTemplate = textTemplate;
  };

  // obvious methods
  _.replaces = function(replacedFragment) {
    replacedFragment.disown();
    this.replacedFragment = replacedFragment;
  };
  _.isEmpty = function() {
    return this.foldChildren(true, function(isEmpty, child) {
      return isEmpty && child.isEmpty();
    });
  };

  _.parser = function() {
    var block = latexMathParser.block;
    var optWhitespace = Parser.optWhitespace;
    var succeed = Parser.succeed;
    var self = this;

    return block.times(self.numBlocks()).map(function(blocks) {
      self.blocks = blocks;

      for (var i = 0; i < blocks.length; i += 1) {
        blocks[i].adopt(self, self.ends[R], 0);
      }

      return self;
    });
  };

  // createLeftOf(cursor) and the methods it calls
  _.createLeftOf = function(cursor) {
    var cmd = this;
    var replacedFragment = cmd.replacedFragment;
    if (cursor.jQ.parent() && cursor.jQ.parent().is(".addhint")) {
      cursor.jQ.parent().removeClass("addhint");
    }
    cmd.createBlocks();
    super_.createLeftOf.call(cmd, cursor);
    if (replacedFragment) {
      replacedFragment.adopt(cmd.ends[L], 0, 0);
      replacedFragment.jQ.appendTo(cmd.ends[L].jQ);
    }
    cmd.finalizeInsert(cursor.options);
    cmd.placeCursor(cursor);
  };
  _.createBlocks = function() {
    var cmd = this,
      numBlocks = cmd.numBlocks(),
      blocks = cmd.blocks = Array(numBlocks);

    for (var i = 0; i < numBlocks; i += 1) {
      var newBlock = blocks[i] = MathBlock();
      newBlock.adopt(cmd, cmd.ends[R], 0);
    }
  };
  _.placeCursor = function(cursor) {
    if (this.ctrlSeq === "\\int") {
      return cursor.insAtRightEnd(this.blocks[2]);
    }
    if (this.ctrlSeq === "\\diffb") {
      return cursor.insAtRightEnd(this.blocks[1]);
    }
    //insert the cursor at the right end of the first empty child, searching
    //left-to-right, or if none empty, the right end child
    cursor.insAtRightEnd(this.foldChildren(this.ends[L], function(leftward, child) {
      return leftward.isEmpty() ? leftward : child;
    }));
  };

  // editability methods: called by the cursor for editing, cursor movements,
  // and selection of the MathQuill tree, these all take in a direction and
  // the cursor
  _.moveTowards = function(dir, cursor, updown) {
    var updownInto = updown && this[updown+'Into'];
    cursor.insAtDirEnd(-dir, updownInto || this.ends[-dir]);
    aria.queueDirEndOf(-dir).queue(cursor.parent, true);
  };
  _.deleteTowards = function(dir, cursor) {
    if (this.isEmpty()) cursor[dir] = this.remove()[dir];
    else this.moveTowards(dir, cursor, null);
  };
  _.selectTowards = function(dir, cursor) {
    cursor[-dir] = this;
    cursor[dir] = this[dir];
  };
  _.selectChildren = function() {
    return Selection(this, this);
  };
  _.unselectInto = function(dir, cursor) {
    cursor.insAtDirEnd(-dir, cursor.anticursor.ancestors[this.id]);
  };
  _.seek = function(pageX, cursor) {
    function getBounds(node) {
      var bounds = {}
      bounds[L] = node.jQ.offset().left;
      bounds[R] = bounds[L] + node.jQ.outerWidth();
      return bounds;
    }

    var cmd = this;
    var cmdBounds = getBounds(cmd);

    if (pageX < cmdBounds[L]) return cursor.insLeftOf(cmd);
    if (pageX > cmdBounds[R]) return cursor.insRightOf(cmd);

    var leftLeftBound = cmdBounds[L];
    cmd.eachChild(function(block) {
      var blockBounds = getBounds(block);
      if (pageX < blockBounds[L]) {
        // closer to this block's left bound, or the bound left of that?
        if (pageX - leftLeftBound < blockBounds[L] - pageX) {
          if (block[L]) cursor.insAtRightEnd(block[L]);
          else cursor.insLeftOf(cmd);
        }
        else cursor.insAtLeftEnd(block);
        return false;
      }
      else if (pageX > blockBounds[R]) {
        if (block[R]) leftLeftBound = blockBounds[R]; // continue to next block
        else { // last (rightmost) block
          // closer to this block's right bound, or the cmd's right bound?
          if (cmdBounds[R] - pageX < pageX - blockBounds[R]) {
            cursor.insRightOf(cmd);
          }
          else cursor.insAtRightEnd(block);
        }
      }
      else {
        block.seek(pageX, cursor);
        return false;
      }
    });
  }

  // methods involved in creating and cross-linking with HTML DOM nodes
  /*
    They all expect an .htmlTemplate like
      '<span>&0</span>'
    or
      '<span><span>&0</span><span>&1</span></span>'

    See html.test.js for more examples.

    Requirements:
    - For each block of the command, there must be exactly one "block content
      marker" of the form '&<number>' where <number> is the 0-based index of the
      block. (Like the LaTeX \newcommand syntax, but with a 0-based rather than
      1-based index, because JavaScript because C because Dijkstra.)
    - The block content marker must be the sole contents of the containing
      element, there can't even be surrounding whitespace, or else we can't
      guarantee sticking to within the bounds of the block content marker when
      mucking with the HTML DOM.
    - The HTML not only must be well-formed HTML (of course), but also must
      conform to the XHTML requirements on tags, specifically all tags must
      either be self-closing (like '<br/>') or come in matching pairs.
      Close tags are never optional.

    Note that &<number> isn't well-formed HTML; if you wanted a literal '&123',
    your HTML template would have to have '&amp;123'.
  */
  _.numBlocks = function() {
    var matches = 0;
    if (this.htmlTemplate) {
       matches = this.htmlTemplate.match(/&\d+/g);
    }
    return matches ? matches.length : 0;
  };
  _.html = function() {
    // Render the entire math subtree rooted at this command, as HTML.
    // Expects .createBlocks() to have been called already, since it uses the
    // .blocks array of child blocks.
    //
    // See html.test.js for example templates and intended outputs.
    //
    // Given an .htmlTemplate as described above,
    // - insert the mathquill-command-id attribute into all top-level tags,
    //   which will be used to set this.jQ in .jQize().
    //   This is straightforward:
    //     * tokenize into tags and non-tags
    //     * loop through top-level tokens:
    //         * add #cmdId attribute macro to top-level self-closing tags
    //         * else add #cmdId attribute macro to top-level open tags
    //             * skip the matching top-level close tag and all tag pairs
    //               in between
    // - for each block content marker,
    //     + replace it with the contents of the corresponding block,
    //       rendered as HTML
    //     + insert the mathquill-block-id attribute into the containing tag
    //   This is even easier, a quick regex replace, since block tags cannot
    //   contain anything besides the block content marker.
    //
    // Two notes:
    // - The outermost loop through top-level tokens should never encounter any
    //   top-level close tags, because we should have first encountered a
    //   matching top-level open tag, all inner tags should have appeared in
    //   matching pairs and been skipped, and then we should have skipped the
    //   close tag in question.
    // - All open tags should have matching close tags, which means our inner
    //   loop should always encounter a close tag and drop nesting to 0. If
    //   a close tag is missing, the loop will continue until i >= tokens.length
    //   and token becomes undefined. This will not infinite loop, even in
    //   production without pray(), because it will then TypeError on .slice().

    var cmd = this;
    var blocks = cmd.blocks;
    var cmdId = ' mathquill-command-id=' + cmd.id;
    var tokens = [];
    if (cmd.htmlTemplate) {
      tokens = cmd.htmlTemplate.match(/<[^<>]+>|[^<>]+/g);
    }

    pray('no unmatched angle brackets', tokens.join('') === this.htmlTemplate);

    // add cmdId and aria-hidden (for screen reader users) to all top-level tags
    // Note: with the RegExp search/replace approach, it's possible that an element which is both a command and block may contain redundant aria-hidden attributes.
    // In practice this doesn't appear to cause problems for screen readers.
    for (var i = 0, token = tokens[0]; token; i += 1, token = tokens[i]) {
      // top-level self-closing tags
      if (token.slice(-2) === '/>') {
        tokens[i] = token.slice(0,-2) + cmdId + ' aria-hidden="true"/>';
      }
      // top-level open tags
      else if (token.charAt(0) === '<') {
        pray('not an unmatched top-level close tag', token.charAt(1) !== '/');

        tokens[i] = token.slice(0,-1) + cmdId + ' aria-hidden="true">';

        // skip matching top-level close tag and all tag pairs in between
        var nesting = 1;
        do {
          i += 1, token = tokens[i];
          pray('no missing close tags', token);
          // close tags
          if (token.slice(0,2) === '</') {
            nesting -= 1;
          }
          // non-self-closing open tags
          else if (token.charAt(0) === '<' && token.slice(-2) !== '/>') {
            nesting += 1;
          }
        } while (nesting > 0);
      }
    }
    return tokens.join('').replace(/>&(\d+)/g, function($0, $1) {
      return ' mathquill-block-id=' + blocks[$1].id + ' aria-hidden="true">' + blocks[$1].join('html');
    });
  };

  // methods to export a string representation of the math tree
  _.latex = function() {
    return this.foldChildren(this.ctrlSeq, function(latex, child) {
      return latex + '{' + (child.latex() || ' ') + '}';
    });
  };
  _.sLatex = function() {
      return this.foldChildren(this.ctrlSeq, function(latex, child) {
          return latex + '{' + (child.sLatex() || ' ') + '}';
        });
  };
  _.nativeLatex = function() {
      return this.foldChildren(this.ctrlSeq, function(latex, child) {
          return latex + '{' + (child.nativeLatex() || ' ') + '}';
        });
  };
  _.sLatexToSVG = function() {
    return this.foldChildren(this.ctrlSeq, function(latex, child) {
      return latex + '{' + (child.sLatexToSVG() || ' ') + '}';
    });
  };
  _.textTemplate = [''];
  _.text = function() {
    var cmd = this, i = 0;
    return cmd.foldChildren(cmd.textTemplate[i], function(text, child) {
      i += 1;
      var child_text = child.text();
      if (text && cmd.textTemplate[i] === '('
          && child_text[0] === '(' && child_text.slice(-1) === ')')
        return text + child_text.slice(1, -1) + cmd.textTemplate[i];
      return text + child_text + (cmd.textTemplate[i] || '');
    });
  };
  _.mathspeakTemplate = [];
  _.mathspeak = function() {
    var cmd = this, i = 0;
    return cmd.foldChildren(cmd.mathspeakTemplate[i] || 'Start'+cmd.ctrlSeq+' ', function(speech, block) {
      i += 1;
      return speech + ' ' + block.mathspeak() + ' ' + (cmd.mathspeakTemplate[i]+' ' || 'End'+cmd.ctrlSeq+' ');
    });
  };
});

/**
 * Lightweight command without blocks or children.
 */
var Symbol = P(MathCommand, function(_, super_) {
  _.init = function(ctrlSeq, html, text, mathspeak) {
    if (!text && !!ctrlSeq) text = ctrlSeq.replace(/^\\/, '');   var greekAlphabetHex = [
      //lower Greek
      "\uED67", "\uED68", "\uED69", "\uED6A", "\uED6B", "\uED6C", "\uED6D", "\uED6E", "\uED6F", "\uED70",
      "\uED71", "\uED72", "\uED73", "\uED74", "\uED75", "\uED76", "\uED77", "\uED78", "\uED79", "\uED7A",
      "\uED7B", "\uED7C", "\uED7D", "\uED7E",
      //upper Greek
      "\uEC67", "\uEC68", "\uEC69", "\uEC6A", "\uEC6B", "\uEC6C", "\uEC6D", "\uEC6E", "\uEC6F", "\uEC70",
      "\uEC71", "\uEC72", "\uEC73", "\uEC74", "\uEC75", "\uEC76",
      "\uEC77", "\uEC78", "\uEC79", "\uEC7A", "\uEC7B", "\uEC7C", "\uEC7D", "\uEC7E"
    ];
    var greekAlphabetName = [
        //lower Greek
        "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta", "theta", "iota", "kappa",
        "lambda", "mu", "nu", "xi", "omicron", "pi", "rho", "sigma", "tau", "upsilon", "phi", "chi", "psi", "omega",
         //upper Greek
        "Alpha", "Beta", "Gamma", "Delta", "Epsilon", "Zeta", "Eta", "Theta", "Iota", "Kappa",
        "Lambda", "Mu", "Nu", "Xi", "Omicron", "Pi", "Rho", "Sigma", "Tau", "Upsilon", "Phi", "Chi", "Psi", "Omega"
    ];
    if (greekAlphabetHex.indexOf(ctrlSeq) > -1) {
      ctrlSeq = "\\" + greekAlphabetName[greekAlphabetHex.indexOf(ctrlSeq)] + " ";
    }

    this.mathspeakName = mathspeak || text;
    super_.init.call(this, ctrlSeq, html, [ text ]);
  };

  _.parser = function() { return Parser.succeed(this); };
  _.numBlocks = function() { return 0; };

  _.replaces = function(replacedFragment) {
    replacedFragment.remove();
  };
  _.createBlocks = noop;

  _.moveTowards = function(dir, cursor) {
     /**
     * @author ThuyLTM
     * clear the wrong bold function name
     */
    if (cursor[L] !== 0 && cursor[L].checkFunctionNameBeforeMoving) {
      cursor[L].checkFunctionNameBeforeMoving();
    }
    if (cursor[R] !== 0 && cursor[R].checkFunctionNameBeforeMoving) {
      cursor[R].checkFunctionNameBeforeMoving();
    }
    cursor.jQ.insDirOf(dir, this.jQ);
    cursor[-dir] = this;
    cursor[dir] = this[dir];
    aria.queue(this);
     /**
     * @author ThuyLTM
     * autobold function name when move to
     */
    if (cursor[L] !== 0 && cursor[L].checkFunctionName) {
      cursor[L].checkFunctionName(cursor.options);
    }
    if (cursor[L] === 0 && cursor.parent.parent instanceof Subscript
      && cursor[R] !== 0 && cursor[R].checkFunctionName) {
      cursor[R].checkFunctionName(cursor.options);
    }
    if (cursor[R] !== 0 && cursor[R].checkFunctionNameBeforeMoving) {
      cursor[R].checkFunctionNameBeforeMoving();
    }
  };
  _.deleteTowards = function(dir, cursor) {
    cursor[dir] = this.remove()[dir];
  };
  _.seek = function(pageX, cursor) {
    // insert at whichever side the click was closer to
    if (pageX - this.jQ.offset().left < this.jQ.outerWidth()/2)
      cursor.insLeftOf(this);
    else
      cursor.insRightOf(this);
  };

  _.latex = function(){ return this.ctrlSeq; };
  _.text = function(){ return this.textTemplate.join(''); };
  _.mathspeak = function(){ return this.mathspeakName; };
  _.placeCursor = noop;
  _.isEmpty = function(){ return true; };
});
var VanillaSymbol = P(Symbol, function(_, super_) {
  _.init = function(ch, html, mathspeak) {
    if (/^\d$/.test(ch)) {
      this.digit = ch;
    }
    var greekAlphabetHex = [
       //lower Greek
      "\uED67", "\uED68", "\uED69", "\uED6A", "\uED6B", "\uED6C", "\uED6D", "\uED6E", "\uED6F", "\uED70",
      "\uED71", "\uED72", "\uED73", "\uED74", "\uED75", "\uED76", "\uED77", "\uED78", "\uED79", "\uED7A",
      "\uED7B", "\uED7C", "\uED7D", "\uED7E",
       //upper Greek
       "\uEC67", "\uEC68", "\uEC69", "\uEC6A", "\uEC6B", "\uEC6C", "\uEC6D", "\uEC6E", "\uEC6F", "\uEC70",
      "\uEC71", "\uEC72", "\uEC73", "\uEC74", "\uEC75", "\uEC76",
      "\uEC77", "\uEC78", "\uEC79", "\uEC7A", "\uEC7B", "\uEC7C", "\uEC7D", "\uEC7E"
    ];
    var greekAlphabetName = [
       //lower Greek
        "alpha", "beta", "gamma", "delta", "epsilon", "zeta", "eta", "theta", "iota", "kappa",
        "lambda", "mu", "nu", "xi", "omicron", "pi", "rho", "sigma", "tau", "upsilon", "phi", "chi", "psi", "omega",
        //upper Greek
        "Alpha", "Beta", "Gamma", "Delta", "Epsilon", "Zeta", "Eta", "Theta", "Iota", "Kappa",
        "Lambda", "Mu", "Nu", "Xi", "Omicron", "Pi", "Rho", "Sigma", "Tau", "Upsilon", "Phi", "Chi", "Psi", "Omega"
    ];
    if (/^\d$/.test(ch)) {
      this.digit = ch;
    }
    if (greekAlphabetHex.indexOf(ch) > -1) {
      this.variable = ch;
      var ctrlSeq = "\\" + greekAlphabetName[greekAlphabetHex.indexOf(ch)] + " ";//need the tail space for succeed parser
      return super_.init.call(this, ctrlSeq, '<span>'+(html || ch)+'</span>');
    }
    super_.init.call(this, ch, '<span>'+(html || ch)+'</span>', undefined, mathspeak);
    
  };
  _.vanillaSymbol = function(){ return this.textTemplate; };
  /**
   * @author ThuyLTM
   * write for make the function name bold
   * START
   */
  _.checkFunctionName = function(opts) {
    if (this.digit || this.variable) {
      MathElement.prototype.autoBoldFunctionName.call(this, opts);
    }
  };
  /**
   * @author ThuyLTM
   * write for make the function name bold
   * END
   */
});
var BinaryOperator = P(Symbol, function(_, super_) {
  _.init = function(ctrlSeq, html, text, mathspeak) {
    super_.init.call(this,
      ctrlSeq, '<span class="mq-binary-operator">'+html+'</span>', text, mathspeak
    );
  };
  _.binaryOperator = function(){ return this.textTemplate; };
});

/**
 * Children and parent of MathCommand's. Basically partitions all the
 * symbols and operators that descend (in the Math DOM tree) from
 * ancestor operators.
 */
var MathBlock = P(MathElement, function(_, super_) {
  _.join = function(methodName) {
    var count = this.jQ.filter(".ph_func.mq-hasCursor").children().length;
    if (count == 1) {
      this.jQ.addClass("addhint");
    } else {
      this.jQ.removeClass("addhint");
    }
    return this.foldChildren('', function(fold, child) {

      //Add 2019/10/30 ttnong start: correct out latex power expression: 2^(2) -> {2}^{2}
      if(methodName == "latex"){

        if(child.sup != null && child[-1]){
          var baseLatex = child[-1].latex();
          var index = fold.lastIndexOf(baseLatex);
          fold = fold.substr(0, index) + "{" + baseLatex  + "}";
        }

        return fold + child[methodName]();
      }
      //Add 2019/10/30 ttnong end: correct out latex power expression: 2^(2) -> {2}^{2}

      return fold + child[methodName]();
    });
  };
  _.html = function() { return this.join('html'); };
  _.latex = function() { return this.join('latex'); };
  _.sLatex = function() { return this.join('sLatex'); };
  _.nativeLatex = function() { return this.join('nativeLatex'); };
  _.sLatexToSVG = function() { return this.join('sLatexToSVG'); };
  _.text = function() {
    return (this.ends[L] === this.ends[R] && this.ends[L] !== 0) ?
      this.ends[L].text() :
      this.join('text')
    ;
  };
  _.mathspeak = function() {
    var tempOp = '';
    var autoOps = {};
    if (this.controller) autoOps = this.controller.options.autoOperatorNames;
    return this.foldChildren([], function(speechArray, cmd) {
      if (cmd.isPartOfOperator) {
        tempOp += cmd.mathspeak();
      }
      else {
        if(tempOp!=='') {
          if(autoOps !== {} && autoOps._maxLength > 0) {
            var x = autoOps[tempOp.toLowerCase()];
            if(typeof x === 'string') tempOp = x;
          }
          speechArray.push(tempOp+' ');
          tempOp = '';
        }
        var mathspeakText = cmd.mathspeak();
        var cmdText = cmd.ctrlSeq;
        if (isNaN(cmdText)) {
          mathspeakText  = ' ' + mathspeakText;
          if (cmdText !== '.') {
            mathspeakText += ' ';
          }
        }
        speechArray.push(mathspeakText);
      }
      return speechArray;
    })
    .join('')
    .replace(/ +(?= )/g,'');
  };
  _.ariaLabel = 'block';

  _.keystroke = function(key, e, ctrlr) {
    console.log("keystroke");
    //Pass the common keyboard move forward, backward, upward, downward, delete
    if (ctrlr.options.preventUpDownKeyBoard && (key === "Up"||key === "Down")) {
      return;
    }
    
    // add start for financial <murayama> 20190611.
    if ( ctrlr.cursor.parent.jQ.hasClass("accept-read-only") /*&& (key !== "Enter")*/ ) {
        return;
    }
    // add end   for financial <murayama> 20190611.
    
    if (key === "Backspace"||key === "Right"||key === "Left"||key === "Up"||key === "Down"||key === "Del"||key === "Ctrl-Control"
       || key === "Ctrl-C")  {
      return super_.keystroke.apply(this, arguments);
    }
    // if (ctrlr.cursor.parent && ctrlr.cursor.parent.parent &&
    //     ctrlr.cursor.parent.parent.ctrlSeq === "parametric" && ctrlr.cursor.parent.jQ.hasClass("accept-only-digit")
    //     && ctrlr.options.restrictInputInParametric) {
    //   //restrictInputInParametric: '^[\\d.-]+$'
    //   var restrictInputInParametric = new RegExp(ctrlr.options.restrictInputInParametric, 'g');
    //   if (restrictInputInParametric.exec(new String(e.originalEvent.key)) === null) {
    //     e.preventDefault();
    //     return;
    //   }
    // }
    if (ctrlr.exportLatex() === "" && [ "+", "-", "/", "*" ].indexOf(e.originalEvent.key) > -1 && ctrlr.container[0]) {
      $(ctrlr.container[0]).attr("character",e.originalEvent.key);
    }
    if (ctrlr.options.restrictInputAfterParametric &&
       (/\bparametric\b/g.test(ctrlr.cursor[L].ctrlSeq) || /\bparametric\b/g.test(ctrlr.cursor[R].ctrlSeq))) {
        e.preventDefault();
        return;
    }
    if (ctrlr.options.autoBreakOutOfSubscript && ctrlr.cursor.parent.parent.sub
      && ctrlr.cursor.parent.jQ.is(".mq-sub")
      && e.originalEvent && !/[A-Za-z0-9_]/g.test(e.originalEvent.key)) {//if a new command in subscript
        var currentParent = ctrlr.cursor.parent.parent;
        while (currentParent.sub) {
          if (currentParent.parent.parent.sub) {
            currentParent = currentParent.parent.parent;
          } else {
            currentParent = currentParent.parent;
          }
        }
        ctrlr.cursor.insAtRightEnd(currentParent);
        return super_.keystroke.apply(this, arguments);
    }
    if (ctrlr.options.restrictInputOfSubscript && ctrlr.cursor.parent.parent.sub
      && ctrlr.cursor.parent.jQ.is(".mq-sub")
      && e.originalEvent && !/[A-Za-z0-9_]/g.test(e.originalEvent.key)) {
        e.preventDefault();
        e.stopPropagation();
        return;
    }
    if (ctrlr.options.restrictInputOnSubOfDiff && /\bdiff\b/g.test(ctrlr.cursor.parent.parent.ctrlSeq)
        && ctrlr.cursor.parent.jQ.parent().is(".mq-sup") && !/^\d+$/g.test(e.originalEvent.key)) {
        e.preventDefault();
        e.stopPropagation();
        return;
    }
    /**
     * @author ThuyLTM
     * comment out for
     * accepting spacebar in mathquill's textbox
     */
    /* if (ctrlr.options.spaceBehavesLikeTab
        && (key === 'Spacebar' || key === 'Shift-Spacebar')) {
      e.preventDefault();
      ctrlr.escapeDir(key === 'Shift-Spacebar' ? L : R, key, e);
      return;
    } */

    return super_.keystroke.apply(this, arguments);
  };

  // editability methods: called by the cursor for editing, cursor movements,
  // and selection of the MathQuill tree, these all take in a direction and
  // the cursor
  _.moveOutOf = function(dir, cursor, updown) {
    var updownInto = updown && this.parent[updown+'Into'];
    if (!updownInto && this[dir]) {
      cursor.insAtDirEnd(-dir, this[dir]);
      aria.queueDirEndOf(-dir).queue(cursor.parent, true);
    }
    //NNANH 1/24/2019 CR#9979 Add - S - dont move out of seq sticky
    else if(!(cursor.jQ.parent().is(".block-sequence"))){
      cursor.insDirOf(dir, this.parent);
      aria.queueDirOf(dir).queue(this.parent);
    }
    //NNANH 1/24/2019 CR#9979 Add - E - dont move out of seq sticky
    cursor.parent.bubble(function(node) { node.moveOutOfBlock && node.moveOutOfBlock(); });
  };
  _.selectOutOf = function(dir, cursor) {
    cursor.insDirOf(dir, this.parent);
  };
  _.deleteOutOf = function(dir, cursor) {
    cursor.unwrapGramp();
  };
  _.seek = function(pageX, cursor) {
    var node = this.ends[R];
    if (!node || node.jQ.offset().left + node.jQ.outerWidth() < pageX) {
      return cursor.insAtRightEnd(this);
    }
    if (pageX < this.ends[L].jQ.offset().left) return cursor.insAtLeftEnd(this);
    while (pageX < node.jQ.offset().left) node = node[L];
    return node.seek(pageX, cursor);
  };
  _.chToCmd = function(ch, options) {
    var cons;
    // exclude f because it gets a dedicated command with more spacing
    if (ch.match(/^[a-eg-zA-Z]$/))
      return Letter(ch);
    else if (/^\d$/.test(ch))
      return Digit(ch);
    else if (options && options.typingSlashWritesDivisionSymbol && ch === '/')
      return LatexCmds['÷'](ch);
    else if (options && options.typingAsteriskWritesTimesSymbol && ch === '*')
      return LatexCmds['×'](ch);
    else if (options && options.typingPercentWritesPercentOf && ch === '%')
      return LatexCmds.percentof(ch);
    else if (cons = CharCmds[ch] || LatexCmds[ch])
      return cons(ch);
    else
      return VanillaSymbol(ch);
  };
  _.write = function(cursor, ch) {
    var cmd = this.chToCmd(ch, cursor.options);
    if (cursor.selection) cmd.replaces(cursor.replaceSelection());
    cmd.createLeftOf(cursor.show());
    this.jQ.removeClass("addhint");
    if (cursor[L] !== 0 && cursor[L].checkFunctionName) {
      cursor[L].checkFunctionName(cursor.options);    
    }
    // special-case the slash so that fractions are voiced while typing
    if (ch === '/') {
      aria.alert('over');
    } else {
      aria.alert(cmd.mathspeak({ createdLeftOf: cursor }));
    }
  };

  _.focus = function() {
    this.jQ.addClass('mq-hasCursor');
    this.jQ.removeClass('mq-empty');
    var count = this.jQ.filter(".ph_func").children().length;
    if (count == 1) {
       this.jQ.addClass("addhint");
    } else {
       this.jQ.removeClass("addhint");
    }
    return this;
  };
  _.blur = function() {
    this.jQ.removeClass('mq-hasCursor');
    this.jQ.removeClass("addhint");    
    if (this.isEmpty()) {
      this.jQ.addClass('mq-empty');
      if (this.isEmptyParens()) {
        this.jQ.addClass('mq-empty-parens');
      } else if (this.isEmptySquareBrackets()) {
        this.jQ.addClass('mq-empty-square-brackets');
      }
    }
    return this;
  };
});

API.StaticMath = function(APIClasses) {
  return P(APIClasses.AbstractMathQuill, function(_, super_) {
    this.RootBlock = MathBlock;
    _.__mathquillify = function(opts, interfaceVersion) {
      this.config(opts);
      super_.__mathquillify.call(this, 'mq-math-mode');
      this.__controller.delegateMouseEvents();
      this.__controller.staticMathTextareaEvents();
      return this;
    };
    _.init = function() {
      super_.init.apply(this, arguments);
      var innerFields = this.innerFields = [];
      this.__controller.root.postOrder(function (node) {
        node.registerInnerField(innerFields, APIClasses.MathField);
      });
    };
    _.latex = function() {
      //NNANH 12/3/2018 - Edit - Bug#8615 - E - Change suggestion with and or xor randseed and "|"
      //NNANH 1/10/2018 start Bug #8615 【formula style】"and" "or" "xor " when load old data with table
      if(arguments[0] && arguments[0] === arguments[0] + ''){
        arguments[0] = arguments[0].replace(/(∧)/g, "\\ and\\ "); //have two side spaces
        arguments[0] = arguments[0].replace(/(⊻)/g, "SxrS"); //convert xor -> different latex and then reset them again
        arguments[0] = arguments[0].replace(/(∨)/g, "\\ or\\ "); //have two side spaces
        arguments[0] = arguments[0].replace(/(SxrS)/g, "\\ xor\\ "); //have two side spaces
        arguments[0] = arguments[0].replace(/(\\left\|)/g, "leftASB"); //have two side spaces
        arguments[0] = arguments[0].replace(/(\\right\|)/g, "rightASB"); //have two side spaces
        if(!(arguments[0].indexOf("\\ \|\\ ") > -1)){
            arguments[0] = arguments[0].replace(/(\|)/g, "\\ |\\ "); //have two side spaces
        }
        arguments[0] = arguments[0].replace(/(leftASB)/g, "\\left|"); //have two side spaces
        arguments[0] = arguments[0].replace(/(rightASB)/g, "\\right|"); //have two side spaces
      }
      //NNANH 1/10/2018 start Bug #8615 【formula style】"and" "or" "xor " when load old data with table
      //NNANH 12/3/2018 - Edit - Bug#8615 - E - Change suggestion with and or xor randseed and "|"
      var returned = super_.latex.apply(this, arguments);
      if (arguments.length > 0) {
        var innerFields = this.innerFields = [];
        this.__controller.root.postOrder(function (node) {
          node.registerInnerField(innerFields, APIClasses.MathField);
        });
      }
      return returned;
    };
    _.sLatex = function() {
      var returned = super_.sLatex.apply(this, arguments);
      if (arguments.length > 0) {
        var innerFields = this.innerFields = [];
        this.__controller.root.postOrder(function (node) {
          node.registerInnerField(innerFields, APIClasses.MathField);
        });
      }
      return returned;
    };
    _.sLatexToSVG = function() {
      var returned = super_.sLatexToSVG.apply(this, arguments);
      if (arguments.length > 0) {
        var innerFields = this.innerFields = [];
        this.__controller.root.postOrder(function (node) {
          node.registerInnerField(innerFields, APIClasses.MathField);
        });
      }
      return returned;
  };    
    _.setAriaLabel = function(ariaLabel) {
      this.__controller.ariaLabel = typeof ariaLabel === 'string' ? ariaLabel : '';
      var prependedLabel = this.__controller.ariaLabel !== 'Math Input' ? this.__controller.ariaLabel + ': ' : '';
      this.__controller.container.attr('aria-label', prependedLabel + this.__controller.root.mathspeak().trim());
      return this;
    };
    _.getAriaLabel = function () {
      return this.__controller.ariaLabel || '';
    };
  });
};

var RootMathBlock = P(MathBlock, RootBlockMixin);
API.MathField = function(APIClasses) {
  return P(APIClasses.EditableField, function(_, super_) {
    this.RootBlock = RootMathBlock;
    _.__mathquillify = function(opts, interfaceVersion) {
      this.config(opts);
      if (interfaceVersion > 1) this.__controller.root.reflow = noop;
      super_.__mathquillify.call(this, 'mq-editable-field mq-math-mode');
      delete this.__controller.root.reflow;
      return this;
    };
  });
};
