/**
 * MathQuill {VERSION}, by Han, Jeanine, and Mary
 * http://mathquill.com | maintainers@mathquill.com
 *
 * This Source Code Form is subject to the terms of the
 * Mozilla Public License, v. 2.0. If a copy of the MPL
 * was not distributed with this file, You can obtain
 * one at http://mozilla.org/MPL/2.0/.
 */

(function() {

var jQuery = window.jQuery,
  undefined,
  min = Math.min,
  max = Math.max;

if (!jQuery) throw 'MathQuill requires jQuery 1.5.2+ to be loaded first';

function noop() {}

/**
 * sugar to make defining lots of commands easier.
 * TODO: rethink this.
 */
var __slice = [].slice;
function bind(cons /*, args... */) {
  var args = __slice.call(arguments, 1);
  return function() {
    return cons.apply(this, args);
  };
}

/**
 * a development-only debug method.  This definition and all
 * calls to `pray` will be stripped from the minified
 * build of mathquill.
 *
 * This function must be called by name to be removed
 * at compile time.  Do not define another function
 * with the same name, and only call this function by
 * name.
 */
function pray(message, cond) {
  if (!cond) throw new Error('prayer failed: '+message);
}

function reflowRoundBracket(blockjQ) {
  var scaleY = blockjQ[0].getBoundingClientRect().height / parseFloat(blockjQ.css('fontSize').slice(0, -2));
  var firstTD = blockjQ.prev();
  var secondTD = blockjQ.next();
  var parens = firstTD.add(secondTD);
  var bigLeftRoundBracket = '<div style="float:left;display:inline-block;font-weight:bold;" class="mq-bracket-round mq-round-left mq-symbol-font">'+
                '<div class="mq-round-head">&#x239B;</div>'+
                '<div style="text-align: center;" class="mq-vline">&#x239C;</div>'+
                '<div class="mq-round-tail">&#x239D;</div>'+
                '<div style="clear:both;"></div></div>';
  var bigRightRoundBracket = '<div style="float:left;display:inline-block;font-weight:bold;" class="mq-bracket-round mq-round-right mq-symbol-font">'+
                '<div class="mq-round-head">&#x239E;</div>'+
                '<div style="text-align: center;" class="mq-vline">&#x239F;</div>'+
                '<div class="mq-round-tail">&#x23A0;</div>'+
                '<div style="clear:both;"></div></div>';
  if (scaleY < 6) {
      firstTD.html("(");
      secondTD.html(")");
      if(scaleY < 1.2){
        scale(parens, min(1 + .2 * (scaleY - 1), 1.2), max(scaleY, 1.3));
      } else {
        scale(parens, min(1 + .2 * (scaleY - 1), 1.2), max(scaleY, 1.3) + 0.3);
      }
      parens.css({"transform-origin": "top left"});
      parens.css("margin-top", "");
      for (var i = 0; i<2; i++) {
          if (firstTD[0].getBoundingClientRect().bottom < blockjQ[0].getBoundingClientRect().bottom + 4) {
              var oldMarginTop = +firstTD.css("margin-top").slice(0,-2);
              var newMarginTop = oldMarginTop + (blockjQ[0].getBoundingClientRect().bottom - firstTD[0].getBoundingClientRect().bottom);
              parens.css({"margin-top": newMarginTop + 2 + "px"});
          } else {
              break;
          }
      }   
      return scaleY;
  }
  parens.removeAttr("style");  
  firstTD.html(bigLeftRoundBracket);
  secondTD.html(bigRightRoundBracket);  
  var blockHeight = blockjQ[0].getBoundingClientRect().height;
  var currentHeightBigBracket = firstTD[0].getBoundingClientRect().height;
  var lineBracketHeight = firstTD.find(".mq-vline").height(); 
  var adjustLineBracket = Math.ceil(Math.abs(blockHeight - currentHeightBigBracket)/(lineBracketHeight));
  for (var i = 0; i < adjustLineBracket; i++) {
      $('<div style="text-align: center;" class="mq-vline">&#x239C;</div>').insertBefore(firstTD.find(".mq-round-tail"));  
      $('<div style="text-align: center;" class="mq-vline">&#x239F;</div>').insertBefore(secondTD.find(".mq-round-tail"));
      if (firstTD[0].getBoundingClientRect().bottom > blockjQ[0].getBoundingClientRect().bottom) {
          break;
      }
  }   
  firstTD.find(".mq-vline:last").addClass("mq-vline-pre");
  secondTD.find(".mq-vline:last").addClass("mq-vline-pre");  
  return scaleY; 
}

            //NNANH end CR#8610 
function reflowPowerSup(blockPower) {
            if(blockPower.length > 0){
              var blockJQ = $(blockPower).find(".mq-power-sup");
              var blockMove = $(blockPower);
              blockMove.css("vertical-align", "0px");
              var heightOfSup = 0;
              var topOfSup = 0;     
              if(blockJQ.find(".mq-matrix").hasClass("mq-matrix") && blockJQ.find(".mq-td").hasClass("mq-td")){
                  var lengthOfFrac = blockJQ.find(".mq-td").length;
                  blockJQ = blockJQ.find(".mq-td")[lengthOfFrac -1];
                  heightOfSup = blockJQ.getBoundingClientRect().height;
                  topOfSup = blockJQ.getBoundingClientRect().top;
                  //code here to handle the case: when sup have mq-matrix and change height the parens.
              }
              blockJQ = $(blockPower).find(".mq-power-sup");
              if(blockJQ.find(".mq-denominator").hasClass("mq-denominator")){
                  blockJQ = $(blockJQ[0].firstElementChild);
                  if(blockJQ.find(".mq-denominator").hasClass("mq-denominator")){
                    var lengthOfFrac = blockJQ.find(".mq-denominator").length;
                    blockJQ = $(blockJQ.find(".mq-denominator")[lengthOfFrac -1]);
                  }
                  heightOfSup = blockJQ[0].getBoundingClientRect().height;
                  topOfSup = blockJQ[0].getBoundingClientRect().top;
                  //code here to handle the case: when sup have fraction and change height the parens.
              }
              if(heightOfSup === 0 && topOfSup === 0){
				          //nnanh add 30/10/2018
                  if(blockJQ[0].firstElementChild){
                  heightOfSup = blockJQ[0].firstElementChild.getBoundingClientRect().height;
                  topOfSup = blockJQ[0].firstElementChild.getBoundingClientRect().top;
                } else {
                  heightOfSup = blockJQ[0].getBoundingClientRect().height;
                  topOfSup = blockJQ[0].getBoundingClientRect().top;
                }
                  //nnanh Add 30/10/2018 
              }
              if(blockPower[0].previousElementSibling){
                if($(blockPower[0].previousElementSibling).hasClass("mq-cursor")){
                  if($(blockPower[0].previousElementSibling)[0].previousElementSibling){
                    var topOfBlockParens =$(blockPower[0].previousElementSibling)[0].previousElementSibling.getBoundingClientRect().top;
                  var valueOfMove = topOfSup - topOfBlockParens + heightOfSup/3;
                  blockMove.css("vertical-align", valueOfMove);
                  }
                } else {
                  var topOfBlockParens = blockPower[0].previousElementSibling.getBoundingClientRect().top;
                  var valueOfMove = topOfSup - topOfBlockParens + heightOfSup/3;
                  blockMove.css("vertical-align", valueOfMove);
                } 
              }
            }
              
            //NNANH end CR#8610 
        };


function isIpad() {
  return navigator.userAgent.toLowerCase().indexOf("iphone") > 0 
         || navigator.userAgent.toLowerCase().indexOf("ipod") > 0
         || navigator.userAgent.toLowerCase().indexOf("ipad") > 0;
}

function isSafariV9() {
  return navigator.userAgent.indexOf("Version\/9") > 0;
}