// Parser MathBlock
var latexMathParser = (function() {
  function commandToBlock(cmd) { // can also take in a Fragment
    var block = MathBlock();
    cmd.adopt(block, 0, 0);
    return block;
  }
  function joinBlocks(blocks) {
    var firstBlock = blocks[0] || MathBlock();

    for (var i = 1; i < blocks.length; i += 1) {
      blocks[i].children().adopt(firstBlock, firstBlock.ends[R], 0);
    }

    return firstBlock;
  }

  var string = Parser.string;
  var regex = Parser.regex;
  var letter = Parser.letter;
  var any = Parser.any;
  var optWhitespace = Parser.optWhitespace;
  var succeed = Parser.succeed;
  var fail = Parser.fail;

  // Parsers yielding either MathCommands, or Fragments of MathCommands
  //   (either way, something that can be adopted by a MathBlock)
  var variable = letter.map(function(c) { return Letter(c); });
  var symbol = regex(/^[^${}\\_^]/).map(function(c) { return VanillaSymbol(c); });

  var controlSequence =
    regex(/^[^\\a-eg-zA-Z]/) // hotfix #164; match MathBlock::write
    .or(string('\\').then(
      regex(/^[a-z]+/i)
      .or(regex(/^\s+/).result(' '))
      .or(any)
    )).then(function(ctrlSeq) {
      var cmdKlass = LatexCmds[ctrlSeq];

      if (cmdKlass) {
        return cmdKlass(ctrlSeq).parser();
      }
      else {
        return fail('unknown command: \\'+ctrlSeq);
      }
    })
  ;

  var command =
    controlSequence
    .or(variable)
    .or(symbol)
  ;

  // Parsers yielding MathBlocks
  var mathGroup = string('{').then(function() { return mathSequence; }).skip(string('}'));
  var mathBlock = optWhitespace.then(mathGroup.or(command.map(commandToBlock)));
  var mathSequence = mathBlock.many().map(joinBlocks).skip(optWhitespace);

  var optMathBlock =
    string('[').then(
      mathBlock.then(function(block) {
        return block.join('latex') !== ']' ? succeed(block) : fail();
      })
      .many().map(joinBlocks).skip(optWhitespace)
    ).skip(string(']'))
  ;

  var latexMath = mathSequence;

  latexMath.block = mathBlock;
  latexMath.optBlock = optMathBlock;
  return latexMath;
})();

Controller.open(function(_, super_) {
  _.exportLatex = function() {
    return this.root.latex().replace(/(\\[a-z]+) (?![a-z])/ig,'$1');
  };
  _.exportSLatex = function() {
    return this.root.sLatex().replace(/(\\[a-z]+) (?![a-z])/ig,'$1');
  };
  _.exportSLatexToSVG= function() {
    return this.root.sLatexToSVG().replace(/(\\[a-z]+) (?![a-z])/ig,'$1');;
  };
  _.exportNativeLatex = function() {
    return this.root.nativeLatex().replace(/(\\[a-z]+) (?![a-z])/ig,'$1');
  };
  _.writeLatex = function(latex) {
    //nnanh start Bug #8609 User input the many sqrt, system is freeze.
    //NNANH 26/9/2018 Start Bug #8615 space is lost when load save "and" "xor" "or"
    //write latex for copy paste drag
    if(latex === latex + ''){
      //NNANH 12/3/2018 - Edit - Bug#8615 - E - Change suggestion with and or xor randseed and "|"
      // latex = latex.replace(/\b(xor)\b/g, "SxrS");
      // latex = latex.replace(/\b(or)\b/g, "\\quad or \\quad");
      // latex = latex.replace(/\b(SxrS)\b/g, "\\quad xor \\quad");
      // latex = latex.replace(/\b(and)\b/g, "\\quad and \\quad");
      // latex = latex.replace(/(\|)/g, " \\quad | \\quad ");
      // latex = latex.replace(/(\\quad\\quad)/g, "\\quad");
      // latex = latex.replace(/(\\quad \\quad)/g, "\\quad");
      //NNANH 12/3/2018 - Edit - Bug#8615 - E - Change suggestion with and or xor randseed and "|"

      //NNANH 12/3/2018 - Edit - Bug#11503  Change "'" When copy paste
      latex = latex.replace(/'/g, "\uEE09");
      if(latex.indexOf("sqrt[") > -1){
      var $nthrootblock = this.container.find(".mq-root-block");
      if($nthrootblock.hasClass("mq-root-block") && $nthrootblock.find(".mq-nthroot").hasClass("mq-nthroot")){
            var lengthofblock = $nthrootblock.children().length;
            var currentblock = $nthrootblock.children();
            for(var i = 0; i < lengthofblock; i++){
                $currentblock = $(currentblock[i]);
                if($currentblock.find(".mq-cursor").hasClass("mq-cursor")){
                    break;
                }
            }
            if($currentblock.find("span.mq-non-leaf.mq-nthroot").hasClass("mq-nthroot")){
                var valuecheckbreak = 4;
                if($currentblock.hasClass("mq-nthroot")){
                    valuecheckbreak = valuecheckbreak - 1;
                }
                if($currentblock.find("span.mq-non-leaf.mq-nthroot").length > valuecheckbreak){
                    return this;
                }
            }
        }
      }
    //nnanh 25/10/2018 start Bug #9189 User input the many ^, system is freeze.
    if(latex.indexOf("^") > -1){
      // limit of the number of sup contained in latex
      var limitOfSupInLatex = 5;

      // Count the number of sup contained in latex (clipboard)

      // Create temporary DOM to paste latex
      var inputFomular = document.createElement("div");
      var $inputFomular = $(inputFomular);
      $inputFomular
        .addClass("input-formular")
        .hide()
        .attr("id", "temporary-mathquill-filed")
        .append(
          '<div class="input-equation-boder"><div class="input-equation-area"><span class="expr"></span></div></div>'
        );
      $("body").append($inputFomular);
      
      var $expr = $inputFomular.find(".expr");
      var htmlElement = "<div class ='mathQuill'> </div>";
      $expr.append(htmlElement);

      // Paste latex to tempral DOM
      var mathField = Classpad.Common.getMathQuillObj($expr.find(".mathQuill"), Classpad.Const.MathQuillType.StaticMath);
      mathField.latex(latex);

      // Count the number of sup (numOfSupInLatex)
      var $fromBlock = $(mathField.el()).find('.mq-root-block');
      var numOfSupInLatex = 0;
      $fromBlock.find(".mq-power-sup-sup").each(function (index, element) {
        var length = $(element).parentsUntil(".mq-root-block", ".mq-power-sup-sup").length + 1;
        if (numOfSupInLatex < length) {
          numOfSupInLatex = length;
        }
      });

      // Remove temporal DOM
      $("#temporary-mathquill-filed").remove();

      // Return when number of sup is more than limit (=5)
      if (numOfSupInLatex > limitOfSupInLatex) {
        return this;
      }
      
      var $supExpBlock = this.container.find(".mq-root-block");
      //NNANH 2/14/2019 - S  - BUg#10502  - Edit - change condition to check mq-power
      if($supExpBlock.hasClass("mq-root-block") && $supExpBlock.find(".mq-power-sup-sup").hasClass("mq-power-sup-sup")){
          var ExpBlock = $supExpBlock.children();
          var lengthOfChildren = ExpBlock.length;
          var $currentExpSup = null;
          var lengthOfLimit = 5;
          var numOfSup = 0;
          for(var i = 0; i< lengthOfChildren; i++){
            $currentExpSup = $(ExpBlock[i]);
            if($currentExpSup.find(".mq-cursor").hasClass("mq-cursor")){

              //akae 2020/05/21 start
              numOfSup = $currentExpSup.find(".mq-cursor").parentsUntil(".mq-root-block", ".mq-power-sup-sup").length;
              numOfSup += numOfSupInLatex;
              if (numOfSup > lengthOfLimit) {
                return this;
              }
              break;
            }
            else if ($currentExpSup.find(".mq-selection").hasClass("mq-selection") || $currentExpSup.hasClass("mq-selection")) {

              var $selectionExpSup = $currentExpSup.hasClass("mq-selection") ? $currentExpSup : $currentExpSup.find(".mq-selection");
              numOfSup = $selectionExpSup.parentsUntil(".mq-root-block", ".mq-power-sup-sup").length;
              
              numOfSup += numOfSupInLatex;

              if (numOfSup > lengthOfLimit) {
                return this;
              }
              break;
            }
            //akae 2020/05/21  end
          }
      }
    }
    //nnanh 25/10/2018 end Bug #9189 User input the many ^, system is freeze.
    }
//NNANH 26/9/2018 End Bug #8615 space is lost when load save "and" "xor" "or"

    //nnanh end Bug #8609 User input the many sqrt, system is freeze.
    var cursor = this.notify('edit').cursor;

    //NNANH Start Bug#6401 enable the number for parametric
    var ENABLE_KEY_PARAMETRIC = /^(-)*(\d)+(\.)*(\d)*$/;
    if(cursor.parent.jQ.hasClass("accept-only-digit") && !ENABLE_KEY_PARAMETRIC.test(latex)){
      return;
    }
    //NNANH Start Bug#6401 enable the number for parametric
    //NNANH 12/21/2018 CR#11812 Add - S - disable event drag drop copy
    var mathquillBlock = cursor.parent.jQ;
    if(mathquillBlock && (mathquillBlock.find(".mq-selection").hasClass("mq-selection") || mathquillBlock.is(".mq-root-block")) && mathquillBlock.find(".mq-recursiveType").hasClass("mq-recursiveType")){
      return this;
    }
    //NNANH 12/21/2018 CR#11812 Add - S - disable event drag drop copy
    var all = Parser.all;
    var eof = Parser.eof;

    var block = latexMathParser.skip(eof).or(all.result(false)).parse(latex);

    if (block && !block.isEmpty()) {
      block.children().adopt(cursor.parent, cursor[L], cursor[R]);
      var jQ = block.jQize();
      jQ.insertBefore(cursor.jQ);
      cursor[L] = block.ends[R];
      block.finalizeInsert(cursor.options, cursor);
      if (block.ends[R][R].siblingCreated) block.ends[R][R].siblingCreated(cursor.options, L);
      if (block.ends[L][L].siblingCreated) block.ends[L][L].siblingCreated(cursor.options, R);
      cursor.parent.bubble(function(node) { node.reflow(); });
      /**
       * @author ThuyLTM
       * make the function name bold when copy/paste latex
       * START
       */
      if (block.ends[L] !== 0) {
        block.ends[L].traverseBoldFunctionName(cursor.options);
      }
      /**
       * @author ThuyLTM
       * make the function name bold when copy/paste latex
       * END
       */
    }

    return this;
  };
  _.renderLatexMath = function(latex) {
    var root = this.root, cursor = this.cursor;

    var all = Parser.all;
    var eof = Parser.eof;

    var block = latexMathParser.skip(eof).or(all.result(false)).parse(latex);

    root.ends[L] = root.ends[R] = 0;

    if (block) {
      block.children().adopt(root, 0, 0);
    }

    var jQ = root.jQ;

    if (block) {
      var html = block.join('html');
      jQ.html(html);
      root.jQize(jQ.children());
      root.finalizeInsert(cursor.options);
      /**
       * @author ThuyLTM
       * make the function name bold when copy/paste latex
       */
      if (root.ends[L] !== 0) {
        root.ends[L].traverseBoldFunctionName(cursor.options);
      }
    }
    else {
      jQ.empty();
    }
    var prependedLabel = this.ariaLabel && this.ariaLabel !== 'Math Input' ? this.ariaLabel + ': ' : '';
    this.container.attr('aria-label', prependedLabel + root.mathspeak().trim());

    delete cursor.selection;
    cursor.insAtRightEnd(root);
  };
  _.renderLatexText = function(latex) {
    var root = this.root, cursor = this.cursor;

    root.jQ.children().slice(1).remove();
    root.ends[L] = root.ends[R] = 0;
    delete cursor.selection;
    cursor.show().insAtRightEnd(root);

    var regex = Parser.regex;
    var string = Parser.string;
    var eof = Parser.eof;
    var all = Parser.all;

    // Parser RootMathCommand
    var mathMode = string('$').then(latexMathParser)
      // because TeX is insane, math mode doesn't necessarily
      // have to end.  So we allow for the case that math mode
      // continues to the end of the stream.
      .skip(string('$').or(eof))
      .map(function(block) {
        // HACK FIXME: this shouldn't have to have access to cursor
        var rootMathCommand = RootMathCommand(cursor);

        rootMathCommand.createBlocks();
        var rootMathBlock = rootMathCommand.ends[L];
        block.children().adopt(rootMathBlock, 0, 0);

        return rootMathCommand;
      })
    ;

    var escapedDollar = string('\\$').result('$');
    var textChar = escapedDollar.or(regex(/^[^$]/)).map(VanillaSymbol);
    var latexText = mathMode.or(textChar).many();
    var commands = latexText.skip(eof).or(all.result(false)).parse(latex);

    if (commands) {
      for (var i = 0; i < commands.length; i += 1) {
        commands[i].adopt(root, root.ends[R], 0);
      }

      root.jQize().appendTo(root.jQ);

      root.finalizeInsert(cursor.options);
    }
  };
});
