/***************************
 * Commands and Operators.
 **************************/

var scale, // = function(jQ, x, y) { ... }
    //will use a CSS 2D transform to scale the jQuery-wrapped HTML elements,
    //or the filter matrix transform fallback for IE 5.5-8, or gracefully degrade to
    //increasing the fontSize to match the vertical Y scaling factor.

    //ideas from http://github.com/louisremi/jquery.transform.js
    //see also http://msdn.microsoft.com/en-us/library/ms533014(v=vs.85).aspx

    forceIERedraw = noop,
    div = document.createElement('div'),
    div_style = div.style,
    transformPropNames = {
        transform: 1,
        WebkitTransform: 1,
        MozTransform: 1,
        OTransform: 1,
        msTransform: 1
  },
    transformPropName;

for (var prop in transformPropNames) {
    if (prop in div_style) {
        transformPropName = prop;
        break;
    }
}

if (transformPropName) {
    scale = function(jQ, x, y) {
        jQ.css(transformPropName, 'scale(' + x + ',' + y + ')');
    };
} else if ('filter' in div_style) { //IE 6, 7, & 8 fallback, see https://github.com/laughinghan/mathquill/wiki/Transforms
    forceIERedraw = function(el) { el.className = el.className; };
    scale = function(jQ, x, y) { //NOTE: assumes y > x
        x /= (1 + (y - 1) / 2);
        jQ.css('fontSize', y + 'em');
        if (!jQ.hasClass('mq-matrixed-container')) {
            jQ.addClass('mq-matrixed-container')
                .wrapInner('<span class="mq-matrixed"></span>');
        }
        var innerjQ = jQ.children()
            .css('filter', 'progid:DXImageTransform.Microsoft' +
                '.Matrix(M11=' + x + ",SizingMethod='auto expand')"
            );

        function calculateMarginRight() {
            jQ.css('marginRight', (innerjQ.width() - 1) * (x - 1) / x + 'px');
  }
        calculateMarginRight();
        var intervalId = setInterval(calculateMarginRight);
        $(window).load(function() {
            clearTimeout(intervalId);
            calculateMarginRight();
        });
    };
} else {
    scale = function(jQ, x, y) {
        jQ.css('fontSize', y + 'em');
    };
}

var Style = P(MathCommand, function(_, super_) {
  _.init = function(ctrlSeq, tagName, attrs, ariaLabel) {
    super_.init.call(this, ctrlSeq, '<'+tagName+' '+attrs+'>&0</'+tagName+'>');
    _.ariaLabel = ariaLabel || ctrlSeq.replace(/^\\/, '');
    _.mathspeakTemplate = ['Start' + _.ariaLabel + ',', 'End' + _.ariaLabel];
  };
});

//fonts
LatexCmds.mathrm = bind(Style, '\\mathrm', 'span', 'class="mq-roman mq-font"', 'Roman Font');
LatexCmds.mathit = bind(Style, '\\mathit', 'i', 'class="mq-font"', 'Italic Font');
LatexCmds.mathbf = bind(Style, '\\mathbf', 'b', 'class="mq-font"', 'Bold Font');
LatexCmds.mathsf = bind(Style, '\\mathsf', 'span', 'class="mq-sans-serif mq-font"', 'Serif Font');
LatexCmds.mathtt = bind(Style, '\\mathtt', 'span', 'class="mq-monospace mq-font"', 'Math Text');
//text-decoration
LatexCmds.underline = bind(Style, '\\underline', 'span', 'class="mq-non-leaf mq-underline"', 'Underline');
LatexCmds.overline = LatexCmds.bar = bind(Style, '\\overline', 'span', 'class="mq-non-leaf mq-overline"', 'Overline');
LatexCmds.overrightarrow = bind(Style, '\\overrightarrow', 'span', 'class="mq-non-leaf mq-overarrow mq-arrow-right"', 'Over Right Arrow');
LatexCmds.overleftarrow = bind(Style, '\\overleftarrow', 'span', 'class="mq-non-leaf mq-overarrow mq-arrow-left"', 'Over Left Arrow');
LatexCmds.overleftrightarrow = bind(Style, '\\overleftrightarrow ', 'span', 'class="mq-non-leaf mq-overarrow mq-arrow-leftright"', 'Over Left and Right Arrow');
LatexCmds.overarc = bind(Style, '\\overarc', 'span', 'class="mq-non-leaf mq-overarc"', 'Over Arc');
LatexCmds.dot = P(MathCommand, function(_, super_) {
    _.init = function() {
        super_.init.call(this, '\\dot', '<span class="mq-non-leaf"><span class="mq-dot-recurring-inner">' +
            '<span class="mq-dot-recurring">&#x2d9;</span>' +
            '<span class="mq-empty-box">&0</span>' +
            '</span></span>'
        );
    };
});

// `\textcolor{color}{math}` will apply a color to the given math content, where
// `color` is any valid CSS Color Value (see [SitePoint docs][] (recommended),
// [Mozilla docs][], or [W3C spec][]).
//
// [SitePoint docs]: http://reference.sitepoint.com/css/colorvalues
// [Mozilla docs]: https://developer.mozilla.org/en-US/docs/CSS/color_value#Values
// [W3C spec]: http://dev.w3.org/csswg/css3-color/#colorunits
var TextColor = LatexCmds.textcolor = P(MathCommand, function(_, super_) {
  _.setColor = function(color) {
    this.color = color;
    this.htmlTemplate =
      '<span class="mq-textcolor" style="color:' + color + '">&0</span>';
  };
  _.latex = function() {
    return '\\textcolor{' + this.color + '}{' + this.blocks[0].latex() + '}';
  };
    _.sLatex = function() {
        return '\\textcolor{' + this.color + '}{' + this.blocks[0].sLatex() + '}';
    };
    _.nativeLatex = function() {
        return '\\textcolor{' + this.color + '}{' + this.blocks[0].nativeLatex() + '}';
    };
  _.parser = function() {
    var self = this;
    var optWhitespace = Parser.optWhitespace;
    var string = Parser.string;
    var regex = Parser.regex;

    return optWhitespace
      .then(string('{'))
      .then(regex(/^[#\w\s.,()%-]*/))
      .skip(string('}'))
      .then(function(color) {
        self.setColor(color);
        return super_.parser.call(self);
            });
  };
  _.isStyleBlock = function() {
    return true;
  };
});

// Very similar to the \textcolor command, but will add the given CSS class.
// Usage: \class{classname}{math}
// Note regex that whitelists valid CSS classname characters:
// https://github.com/mathquill/mathquill/pull/191#discussion_r4327442
var Class = LatexCmds['class'] = P(MathCommand, function(_, super_) {
  _.parser = function() {
        var self = this,
            string = Parser.string,
            regex = Parser.regex;
    return Parser.optWhitespace
      .then(string('{'))
      .then(regex(/^[-\w\s\\\xA0-\xFF]*/))
      .skip(string('}'))
      .then(function(cls) {
        self.cls = cls || '';
                self.htmlTemplate = '<span class="mq-class ' + cls + '">&0</span>';
        return super_.parser.call(self);
            });
  };
  _.latex = function() {
    return '\\class{' + this.cls + '}{' + this.blocks[0].latex() + '}';
  };
  _.isStyleBlock = function() {
    return true;
  };
});

var SupSub = P(MathCommand, function(_, super_) {
  _.ctrlSeq = '_{...}^{...}';
  _.createLeftOf = function(cursor) {
    if (!this.replacedFragment && !cursor[L] && cursor.options.supSubsRequireOperand) return;
        if (cursor.options.restrictSubscriptLevel) {
            var currentSub = cursor.parent.parent;
            var i = 0;
            while (i < cursor.options.restrictSubscriptLevel) {
                if (!currentSub.sub) {
                    break;
                }
                i++;
                currentSub = currentSub.parent.parent;
            }
            if (i === cursor.options.restrictSubscriptLevel) return;
        }
    return super_.createLeftOf.apply(this, arguments);
  };
  _.contactWeld = function(cursor) {
    // Look on either side for a SupSub, if one is found compare my
    // .sub, .sup with its .sub, .sup. If I have one that it doesn't,
    // then call .addBlock() on it with my block; if I have one that
    // it also has, then insert my block's children into its block,
    // unless my block has none, in which case insert the cursor into
    // its block (and not mine, I'm about to remove myself) in the case
    // I was just typed.
    // TODO: simplify

    // equiv. to [L, R].forEach(function(dir) { ... });
    for (var dir = L; dir; dir = (dir === L ? R : false)) {
      if (this[dir] instanceof SupSub) {
        // equiv. to 'sub sup'.split(' ').forEach(function(supsub) { ... });
        for (var supsub = 'sub'; supsub; supsub = (supsub === 'sub' ? 'sup' : false)) {
                    var src = this[supsub],
                        dest = this[dir][supsub];
          if (!src) continue;
          if (!dest) this[dir].addBlock(src.disown());
          else if (!src.isEmpty()) { // ins src children at -dir end of dest
            src.jQ.children().insAtDirEnd(-dir, dest.jQ);
            var children = src.children().disown();
            var pt = Point(dest, children.ends[R], dest.ends[L]);
            if (dir === L) children.adopt(dest, dest.ends[R], 0);
            else children.adopt(dest, 0, dest.ends[L]);
                    } else var pt = Point(dest, 0, dest.ends[L]);
          this.placeCursor = (function(dest, src) { // TODO: don't monkey-patch
            return function(cursor) { cursor.insAtDirEnd(-dir, dest || src); };
          }(dest, src));
        }
        this.remove();
        if (cursor && cursor[L] === this) {
          if (dir === R && pt) {
            pt[L] ? cursor.insRightOf(pt[L]) : cursor.insAtLeftEnd(pt.parent);
                    } else cursor.insRightOf(this[dir]);
        }
        break;
      }
    }
  };
  Options.p.charsThatBreakOutOfSupSub = '';
  _.finalizeTree = function() {
    this.ends[L].write = function(cursor, ch) {
            if (cursor.parent.parent.sub && cursor.options.autoBreakOutOfSubscript &&
                cursor.parent.jQ.is(".mq-sub") &&
                !(ch.match(/^[A-Za-z0-9_]$/) || /^\d$/.test(ch))) {
                var currentParent = cursor.parent.parent;
                while (currentParent.sub) {
                    if (currentParent.parent.parent.sub) {
                        currentParent = currentParent.parent.parent;
                    } else {
                        currentParent = currentParent.parent;
                    }
                }
                cursor.insAtRightEnd(currentParent);
                MathBlock.p.write.apply(this, arguments);
                return;
            }
            if (cursor.parent.parent.sub && cursor.options.restrictInputOfSubscript &&
                cursor.parent.jQ.is(".mq-sub") &&
                !(ch.match(/^[A-Za-z0-9_]$/) || /^\d$/.test(ch))) {
                return;
            }
      if (cursor.options.autoSubscriptNumerals && this === this.parent.sub) {
        if (ch === '_') return;
        var cmd = this.chToCmd(ch, cursor.options);
        if (cmd instanceof Symbol) cursor.deleteSelection();
        else cursor.clearSelection().insRightOf(this.parent);
        cmd.createLeftOf(cursor.show());
        aria.queue('Baseline').alert(cmd.mathspeak({ createdLeftOf: cursor }));
        return;
      }
      if (cursor[L] && !cursor[R] && !cursor.selection
          && cursor.options.charsThatBreakOutOfSupSub.indexOf(ch) > -1) {
        cursor.insRightOf(this.parent);
        aria.queue('Baseline');
      }
      MathBlock.p.write.apply(this, arguments);
    };
  };
  _.moveTowards = function(dir, cursor, updown) {
    if (cursor.options.autoSubscriptNumerals && !this.sup) {
      cursor.insDirOf(dir, this);
        } else super_.moveTowards.apply(this, arguments);
  };
  _.deleteTowards = function(dir, cursor) {
    if (cursor.options.autoSubscriptNumerals && this.sub) {
      var cmd = this.sub.ends[-dir];
      if (cmd instanceof Symbol) cmd.remove();
      else if (cmd) cmd.deleteTowards(dir, cursor.insAtDirEnd(-dir, this.sub));

      // TODO: factor out a .removeBlock() or something
      if (this.sub.isEmpty()) {
        this.sub.deleteOutOf(L, cursor.insAtLeftEnd(this.sub));
        if (this.sup) cursor.insDirOf(-dir, this);
        // Note `-dir` because in e.g. x_1^2| want backspacing (leftward)
        // to delete the 1 but to end up rightward of x^2; with non-negated
        // `dir` (try it), the cursor appears to have gone "through" the ^2.
      }
        } else super_.deleteTowards.apply(this, arguments);
    };
    _.latex = function() {
        var result = "";
        if (this.sub) {
            var l = this.sub && this.sub.latex();
            if (l.length > 0) {
                result += '_' + this.sub.latex();
            }
        }
        if (this.sup) {
            var l = this.sup && this.sup.latex();
            if (l.length > 0) {
                //Edit 2019/10/30 ttnong start: correct out latex power expression: 2^(2) -> 2^{2}
                //result += '^' + '(' + this.sup.latex() + ')';
                result += '^' + '{' + this.sup.latex() + '}';
            }
        }
        return result;
    };
    _.sLatex = function() {
        function latex(prefix, block) {
            var l = block && block.sLatex();
            return block ? prefix + (l.length === 1 ? l : '{' + (l || ' ') + '}') : '';
    }
        return latex('_', this.sub) + latex('^', this.sup);
  };
    _.sLatexToSVG = function() {
    function latex(prefix, block) {
            var l = block && block.sLatexToSVG();
            return block ? prefix + (l.length === 1 ? l : '{' + (l || ' ') + '}') : '';
    }
    return latex('_', this.sub) + latex('^', this.sup);
  };
    _.nativeLatex = function() {
        function nativeLatex(prefix, block) {
            var l = block && block.nativeLatex();
            return block ? prefix + (l.length === 1 ? l : '{' + (l || ' ') + '}') : '';
        }
        return nativeLatex('_', this.sub) + nativeLatex('^', this.sup);
    };
  _.addBlock = function(block) {
    if (this.supsub === 'sub') {
      this.sup = this.upInto = this.sub.upOutOf = block;
      block.adopt(this, this.sub, 0).downOutOf = this.sub;
      block.jQ = $('<span class="mq-sup"/>').append(block.jQ.children()).prependTo(this.jQ);
      Node.linkElementByBlockNode(block.jQ[0], block);
    }
    else {
      this.sub = this.downInto = this.sup.downOutOf = block;
      block.adopt(this, 0, this.sup).upOutOf = this.sup;
      block.jQ = $('<span class="mq-sub"></span>').append(block.jQ.children())
        .appendTo(this.jQ.removeClass('mq-sup-only'));
      Node.linkElementByBlockNode(block.jQ[0], block);
      this.jQ.append('<span style="display:inline-block;width:0">&#8203;</span>');
    }


    // like 'sub sup'.split(' ').forEach(function(supsub) { ... });
    for (var i = 0; i < 2; i += 1) (function(cmd, supsub, oppositeSupsub, updown) {
      cmd[supsub].deleteOutOf = function(dir, cursor) {
        cursor.insDirOf((this[dir] ? -dir : dir), this.parent);
        if (!this.isEmpty()) {
          var end = this.ends[dir];
          this.children().disown()
            .withDirAdopt(dir, cursor.parent, cursor[dir], cursor[-dir])
            .jQ.insDirOf(-dir, cursor.jQ);
          cursor[-dir] = end;
        }
        cmd.supsub = oppositeSupsub;
        delete cmd[supsub];
                delete cmd[updown + 'Into'];
                cmd[oppositeSupsub][updown + 'OutOf'] = insLeftOfMeUnlessAtEnd;
        delete cmd[oppositeSupsub].deleteOutOf;
        if (supsub === 'sub') $(cmd.jQ.addClass('mq-sup-only')[0].lastChild).remove();
        this.remove();
      };
    }(this, 'sub sup'.split(' ')[i], 'sup sub'.split(' ')[i], 'down up'.split(' ')[i]));
  };
});

function insLeftOfMeUnlessAtEnd(cursor) {
  // cursor.insLeftOf(cmd), unless cursor at the end of block, and every
  // ancestor cmd is at the end of every ancestor block
    var cmd = this.parent,
        ancestorCmd = cursor;
  do {
    if (ancestorCmd[R]) return cursor.insLeftOf(cmd);
    ancestorCmd = ancestorCmd.parent.parent;
  } while (ancestorCmd !== cmd);
  cursor.insRightOf(cmd);
}

var Subscript =
LatexCmds.subscript =
LatexCmds._ = P(SupSub, function(_, super_) {
  _.supsub = 'sub';
  _.htmlTemplate =
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub">&0</span>' +
            '<span style="display:inline-block;width:0">&#8203;</span>' +
            '</span>';
        _.textTemplate = ['_'];
  _.mathspeakTemplate = [ 'Subscript,', ', Baseline'];
  _.ariaLabel = 'subscript';
  _.finalizeTree = function() {
    this.downInto = this.sub = this.ends[L];
    this.sub.upOutOf = insLeftOfMeUnlessAtEnd;
    super_.finalizeTree.call(this);
  };
        /**
         * @author ThuyLTM
         * write for make the function name bold
         */
        _.checkFunctionName = function(opts) {
            //NNANH - 4/18/2019 - S - Add - Bug#11445 check error js
            if(this.sub){
                if (this.sub.isEmpty()) {
                    MathElement.prototype.autoBoldFunctionName.call(this[L], opts);
                } else {
                    MathElement.prototype.autoBoldFunctionName.call(this.sub.ends[R], opts);
                }
            }
            
        };
        _.getFirstRightSubscript = function() {
            //NNANH - 4/18/2019 - S - Add - Bug#11445 check error js
            if (this.sub && !this.sub.isEmpty()) {
                return this.sub.ends[R];
            }
        };
        _.getFirstLeftSubscript = function() {
            //NNANH - 4/18/2019 - S - Add - Bug#11445 check error js
            if (this.sub && !this.sub.isEmpty()) {
                return this.sub.ends[L];
            }
        };
        _.getLeftOutOfSubscript = function() {
            return this[L];
        };
        _.getRightOutOfSubscript = function() {
            return this[R];
        };
    });

LatexCmds.superscript =
LatexCmds.supscript =
LatexCmds['^'] = P(SupSub, function(_, super_) {
  _.supsub = 'sup';
  _.htmlTemplate =
            '<span class="mq-power-sup-sup mq-supsub mq-non-leaf mq-sup-only">' +
            '<span class="mq-sup mq-power-sup">&0</span>' +
            '</span>';
        _.textTemplate = ['^'];
  _.mathspeakTemplate = [ 'Superscript,', ', Baseline'];
  _.ariaLabel = 'superscript';
  _.finalizeTree = function() {
    this.upInto = this.sup = this.ends[R];
    this.sup.downOutOf = insLeftOfMeUnlessAtEnd;
    super_.finalizeTree.call(this);
  };
        //NNANH start CR#8610 
        _.reflow = function(){
            // var blockJQ = this.jQ.find(".mq-power-sup");
            // var blockMove = this.jQ;
            // blockMove.css("vertical-align", "0px");
            // var heightOfSup = 0;
            // var topOfSup = 0;     
            // if(blockJQ.find(".mq-matrix").hasClass("mq-matrix") && blockJQ.find(".mq-td").hasClass("mq-td")){
            //     var lengthOfFrac = blockJQ.find(".mq-td").length;
            //     blockJQ = blockJQ.find(".mq-td")[lengthOfFrac -1];
            //     heightOfSup = blockJQ.getBoundingClientRect().height;
            //     topOfSup = blockJQ.getBoundingClientRect().top;
            //     //code here to handle the case: when sup have mq-matrix and change height the parens.
            // }
            // blockJQ = this.jQ.find(".mq-power-sup");
            // if(blockJQ.find(".mq-denominator").hasClass("mq-denominator")){
            //     var lengthOfFrac = blockJQ.find(".mq-denominator").length;
            //     blockJQ = blockJQ.find(".mq-denominator")[lengthOfFrac -1];
            //     heightOfSup = blockJQ.getBoundingClientRect().height;
            //     topOfSup = blockJQ.getBoundingClientRect().top;
            //     //code here to handle the case: when sup have fraction and change height the parens.
            // }
            // if(heightOfSup === 0 && topOfSup === 0){
            //     heightOfSup = blockJQ[0].getBoundingClientRect().height;
            //     topOfSup = blockJQ[0].getBoundingClientRect().top;
            // }
            // if(this[-1].jQ){
            //     var lengthOfParens = this[-1].jQ.length;
            //     topOfBlockParens = this[-1].jQ[lengthOfParens-1].getBoundingClientRect().top;
            //     var valueOfMove = topOfSup - topOfBlockParens + heightOfSup/3;
            //     blockMove.css("vertical-align", valueOfMove);
            // }
            reflowPowerSup(this.jQ);
            //NNANH end CR#8610 
        };
    });

var SummationNotation = P(MathCommand, function(_, super_) {
  _.init = function(ch, html, ariaLabel) {
    _.ariaLabel = ariaLabel || ctrlSeq.replace(/^\\/, '');
    var htmlTemplate =
            '<span class="mq-large-operator mq-non-leaf">' +
            '<span class="mq-to"><span>&1</span></span>' +
            '<big>' + html + '</big>' +
            '<span class="mq-from"><span>&0</span></span>' +
            '</span>';
    Symbol.prototype.init.call(this, ch, htmlTemplate);
  };
  _.createLeftOf = function(cursor) {
    super_.createLeftOf.apply(this, arguments);
    if (cursor.options.sumStartsWithNEquals) {
      Letter('n').createLeftOf(cursor);
      Equality().createLeftOf(cursor);
    }
  };
  _.latex = function() {
    function simplify(latex) {
            return latex.length === 1 ? latex : '{' + (latex || ' ') + '}';
    }
    return this.ctrlSeq + '_' + simplify(this.ends[L].latex()) +
      '^' + simplify(this.ends[R].latex());
  };
  _.mathspeak = function() {
    return 'Start ' + this.ariaLabel + ' from ' + this.ends[L].mathspeak() +
      ' to ' + this.ends[R].mathspeak() + ', end ' + this.ariaLabel + ', ';
  };
  _.parser = function() {
    var string = Parser.string;
    var optWhitespace = Parser.optWhitespace;
    var succeed = Parser.succeed;
    var block = latexMathParser.block;

    var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
    for (var i = 0; i < blocks.length; i += 1) {
      blocks[i].adopt(self, self.ends[R], 0);
    }

    return optWhitespace.then(string('_').or(string('^'))).then(function(supOrSub) {
      var child = blocks[supOrSub === '_' ? 0 : 1];
      return block.then(function(block) {
        block.children().adopt(child, child.ends[R], 0);
        return succeed(self);
      });
    }).many().result(self);
  };
  _.finalizeTree = function() {
    this.ends[L].ariaLabel = 'lower bound';
    this.ends[R].ariaLabel = 'upper bound';
    this.downInto = this.ends[L];
    this.upInto = this.ends[R];
    this.ends[L].upOutOf = this.ends[R];
    this.ends[R].downOutOf = this.ends[L];
  };
});

LatexCmds['∑'] =
LatexCmds.sum =
LatexCmds.summation = bind(SummationNotation,'\\sum ','&sum;', 'sum');

var CustomizeSummation = P(MathCommand, function(_, super_) {
    _.init = function(ch, html) {
        var htmlTemplate =
            '<span class="mq-large-operator mq-non-leaf">' +
            '<span class="mq-to"><span>&2</span></span>' +
            '<big>' + html + '</big>' +
            '<span class="mq-from">' +
            '<span>&0</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '<span>&1</span>' +
            '</span>' +
            '</span>' +
            '<span class="mq-non-leaf">' +
            '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
            '<span class="mq-non-leaf">&3</span>' +
            '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
            '</span>';
        Symbol.prototype.init.call(this, ch, htmlTemplate);
    };
    _.latex = function() {
        return this.ctrlSeq + '(' + this.blocks[3].latex() + "," + this.blocks[0].latex() +
            "," + this.blocks[1].latex() + "," + this.blocks[2].latex() + ")";
    };
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\sum ^{3}_{n=1}\left( x\right)
        //\sum_{=}^{}{{()}}
        return optWhitespace.then(string('_').or(string('^')).or(string('{')).or(string('\\left('))).then(function(supOrSub) {
            if (supOrSub === "_") {
                return block.then(function(block) {
                    var firstBlock = block.ends[L];
                    var endBlock = block.ends[R];
                    var splitBlock = firstBlock[R];
                    while (splitBlock && splitBlock.ctrlSeq !== '=') {
                        splitBlock = splitBlock[R];
                    }
                    if (splitBlock) {
                        splitBlock = splitBlock[L];
                    }                   
                    var nextBlock = undefined;
                    if (splitBlock && splitBlock[R] && splitBlock[R][R]) {
                        nextBlock = splitBlock[R][R];
                    }
                    if (splitBlock) {
                        Fragment(firstBlock, splitBlock).adopt(blocks[0], blocks[0].ends[R], 0);
                        if (nextBlock) {
                            Fragment(nextBlock, endBlock).adopt(blocks[1], blocks[1].ends[R], 0);
                        }
                    } else {
                        //the left side is not null
                        if (firstBlock.ctrlSeq === '=' && endBlock !== 0 && endBlock.ctrlSeq !== '=') {
                            Fragment(firstBlock[R], endBlock).adopt(blocks[1], blocks[1].ends[R], 0);
                        }
                    }
                    return succeed(self);
                });
            } else if (supOrSub === "^") {
                return block.then(function(block) {
                    var child = blocks[2];
                    block.children().adopt(child, child.ends[R], 0);
                    return succeed(self);
                });
            } else if (supOrSub === "\\left(") {
                  return latexMathParser.then(function(block) {
                       return string('\\right)').map(function(end) {
                           var child = blocks[3];
                           var firstBlock = block.ends[L]; 
                           var endBlock = block.ends[R]; 
                           if (firstBlock && firstBlock !== 0 && 
                                endBlock && endBlock !== 0) {
                                Fragment(firstBlock, endBlock).adopt(child, child.ends[R], 0);
                           }
                           return succeed(self);
                       });
                  }).or(latexMathParser.then(function(block) {
                      var child = blocks[3];
                           var firstBlock = block.ends[L]; 
                           var endBlock = block.ends[R]; 
                           if (firstBlock && firstBlock !== 0 && 
                                endBlock && endBlock !== 0) {
                                Fragment(firstBlock, endBlock).adopt(child, child.ends[R], 0);
                           }
                           return succeed(self);
                  }));
            } else {
                return block.then(function(block) {
                        var child = blocks[3];
                        var firstBlock = block.ends[L][R]; //remove (
                        var endBlock = block.ends[R][L]; //remove )
                        if (firstBlock && firstBlock !== 0 && firstBlock.ctrlSeq !== ")" && 
                            endBlock && endBlock !== 0 && endBlock.ctrlSeq !== "(") {
                            Fragment(firstBlock, endBlock).adopt(child, child.ends[R], 0);
                        }
                        return succeed(self);
                    })
                    .then(string('}'));
            }

        }).atMost(3).result(self);
    };
    _.reflow = function() {
        reflowRoundBracket(this.blocks[3].jQ);
    };
    _.finalizeTree = function() {
        this.ends[L][R].upOutOf = this.ends[L][R][R];
        this.ends[L][R][R].downOutOf = this.ends[L];
        this.ends[L][R][R][R].upOutOf = this.ends[L][R][R];
        this.ends[L][R][R][R].downOutOf = this.ends[L];
    };
});
LatexCmds['\uEE51'] =
    LatexCmds.sum =
    LatexCmds.summation = P(CustomizeSummation, function(_, super_) {
        _.init = function() {
            CustomizeSummation.prototype.init.call(this, '\\xee\\x51', '&#xEE51');
        };
        _.sLatex = function() {
            return "\\sum" + '_{' + this.blocks[0].sLatex() + "=" +
                this.blocks[1].sLatex() + "}^{" + this.blocks[2].sLatex() +
                "}{{(" + this.blocks[3].sLatex() + ')}}';
        };
        _.sLatexToSVG = function() {
            return "\\sum" + '_{' + this.blocks[0].sLatexToSVG() + "=" +
                this.blocks[1].sLatexToSVG() + "}^{" + this.blocks[2].sLatexToSVG() +
                "}{{(" + this.blocks[3].sLatexToSVG() + ')}}';
        };
        _.nativeLatex = function() {
            return "\\sum" + '_{' + this.blocks[0].nativeLatex() + "=" +
                this.blocks[1].nativeLatex() + "}^{" + this.blocks[2].nativeLatex() +
                "}{{(" + this.blocks[3].nativeLatex() + ')}}';
        };
    });
LatexCmds['∏'] =
    LatexCmds.prod =
    LatexCmds.product = P(CustomizeSummation, function(_, super_) {
        _.init = function() {
            CustomizeSummation.prototype.init.call(this, '\\xee\\x52', '&#xEE52');
        };
        _.sLatex = function() {
            return "\\prod" + '_{' + this.blocks[0].sLatex() + "=" +
                this.blocks[1].sLatex() + "}^{" + this.blocks[2].sLatex() +
                "}{{(" + this.blocks[3].sLatex() + ')}}';
        };
        _.sLatexToSVG = function() {
            return "\\prod" + '_{' + this.blocks[0].sLatexToSVG() + "=" +
                this.blocks[1].sLatexToSVG() + "}^{" + this.blocks[2].sLatexToSVG() +
                "}{{(" + this.blocks[3].sLatexToSVG() + ')}}';
        };
        _.nativeLatex = function() {
            return "\\prod" + '_{' + this.blocks[0].nativeLatex() + "=" +
                this.blocks[1].nativeLatex() + "}^{" + this.blocks[2].nativeLatex() +
                "}{{(" + this.blocks[3].nativeLatex() + ')}}';
        };
    });

LatexCmds.coprod =
LatexCmds.coproduct = bind(SummationNotation,'\\coprod ','&#8720;', 'co product');

LatexCmds['∫'] =
    LatexCmds['int'] =
    LatexCmds['integral'] =
    LatexCmds.integral = P(MathCommand, function(_, super_) {
  _.init = function() {
    _.ariaLabel = 'integral';
    var htmlTemplate =
           '<span class="mq-new-int mq-non-leaf">' +           
                '<span style="display:block;text-align: left;"><span class="mq-new-int-supsub">&1</span></span>'+
                '<span style="display:block;text-align:left;" class="mq-int"><big>&#xEE53</big></span>' +
                '<span style="display:block;float: right;width: 100%;margin-top: -0.5em;text-align: left;"><span class="mq-new-int-supsub">&0</span></span>'+
            '</span>' +
            '<span class="mq-non-leaf">' +
                '<span class="mq-non-leaf">&2</span>' +
                '<var>d</var>' +
                '<span class="mq-non-leaf">&3</span>' +
                '</span>';
                
            Symbol.prototype.init.call(this, '\\int', htmlTemplate, 'integral');
  };
  // FIXME: refactor rather than overriding
  _.createLeftOf = MathCommand.p.createLeftOf;
        _.latex = function() { //post to Server
            if (this.blocks[0].latex() == "" && this.blocks[1].latex() == "") {
                return '\\xee\\x53' + '(' + this.blocks[2].latex() + "," +
                    this.blocks[3].latex() + ')';
            }
            return '\\xee\\x53' + '(' + this.blocks[2].latex() + "," +
                this.blocks[3].latex() + "," + this.blocks[0].latex() +
                "," + this.blocks[1].latex() + ')';
        };
        _.sLatex = function() { //original
            // return "\\int" + '_{' + this.blocks[0].sLatex() + "}^{" +
            //     this.blocks[1].sLatex() + "}{{{" + this.blocks[2].sLatex() +
            //     "}d{" + this.blocks[3].sLatex() + '}}}';
            return "\\int" + '_{' + this.blocks[0].sLatex() + "}^{" +
                this.blocks[1].sLatex() + "}{" + this.blocks[2].sLatex() +
                "}d{" + this.blocks[3].sLatex() + '}';
        };
        _.sLatexToSVG = function() {
            return "\\int" + '_{' + this.blocks[0].sLatexToSVG() + "}^{" +
                this.blocks[1].sLatexToSVG() + "}{{" + this.blocks[2].sLatexToSVG() +
                "d" + this.blocks[3].sLatexToSVG() + '}}';
        };
        _.nativeLatex = function() { //original
            // return "\\int" + '_{' + this.blocks[0].nativeLatex() + "}^{" +
            //     this.blocks[1].nativeLatex() + "}{{" + this.blocks[2].nativeLatex() +
            //     "d" + this.blocks[3].nativeLatex() + '}}';
            return "\\int" + '_{' + this.blocks[0].sLatex() + "}^{" +
                this.blocks[1].sLatex() + "}{" + this.blocks[2].sLatex() +
                "}d{" + this.blocks[3].sLatex() + '}';
        };
        _.parser = function() {
            var string = Parser.string;
            var optWhitespace = Parser.optWhitespace;
            var succeed = Parser.succeed;
            var block = latexMathParser.block;

            var self = this;
            var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock(), MathBlock()];
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            //NNANH - 4/9/2019 - End - Edit - change parser data integral Bug#10472 
            //\int ^{3}_{2}xdx+3
            //\int_{}^{}{{d}}
            // return optWhitespace.then(string('_').or(string('^')).or(string('{')).or(latexMathParser)).then(function(supOrSub) {
            //     if (supOrSub === "{") {
            //         return block.then(function(block) {
            //             var firstBlock = block.ends[L];
            //             var endBlock = block.ends[R];
            //             var splitBlock = firstBlock[R];
            //             while (splitBlock && splitBlock.ctrlSeq !== 'd') {
            //                 splitBlock = splitBlock[R];
            //             }
            //             splitBlock = splitBlock[L];
            //             var nextBlock = undefined;
            //             if (splitBlock && splitBlock[R] && splitBlock[R][R]) {
            //                 nextBlock = splitBlock[R][R];
            //             }
            //             if (splitBlock) {
            //                 Fragment(firstBlock, splitBlock).adopt(blocks[2], blocks[2].ends[R], 0);
            //                 if (nextBlock) {
            //                     Fragment(nextBlock, endBlock).adopt(blocks[3], blocks[3].ends[R], 0);
            //                 }
            //             } else {
            //                 if (firstBlock === endBlock && firstBlock && firstBlock !== 0 && firstBlock.ctrlSeq === 'd') {
            //                     return succeed(self);
            //                 }
            //                 if (firstBlock !== endBlock && firstBlock && firstBlock !== 0 && firstBlock.ctrlSeq === 'd') {
            //                     // Fragment(endBlock, endBlock).adopt(blocks[3], blocks[3].ends[R], 0);
            //                     //NNANH - S - Add - 3/4/2019 Bug#10472 parser in integral
            //                     Fragment(firstBlock[R], endBlock).adopt(blocks[3], blocks[3].ends[R], 0);
            //                 } else if (firstBlock && firstBlock !== 0 &&
            //                     endBlock && endBlock !== 0 && firstBlock.ctrlSeq !== 'd' && endBlock.ctrlSeq !== 'd') {
            //                     Fragment(firstBlock, endBlock).adopt(blocks[2], blocks[2].ends[R], 0);
            //                 }
            //             }
            //             return succeed(self);
            //         }).then(string('}'));
            //     } else if (supOrSub === "^") {
            //         return block.then(function(block) {
            //             var child = blocks[1];
            //             block.children().adopt(child, child.ends[R], 0);
            //             return succeed(self);
            //         });
            //     } else if (supOrSub === "_"){
            //         return block.then(function(block) {
            //             var child = blocks[0];
            //             block.children().adopt(child, child.ends[R], 0);
            //             return succeed(self);
            //         });
            //     }
            //     else {
            //         var firstBlock = supOrSub.ends[L];
            //         var endBlock = supOrSub.ends[R];
            //         var splitBlock = firstBlock[R];
            //         while (splitBlock && splitBlock.ctrlSeq !== 'd') {
            //             splitBlock = splitBlock[R];
            //         }
            //         if (splitBlock && splitBlock !== 0) {
            //              splitBlock = splitBlock[L];
            //         }                   
            //         var nextBlock = undefined;
            //         if (splitBlock && splitBlock[R] && splitBlock[R][R]) {
            //             nextBlock = splitBlock[R][R];
            //         }
            //         if (splitBlock) {
            //             Fragment(firstBlock, splitBlock).adopt(blocks[2], blocks[2].ends[R], 0);
            //             if (nextBlock) {
            //                 Fragment(nextBlock, endBlock).adopt(blocks[3], blocks[3].ends[R], 0);
            //             }
            //         } else {
            //             if (firstBlock && firstBlock !== 0 && firstBlock === endBlock && firstBlock.ctrlSeq === 'd') {
            //                 return succeed(self);
            //             }
            //             if (firstBlock !== endBlock && firstBlock && firstBlock !== 0 && firstBlock.ctrlSeq === 'd') {
            //                 Fragment(endBlock, endBlock).adopt(blocks[3], blocks[3].ends[R], 0);
            //             } else if (firstBlock && firstBlock !== 0 &&
            //                 endBlock && endBlock !== 0 && firstBlock.ctrlSeq !== 'd' && endBlock.ctrlSeq !== 'd') {
            //                 Fragment(firstBlock, endBlock).adopt(blocks[2], blocks[2].ends[R], 0);
            //             }
            //         }
            //         return succeed(self);
            //     }

            // }).atMost(3).result(self);
            //\int_{}^{}{}d{}
            return (optWhitespace.then(string('_').or(string('^')).or(string('d'))).then(function(supOrSub) {
                if (supOrSub === "^") {
                    return block.then(function(block) {
                        var child = blocks[1];
                        block.children().adopt(child, child.ends[R], 0);
                        return succeed(self);
                    });
                } else if (supOrSub === "_"){
                    return block.then(function(block) {
                        var child = blocks[0];
                        block.children().adopt(child, child.ends[R], 0);
                        return succeed(self);
                    });
                } else if (supOrSub === "d"){
                    return block.then(function(block) {
                        var child = blocks[3];
                        block.children().adopt(child, child.ends[R], 0);
                        return succeed(self);
                    });
                }

            })).or(                
                block.then(function(block) {
                    block.children().adopt(blocks[2], blocks[2].ends[R], 0);
                return succeed(self);
            })
        ).atMost(4).result(self);
        };
        //NNANH - 4/9/2019 - End - Edit - change parser data integral Bug#10472 
    });
//Hoang write more 160 functions
var AllFunction = P(MathCommand, function(_, super_) {
    _.init = function(name, htmlCode, ctrlSeq, parameter, optPara, hint) {
        this.htmlCode = htmlCode;
        this.name = name;
        this.parameter = parameter;
        this.countOp = optPara;
        if (hint == null) {
            hint = [];
        }
        if (hint.length != parameter + optPara) {
            for (var i = 0; i < parameter + optPara; i++) {
                hint[i] = "";
            }
        }
        var htmlTemplate =
            '<var>' + htmlCode + '</var>' +
            '<span class="mq-int mq-non-leaf">' +
            '<span class="mq-scaled mq-paren">(</span>' +
            '<span class="mq-non-leaf">';
        var k = 0;
        if (parameter > 0) {
            htmlTemplate += '<span class="mq-scaled  ph_func active" data-placeholder="' + hint[k] + '">&' + k + '</span>';
            k++;
        }
        if (parameter > 1) {
            for (var i = 0; i < parameter - 1; i++) {
                htmlTemplate += '<var>, </var>' +
                    '<span class="mq-scaled  ph_func active" data-placeholder="' + hint[k] + '">&' + k + '</span>';
                k++;
            }
        }
        for (var i = 0; i < optPara; i++) {
            htmlTemplate += '<var> </var>' +
                '<span class="mq-scaled  ph_func active" data-placeholder="[,' + hint[k] + ']">&' + k + '</span>';
            k++;
        }
        htmlTemplate += '</span>' +
            '<span class="mq-scaled mq-paren">)</span>' +
            '</span>';
        Symbol.prototype.init.call(this, ctrlSeq, htmlTemplate);
    };
    _.latex = function() {
        var result = this.ctrlSeq + '(';
        var k = 0;
        for (var i = 0; i < this.blocks.length - this.countOp; i++) {
            result += this.blocks[k].latex() + ',';
            k++;
        }
        result = result.slice(0, -1);
        for (var j = 0; j < this.countOp; j++) {
            var temp = this.blocks[k].latex();
            result += temp;
            k++;
        }
        result += ')';
        return result;
    };
    _.sLatex = function() {
        var result = "\\" + this.name + '$${(';
        var k = 0;
        var fixBlocks = [];
        for (var i = 0; i < this.parameter; i++) {
            fixBlocks.push(this.blocks[k].sLatex());
            k++;
        }
        result += "{" + fixBlocks.join("},{") + "}";
        var optBlocks = [];
        for (var j = 0; j < this.countOp; j++) {
            optBlocks.push(this.blocks[k].sLatex());
            k++;
        }
        result += "[{" + optBlocks.join("},{") + "}]";
        result += ')}$$';
        return result;       
    };
    _.nativeLatex = function() {
        var functionName = this.htmlCode;         
        switch(this.name) {
            case "finvsin":
                functionName = "\\arcsin"
                break;
            case "finvcos":
                functionName = "\\arccos"
                break;
            case "finvtan":
                functionName = "\\arctan"
                break;
            case "finvsinh":
                functionName = "\\arcsinh"
                break;
           case "finvcosh":
                functionName = "\\arccosh"
                break;
            case "finvtanh":
                functionName = "\\arctanh"
                break;
        }
        var result = functionName + '\\left(';
        var k = 0;
        var fixBlocks = [];
        for (var i = 0; i < this.parameter; i++) {
            fixBlocks.push(this.blocks[k].nativeLatex());
            k++;
        }
        result += fixBlocks.join(",");       
        for (var j = 0; j < this.countOp; j++) {            
            result += this.blocks[k].nativeLatex();
            k++;
        }        
        result += '\\right)';
        return result;       
    };
    _.parser = function() {
        var self = this;
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var blockData = [];
        //\taylor$${({},{},{}[{}])}$$
        //\taylor{{({},{},{}[{}])}}
        //\taylor\left(x,x,y,5\right)
        return (optWhitespace.then(string('$${').or(string('}$$')).or(string('{')).or(string('}').or(string('\\left(')))).then(function(barrierSymbol) {
            if (barrierSymbol === "$${" || barrierSymbol === "{") {
                return optWhitespace
                .then(latexMathParser.block)
                .many()
                .skip(optWhitespace)
                .then(function(items) {
                    var lastIndex = items.length - 1;
                    self.blocks = [];
                    var paraIndex = 1;
                    for (var i = 0; i < self.parameter; i++) {
                        var node = items[paraIndex];
                        self.blocks.push(node);
                        node.adopt(self, self.ends[R], 0);
                        paraIndex += 2;
                    }
                    var optIndex = self.parameter * 2 + 1;
                    for (var i = 0; i < self.countOp; i++) {
                        var node = items[optIndex];
                        if (optIndex === lastIndex) {
                            break;
                        }
                        self.blocks.push(node);
                        node.adopt(self, self.ends[R], 0);
                        optIndex += 2;
                    }
                    return Parser.succeed(self);
                });
            } else if (barrierSymbol === "\\left(") {
                return latexMathParser.then(function(block) {
                    blockData.push(block);
                    return (string('\\right)').map(function(end) {
                        var blocks = self.blocks = [];  
                        for (var i = 0; i < self.numBlocks(); i++) {
                            blocks.push(MathBlock());
                        }
                        
                        var startBlock = block.ends[L];
                        if (startBlock !== 0) {
                            var i = 0;
                            while (startBlock && startBlock != 0 && i < self.numBlocks()) {
                                if (startBlock.ctrlSeq === "," && i < self.parameter - 1) {
                                     blocks[i].adopt(self, self.ends[R], 0);
                                     startBlock = startBlock[R];
                                     i++;
                                     continue;
                                }
                                var endBlock = startBlock[R];
                                var leftEndBlock = endBlock[L];
                                while (endBlock !== 0 && endBlock.ctrlSeq !== ",") {
                                    leftEndBlock = endBlock;
                                    endBlock = endBlock[R];
                                }                               
                                if (endBlock !== 0 && endBlock.ctrlSeq === ",") {
                                    Fragment(startBlock, endBlock[L]).adopt(blocks[i], blocks[i].ends[R], 0);
                                     blocks[i].adopt(self, self.ends[R], 0);
                                    if (i < self.parameter - 1) {
                                        startBlock = endBlock[R];
                                    } else {
                                        startBlock = endBlock;
                                    }
                                    i++;
                                    continue;
                                } else if (leftEndBlock) {
                                    Fragment(startBlock, leftEndBlock).adopt(blocks[i], blocks[i].ends[R], 0);
                                     blocks[i].adopt(self, self.ends[R], 0);
                                    startBlock = 0;
                                    i++;
                                    continue;
                                } else {
                                    Fragment(startBlock, startBlock).adopt(blocks[i], blocks[i].ends[R], 0);
                                     blocks[i].adopt(self, self.ends[R], 0);
                                    startBlock = 0;
                                    i++;
                                    continue;
                                }
                            }
                            for (var j = i; j < blocks.length; j++) {
                                 blocks[j].adopt(self, self.ends[R], 0);
                            }
                        } else {
                            for (var i = 0; i < blocks.length; i += 1) {
                               blocks[i].adopt(self, self.ends[R], 0);
                            }
                        }
                        return self;
                    })).or(succeed(self));
                });
            } else {
                return Parser.succeed(self);
            }            
        })).or(optWhitespace.then(function() {
            if (self.blocks && self.blocks.length === self.numBlocks()) {
                return succeed(self);
            }
            var blocks = self.blocks = [];
            for (var i = 0; i < self.numBlocks(); i++) {
                blocks.push(MathBlock());
            }
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            if (blockData.length > 0) {
                var startBlock = blockData[0].ends[L];
                if (startBlock !== 0) {
                    var i = 0;
                    while (startBlock && startBlock != 0 && i < self.numBlocks()) {
                        if (startBlock.ctrlSeq === "," && i < self.parameter - 1) {                            
                            startBlock = startBlock[R];
                            i++;
                            continue;
                        }
                        var endBlock = startBlock[R];
                        var leftEndBlock = endBlock[L];
                        while (endBlock !== 0 && endBlock.ctrlSeq !== ",") {
                            leftEndBlock = endBlock;
                            endBlock = endBlock[R];
                        }                               
                        if (endBlock !== 0 && endBlock.ctrlSeq === ",") {
                            Fragment(startBlock, endBlock[L]).adopt(blocks[i], blocks[i].ends[R], 0);
                            if (i < self.parameter - 1) {
                                startBlock = endBlock[R];
                            } else {
                                startBlock = endBlock;
                            }
                            i++;
                            continue;
                        } else if (leftEndBlock) {
                            Fragment(startBlock, leftEndBlock).adopt(blocks[i], blocks[i].ends[R], 0);
                            startBlock = 0;
                            i++;
                            continue;
                        } else {
                            Fragment(startBlock, startBlock).adopt(blocks[i], blocks[i].ends[R], 0);
                            startBlock = 0;
                            i++;
                            continue;
                        }
                    }
                }
            }
             return succeed(self);
        })).atMost(2).result(self);
    };
    // _.reflow = function() {   
    //   var i = this.parameter;
    //   for (var j=0; j<this.optPara; j++) {
    //     var currentBlock = this.blocks[i+j].jQ;
    //     var height = currentBlock.parent().outerHeight()/ parseFloat(currentBlock.css('fontSize').slice(0,-2));
    //     var parens = currentBlock.siblings(".mq-paren");
    //     if (parens.length) {
    //       var sy = 1.2*height;
    //       scale(parens, min(1 + .2*(height - 1), 1.2), sy);          
    //     } 
    //   }
    //   var delimjQs = this.jQ.children(':first').add(this.jQ.children(':last'));
    //   var contentjQ = this.jQ.children(':eq(1)');
    //   var height = contentjQ.outerHeight() / parseFloat(contentjQ.css('fontSize'));
    //   scale(delimjQs, min(1 + .2*(height - 1), 1.2), 1.2*height);
    // };
    _.reflow = function() {
        var count = this.jQ.find(".ph_func.mq-hasCursor").children().length;
        if (count != 1) {
            this.jQ.find(".mq-cursor").show();
        }
        reflowRoundBracket(this.jQ.children(':eq(1)'));   
    };
    _.sLatexToSVG = function() {
        var result = this.htmlCode + '\\left(';
        var k = 0;
        var fixBlocks = [];
        for (var i = 0; i < this.parameter; i++) {
            fixBlocks.push(this.blocks[k].sLatexToSVG());
            k++;
        }
        result += "{" + fixBlocks.join("},{") + "}";
        if (this.countOp && this.countOp > 0) {
            var optBlocks = [];
            for (var j = 0; j < this.countOp; j++) {
                if (this.blocks[k].sLatexToSVG()) {
                    optBlocks.push(this.blocks[k].sLatexToSVG());
                }
                k++;
            }
            if (optBlocks.length > 0) {
                result += "{" + optBlocks.join("") + "}";
            }
        }
        result += '\\right)';
        return result;
    };
});

var SymbolFunction = P(AllFunction, function(_, super_) {
    _.init = function(ctrlSeq, htmlCode, parameter, optPara) {
        this.ctrlSeq = ctrlSeq;
        this.name = ctrlSeq;
        this.htmlCode = htmlCode;
        this.parameter = parameter;
        this.optPara = optPara;
        var htmlTemplate =
            '<var>' + htmlCode + '</var>' +
            '<span class="mq-int mq-non-leaf">' +
            '<span class="mq-scaled mq-paren">(</span>' +
            '<span class="mq-non-leaf">';
        var k = 0;
        if (parameter > 0) {
            htmlTemplate += '<span class="mq-scaled  ">&' + k + '</span>';
            k++;
        }
        if (parameter > 1) {
            for (i = 0; i < parameter - 1; i++) {
                htmlTemplate += '<var>, </var>' +
                    '<span class="mq-scaled ">&' + k + '</span>';
                k++;
            }
        }
        for (var j = 0; j < optPara; j++) {
            htmlTemplate += '<var> </var>' +
                '<span class=" mq-scaled">' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren">[</span>' +
                '<span class="mq-non-leaf">' +
                '<var>,</var>' +
                '<span>&' + k + '</span>' +
                '</span>' +
                '<span class="mq-scaled mq-paren">]</span>' +
                '</span>' +
                '</span>';
            k++;
        }
        htmlTemplate += '</span>' +
            '<span class="mq-scaled mq-paren">)</span>' +
            '</span>';
        Symbol.prototype.init.call(this, ctrlSeq, htmlTemplate);
    };
    _.latex = function() {
        var result = this.ctrlSeq + '(';
        var k = 0;
        for (var i = 0; i < this.parameter; i++) {
            result += this.blocks[i].latex();
            if (i < this.parameter - 1) {
                result += ",";
            }
        }
        for (var j = 0; j < this.optPara; j++) {
            result += this.blocks[i].latex();
            if (i < this.optPara - 1) {
                result += ",";
            }
        }
        result += ')';
        return OPEN_BRACKET_SERVER + result + CLOSE_BRACKET_SERVER;
    };
    _.parser = function() {
        var block = latexMathParser.block;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var string = Parser.string;
        var self = this;
        //\Gamma\left(0\right)
        //\Gamma
        //\gamma{x}
        //\gamma{({x}[{}])} old paper
        return (string("\\left(").then(function() {
            return latexMathParser.then(function(block) {
                return string('\\right)').map(function(end) {
                    var blocks = self.blocks = [MathBlock()];  
                    block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                    blocks[0].adopt(self, self.ends[R], 0);
                    return self;
                });
            });
        })).or(
            string("{").then(string('(')).then(block)
            .map(function(block) {
                var blocks = self.blocks = [MathBlock()];  
                block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                blocks[0].adopt(self, self.ends[R], 0);
                return self;
            })
            .then(string('[').then(block).map(function(block) {
                return self;
            }).then(string(']'))).then(string(')')).then(string('}')).result(self)
        ).or(block.times(1).map(function(block) { 
            var blocks = self.blocks = [MathBlock()];  
            block[0].children().adopt(blocks[0], blocks[0].ends[R], 0);
            blocks[0].adopt(self, self.ends[R], 0);
            return self;
        })).or(optWhitespace.then(function() {
            var blocks = self.blocks = [];
            for (var i = 0; i < self.numBlocks(); i++) {
                blocks.push(MathBlock());
            }
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            return succeed(self);
        }));
    };
    _.sLatexToSVG = function() {
        var result = this.htmlCode + '\\left(';
        var k = 0;
        for (var i = 0; i < this.parameter; i++) {
            result += this.blocks[i].sLatexToSVG();
            if (i < this.parameter - 1) {
                result += ",";
            }
        }
        for (var j = 0; j < this.optPara; j++) {
            result += this.blocks[i + j].sLatexToSVG();
            if (j < this.optPara - 1) {
                result += ",";
            }
        }
        result += '\\right)';
        return result;
    };
});

LatexCmds['heaviside'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, 'heaviside', '\uECA8', 1, 0);
    };
    _.sLatex = function() {
        return "\\heaviside{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\heaviside{" + this.blocks[0].nativeLatex() + "}";
    };
});
LatexCmds['invsin'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, 'sin\uEDFC', 'sin\uEDFC', 1, 0);
    };
    _.sLatex = function() {
        return "\\invsin{" + this.blocks[0].sLatex() + "}";
    };
     _.nativeLatex = function() {
        return "\\arcsin\\left(" + this.blocks[0].nativeLatex() + "\\right)";
    };
});
LatexCmds['invcos'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, 'cos\uEDFC', 'cos\uEDFC', 1, 0);
    };
    _.sLatex = function() {
        return "\\invcos{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\arccos\\left(" + this.blocks[0].nativeLatex() + "\\right)";
    };
});
LatexCmds['invtan'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, 'tan\uEDFC', 'tan\uEDFC', 1, 0);
    };
    _.sLatex = function() {
        return "\\invtan{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\arctan\\left(" + this.blocks[0].nativeLatex() + "\\right)";
    };
});
LatexCmds['invsinh'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, 'sinh\uEDFC', 'sinh\uEDFC', 1, 0);
    };
    _.sLatex = function() {
        return "\\invsinh{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\arcsinh\\left(" + this.blocks[0].nativeLatex() + "\\right)";
    };
});
LatexCmds['invcosh'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, 'cosh\uEDFC', 'cosh\uEDFC', 1, 0);
    };
    _.sLatex = function() {
        return "\\invcosh{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\arccosh\\left(" + this.blocks[0].nativeLatex() + "\\right)";
    };
});
LatexCmds['invtanh'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, 'tanh\uEDFC', 'tanh\uEDFC', 1, 0);
    };
    _.sLatex = function() {
        return "\\invtanh{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\arctanh\\left(" + this.blocks[0].nativeLatex() + "\\right)";
    };
});
LatexCmds.Gamma =
LatexCmds['gamma'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, '\\gamma', '\uEC69', 1, 0);
    };
    _.sLatex = function() {
        return "\\gamma{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\Gamma\\left(" + this.blocks[0].nativeLatex() + "\\right)";
    };
});
LatexCmds['delta'] = P(SymbolFunction, function(_, supper_) {
    _.init = function() {
        SymbolFunction.prototype.init.call(this, '\\delta', '\uED6A', 1, 0);
    };
    _.sLatex = function() {
        return "\\delta{" + this.blocks[0].sLatex() + "}";
    };
    _.nativeLatex = function() {
        return "\\delta{" + this.blocks[0].nativeLatex() + "}";
    };
});
LatexCmds['deltab'] = P(MathCommand, function(_, supper_) {
    _.init = function() {
        var htmlTemplate =
            '<span class="mq-int mq-non-leaf">' +
            '<span>&#xED6A;</span>' +
            '<span class="mq-supsub mq-non-leaf" style="vertical-align: -0.2em;">' +
            '<span class="mq-sup">' +
            '<span class="mq-scaled mq-paren">(</span>' +
            '<span class="mq-non-leaf">&0</span>' +
            '<span class="mq-scaled mq-paren">)</span>' +
            '</span>' +
            '<span style="display:inline-block;width:0">&#8203</span>' +
            '</span>' +
            '</span>' +
            '<span class="mq-non-leaf">' +
            '<span class="mq-scaled mq-paren">(</span>' +
            '<span class="mq-non-leaf">&1</span>' +
            '<span class="mq-scaled mq-paren">)</span>' +
            '</span>';
        Symbol.prototype.init.call(this, '\\delta', htmlTemplate);
    };
    _.latex = function() {
        if (this.blocks[0].latex() == "") {
            return OPEN_BRACKET_SERVER + this.ctrlSeq + '(' + this.blocks[1].latex() + ')' + CLOSE_BRACKET_SERVER;
        }
        return OPEN_BRACKET_SERVER + this.ctrlSeq + '(' + this.blocks[1].latex() + "," +
            this.blocks[0].latex() + ')' + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\deltab' + '^{(' + this.blocks[0].sLatex() + ')}${(' + this.blocks[1].sLatex() + ")}$";
    };
    _.nativeLatex = function() {
        return '\\deltab' + '^{(' + this.blocks[0].nativeLatex() + ')}${(' + this.blocks[1].nativeLatex() + ")}$";
    };
    _.sLatexToSVG = function() {
        return '\uED6A' + '^{\\left(' + this.blocks[0].sLatexToSVG() + '\\right)}{\\left(' + this.blocks[1].sLatexToSVG() + "\\right)}";
    };
    _.reflow = function() {
        reflowRoundBracket(this.blocks[1].jQ);   
        reflowRoundBracket(this.blocks[0].jQ);   
    };
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\deltab^{\left(2\right)}{\left(3\right)}
        //\deltab^{()}${()}$
        return optWhitespace.then(string('^').or(string('$')).or(string('\\').then(regex(/^\bleft\b\((?!\{)/g))).or(string("{\\left("))).then(function(supOrSub) {
            if (supOrSub === "^") {
                return (
                    string("{\\left(").then(function() {
                        return latexMathParser.then(function(block) {
                            return string('\\right)}').map(function(end) {
                                block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                                return self;
                            });
                        });
                    })
                ).or(
                    block.then(function(block) {
                        if (block.ends[L][R].id === block.ends[R].id) {
                            Fragment(0, 0).adopt(blocks[0], blocks[0].ends[R], 0);
                        } else {
                            var firstBlock = block.ends[L][R]; //ignore (
                            var endBlock = block.ends[R][L]; //ignore )          
                            Fragment(firstBlock, endBlock).adopt(blocks[0], blocks[0].ends[R], 0);
                        }
                        return succeed(self);
                    })
                );
            } else if (supOrSub === "$") {
                return block.then(function(block) {
                        if (block.ends[L][R].id === block.ends[R].id) {
                            Fragment(0, 0).adopt(blocks[0], blocks[0].ends[R], 0);
                        } else {
                            var firstBlock = block.ends[L][R]; //ignore (
                            var endBlock = block.ends[R][L]; //ignore )    
                            Fragment(firstBlock, endBlock).adopt(blocks[1], blocks[1].ends[R], 0);
                        }
                        return succeed(self);
                    })
                    .then(string('$'));
            } else if (supOrSub === "left(") {
                 return latexMathParser.then(function(block) {
                       return string('\\right)').map(function(end) {
                           block.children().adopt(blocks[1], blocks[1].ends[R], 0)
                           return succeed(self);
                       });
                  }).or(latexMathParser.then(function(block) {
                       block.children().adopt(blocks[1], blocks[1].ends[R], 0)
                       return succeed(self);
                  }));
            } else if (supOrSub === "{\\left(") {
               return latexMathParser.then(function(block) {
                    return string('\\right)}').map(function(end) {
                        block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                        return self;
                    });
                });
            }
        }).atMost(2).result(self);
    };
});

var ArcFunction = P(MathCommand, function(_, supper_){
    _.init = function(ctrlSeq, htmlCode) {
        this.ctrlSeq = ctrlSeq;
        this.htmlCode = htmlCode;
        var htmlTemplate = 
        '<span class="mq-int mq-non-leaf" style="marginTop: 0.05em">' +
        '<span>'+ htmlCode +'</span>' +
        '<span class="mq-non-leaf">' +
        '<span class="mq-scaled mq-paren">(</span>' +
        '<span class="mq-non-leaf">&0</span>' +
        '<span class="mq-scaled mq-paren">)</span>' +
        '</span>' +
        '</span>';
        Symbol.prototype.init.call(this, ctrlSeq, htmlTemplate);
    };
    _.sLatex = function() {
        return this.ctrlSeq + "{" + this.blocks[0].sLatex() + "}";
    };
    _.sLatexToSVG = function() {
        return this.htmlCode + "\\left("+ this.blocks[0].sLatexToSVG() +"\\right)";
    };
    _.nativeLatex = function() {
        return  this.ctrlSeq + "\\left("+ this.blocks[0].nativeLatex() +"\\right)";
    };
    _.parser = function(){
        var string = Parser.string;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var self = this;
        var blocks = self.blocks = [MathBlock()];
        blocks[0].adopt(self, self.ends[R], 0);
        //\arcsinh\left(x\right)
        //\arcsinh{x}
        return (string("\\left(").then(function(){
                return latexMathParser.then(function(block) {
                    return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                            return self;
                        });
                    });
                })
        ).or(
            block.then(function(block) { 
            var firstBlock = block.ends[L];
            var endBlock = block.ends[R];         
            Fragment(firstBlock, endBlock).adopt(blocks[0], blocks[0].ends[R], 0);
            return succeed(self);
        }));
    };
    _.reflow = function(){
       reflowRoundBracket(this.blocks[0].jQ);
    }
});
LatexCmds['arcsin'] = P(ArcFunction, function(_, supper_){
    _.init = function() {
        ArcFunction.prototype.init.call(this, '\\arcsin', 'sin\uEDFC');
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + "sin\uEDFC("+ this.blocks[0].latex() + ")" + CLOSE_BRACKET_SERVER;
    };
});
LatexCmds['arccos'] = P(ArcFunction, function(_, supper_){
    _.init = function() {
        ArcFunction.prototype.init.call(this, '\\arccos', 'cos\uEDFC');
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + "cos\uEDFC("+ this.blocks[0].latex() + ")" + CLOSE_BRACKET_SERVER;
    };
});
LatexCmds['arctan'] = P(ArcFunction, function(_, supper_){
    _.init = function() {
        ArcFunction.prototype.init.call(this, '\\arctan', 'tan\uEDFC');
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + "tan\uEDFC("+ this.blocks[0].latex() + ")" + CLOSE_BRACKET_SERVER;
    };
});
LatexCmds['arcsinh'] = P(ArcFunction, function(_, supper_){
    _.init = function() {
        ArcFunction.prototype.init.call(this, '\\arcsinh', 'sinh\uEDFC');
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + "sinh\uEDFC("+ this.blocks[0].latex() + ")" + CLOSE_BRACKET_SERVER;
    };
});
LatexCmds['arccosh'] = P(ArcFunction, function(_, supper_){
    _.init = function() {
        ArcFunction.prototype.init.call(this, '\\arccosh', 'cosh\uEDFC');
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + "cosh\uEDFC("+ this.blocks[0].latex() + ")" + CLOSE_BRACKET_SERVER;
    };
});
LatexCmds['arctanh'] = P(ArcFunction, function(_, supper_){
    _.init = function() {
        ArcFunction.prototype.init.call(this, '\\arctanh', 'tanh\uEDFC');
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + "tanh\uEDFC("+ this.blocks[0].latex() + ")" + CLOSE_BRACKET_SERVER;
    };
});
var functionLists = [{ name: "taylor", htmlCode: 'taylor', para: 3, option: 1, hint: ['Exp/List', 'Variable', 'order', 'center point'] },
    { name: "randomb", htmlCode: 'rand', para: 2, option: 0, hint: ['min value', 'max value'], ctrlSeq: 'rand' },
    { name: "randList", htmlCode: 'randList', para: 1, option: 1, hint: ['n', 'a,b'] },
    { name: "randNorm", htmlCode: 'randNorm', para: 2, option: 1, hint: ['\uED78', '\uED72', 'n'] },
    { name: "randBin", htmlCode: 'randBin', para: 2, option: 1, hint: ['n', ' P', 'm'] },
    { name: "fsin", htmlCode: 'sin', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'sin' },
    { name: "fcos", htmlCode: 'cos', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'cos' },
    { name: "ftan", htmlCode: 'tan', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'tan' },
    { name: "fsinh", htmlCode: 'sinh', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'sinh' },
    { name: "fcosh", htmlCode: 'cosh', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'cosh' },
    { name: "ftanh", htmlCode: 'tanh', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'tanh' },
    { name: "finvsin", htmlCode: 'sin\uEDFC', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'sin\uEDFC' },
    { name: "finvcos", htmlCode: 'cos\uEDFC', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'cos\uEDFC' },
    { name: "finvtan", htmlCode: 'tan\uEDFC', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'tan\uEDFC' },
    { name: "finvsinh", htmlCode: 'sinh\uEDFC', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'sinh\uEDFC' },
    { name: "finvcosh", htmlCode: 'cosh\uEDFC', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'cosh\uEDFC' },
    { name: "finvtanh", htmlCode: 'tanh\uEDFC', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'tanh\uEDFC' },
    { name: "signum", htmlCode: 'signum', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "fln", htmlCode: 'ln', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'ln' },
    { name: "approx", htmlCode: 'approx', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "expanda", htmlCode: 'expand', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'expand' },
    { name: "expandb", htmlCode: 'expand', para: 2, option: 0, hint: ['Exp', 'Variable'], ctrlSeq: 'expand' },
    { name: "combine", htmlCode: 'combine', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "collect", htmlCode: 'collect', para: 1, option: 1, hint: ['Exp/Eq/List/Mat/Variable', 'Exp'] },
    { name: "simplify", htmlCode: 'simplify', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "numerator", htmlCode: 'numerator', para: 1, option: 0, hint: ['Exp/List'] },
    { name: "denominator", htmlCode: 'denominator', para: 1, option: 0, hint: ['Exp/List'] },
    { name: "factor", htmlCode: 'factor', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "rFactor", htmlCode: 'rFactor', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "gcd", htmlCode: 'gcd', para: 2, option: 0, hint: ['Exp/List-1', 'Exp/List-2'] },
    { name: "lcm", htmlCode: 'lcm', para: 2, option: 0, hint: ['Exp/List-1', 'Exp/List-2'] },
    { name: "invert", htmlCode: 'invert', para: 1, option: 1, hint: ['Exp/Eq/Ineq/List', 'Variable-1, Variable-2'] },
    { name: "solvea", htmlCode: 'solve', para: 1, option: 1, hint: ['Exp/Eq/Ineq', 'Variable'], ctrlSeq: 'solve' },
    { name: "solveb", htmlCode: 'solve', para: 2, option: 1, hint: ['Exp/Eq/Ineq\uEE11', 'Variable', 'value, lower limit, upper limit'], ctrlSeq: 'solve' },
    { name: "dsolvea", htmlCode: 'dSolve', para: 3, option: 3, hint: ['Eq', 'independent variable', 'dependent variable', 'initial condition-1, initial condition-2', 'initial condition-3, initial condition-4', 'initial condition-5, initial condition-6'], ctrlSeq: 'dsolve' },
    { name: "dsolveb", htmlCode: 'dSolve', para: 3, option: 1, hint: ['{Eq-1, Eq-2}', 'independent variable', '{dependent variable-1, dependent variable-2}', 'initial condition-1, initial condition-2, initial condition-3, initial condition-4'], ctrlSeq: 'dsolve' },
    { name: "arcLen", htmlCode: 'arcLen', para: 4, option: 0, hint: ['Exp/List', 'Variable', 'start value', 'end value'] },
    { name: "tanLine", htmlCode: 'tanLine', para: 3, option: 0, hint: ['Exp/List', 'Variable', 'variable value at point of tangency'] },
    { name: "normal", htmlCode: 'normal', para: 3, option: 0, hint: ['Exp/List', 'Variable', 'variable value at point of normal'] },
    { name: "seq", htmlCode: 'seq', para: 4, option: 1, hint: ['Exp', 'Variable', 'start value', 'end value', 'step size'] },
    { name: "sequence", htmlCode: 'sequence', para: 1, option: 2, hint: ['List-1', 'List-2', 'Variable'] },
    { name: "sumseq", htmlCode: 'sumSeq', para: 1, option: 2, hint: ['List-1', 'List-2', 'Variable'] },
    { name: "tCollect", htmlCode: 'tCollect', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List/Mat'] },
    { name: "tExpand", htmlCode: 'tExpand', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List/Mat'] },
    { name: "expToTrig", htmlCode: 'expToTrig', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List/Mat'] },
    { name: "trigToExp", htmlCode: 'trigToExp', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List/Mat'] },
    { name: "det", htmlCode: 'det', para: 1, option: 0, hint: ['Mat'] },
    { name: "trn", htmlCode: 'trn', para: 1, option: 0, hint: ['Mat'] },
    { name: "dima", htmlCode: 'dim', para: 1, option: 0, hint: ['Mat'], ctrlSeq: 'dim' },
    { name: "dimb", htmlCode: 'dim', para: 1, option: 0, hint: ['List'], ctrlSeq: 'dim' },
    { name: "augmenta", htmlCode: 'augment', para: 2, option: 0, hint: ['List-1', 'List-2'], ctrlSeq: 'augment' },
    { name: "augmentb", htmlCode: 'augment', para: 2, option: 0, hint: ['Mat-1', 'Mat-2'], ctrlSeq: 'augment' },
    { name: "ident", htmlCode: 'ident', para: 1, option: 0, hint: ['natural number'], ctrlSeq: 'ident' },
    { name: "filla", htmlCode: 'fill', para: 2, option: 0, hint: ['Exp/Eq/Ineq', 'number of elements'], ctrlSeq: 'fill' },
    { name: "fillb", htmlCode: 'fill', para: 2, option: 0, hint: ['Exp/Eq/Ineq', 'List'], ctrlSeq: 'fill' },
    { name: "fillc", htmlCode: 'fill', para: 2, option: 0, hint: ['List', 'List'], ctrlSeq: 'fill' },
    { name: "filld", htmlCode: 'fill', para: 3, option: 0, hint: ['Exp', 'number of rows', 'number of columns'], ctrlSeq: 'fill' },
    { name: "fille", htmlCode: 'fill', para: 2, option: 0, hint: ['Exp', 'Mat'], ctrlSeq: 'fill' },
    { name: "fillf", htmlCode: 'fill', para: 2, option: 0, hint: ['Exp', 'Mat'], ctrlSeq: 'fill' },
    { name: "fillg", htmlCode: 'fill', para: 3, option: 0, hint: ['Exp', '1', 'number of columns'], ctrlSeq: 'fill' },
    { name: "listtomat", htmlCode: 'listToMat', para: 1, option: 1, hint: ['List-1', 'List-2, ..., List-N'], ctrlSeq: 'listToMat' },
    { name: "mattolist", htmlCode: 'matToList', para: 2, option: 0, hint: ['Mat', 'column number'], ctrlSeq: 'matToList' },
    { name: "crossP", htmlCode: 'crossP', para: 2, option: 0, hint: ['Mat-1', 'Mat-2'] },
    { name: "dotP", htmlCode: 'dotP', para: 2, option: 0, hint: ['Mat-1', 'Mat-2'] },
    { name: "colDim", htmlCode: 'colDim', para: 1, option: 0, hint: ['Mat'] },
    { name: "rowDim", htmlCode: 'rowDim', para: 1, option: 0, hint: ['Mat'] },
    { name: "rowswap", htmlCode: 'swap', para: 3, option: 0, hint: ['Mat', 'row number-1', 'row number-2'], ctrlSeq: 'swap' },
    { name: "rotate", htmlCode: 'rotate', para: 1, option: 1, hint: ['List', 'number of rotations'] },
    { name: "shift", htmlCode: 'shift', para: 1, option: 1, hint: ['List', 'number of shifts'] },
    { name: "polyEval", htmlCode: 'polyEval', para: 1, option: 1, hint: ['List', 'Exp/List'] },
    { name: "diag", htmlCode: 'diag', para: 1, option: 0, hint: ['Mat'] },
    { name: "mRow", htmlCode: 'mRow', para: 3, option: 0, hint: ['Exp', 'Mat', 'row number'] },
    { name: "mRowAdd", htmlCode: 'mRowAdd', para: 4, option: 0, hint: ['Exp', 'Mat', 'row number-1', 'row number-2'] },
    { name: "rowAdd", htmlCode: 'rowAdd', para: 3, option: 0, hint: ['Mat', 'row number-1', 'row number-2'] },
    { name: "ref", htmlCode: 'ref', para: 1, option: 0, hint: ['Mat'] },
    { name: "rref", htmlCode: 'rref', para: 1, option: 0, hint: ['Mat'] },
    { name: "stdDev", htmlCode: 'stdDev', para: 1, option: 0, hint: ['List'] },
    { name: "variance", htmlCode: 'variance', para: 1, option: 0, hint: ['List'] },
    { name: "submat", htmlCode: 'subMat', para: 1, option: 4, hint: ['Mat', 'start row', 'start column', 'end row', 'end column'] },
    { name: "norm", htmlCode: 'norm', para: 1, option: 0, hint: ['Mat'] },
    { name: "colNorm", htmlCode: 'colNorm', para: 1, option: 0, hint: ['Mat'] },
    { name: "rowNorm", htmlCode: 'rowNorm', para: 1, option: 0, hint: ['Mat'] },
    { name: "unitV", htmlCode: 'unitV', para: 1, option: 0, hint: ['Mat'] },
    { name: "eigVl", htmlCode: 'eigVl', para: 1, option: 0, hint: ['Mat'] },
    { name: "eigVc", htmlCode: 'eigVc', para: 1, option: 0, hint: ['Mat'] },
    { name: "lu", htmlCode: 'LU', para: 3, option: 0, hint: ['Mat', 'lVariableMem', 'uVariableMem'] },
    { name: "qr", htmlCode: 'QR', para: 3, option: 0, hint: ['Mat', 'qVariableMem', 'rVariableMem'] },
    { name: "fangle", htmlCode: 'angle', para: 2, option: 0, hint: ['Mat-1', 'Mat-2'], ctrlSeq: 'angle' },
    { name: "cuml", htmlCode: 'cuml', para: 1, option: 0, hint: ['List'] },
    { name: "max", htmlCode: 'max', para: 1, option: 1, hint: ['Exp/List-1', 'Exp/List-2'] },
    { name: "min", htmlCode: 'min', para: 1, option: 1, hint: ['Exp/List-1', 'Exp/List-2'] },
    { name: "mean", htmlCode: 'mean', para: 1, option: 1, hint: ['List-1', 'List-2'] },
    { name: "median", htmlCode: 'median', para: 1, option: 1, hint: ['List-1', 'List-2'] },
    { name: "percent", htmlCode: 'percent', para: 1, option: 0, hint: ['List'] },
    { name: "sortA", htmlCode: 'sortA', para: 1, option: 0, hint: ['List'] },
    { name: "sortD", htmlCode: 'sortD', para: 1, option: 0, hint: ['List'] },
    { name: "sublist", htmlCode: 'subList', para: 1, option: 2, hint: ['List', 'start number', 'end number'] },
    { name: "summ", htmlCode: 'sum', para: 1, option: 1, hint: ['List-1', 'List-2'] },
    { name: "prodm", htmlCode: 'prod', para: 1, option: 1, hint: ['List-1', 'List-2'] },
    { name: "fnpr", htmlCode: 'nPr', para: 2, option: 0, hint: ['n', 'r'], ctrlSeq: 'npr' },
    { name: "fncr", htmlCode: 'nCr', para: 2, option: 0, hint: ['n', 'r'], ctrlSeq: 'ncr' },
    { name: "mod", htmlCode: 'mod', para: 2, option: 0, hint: ['{Exp/List}-1', '{Exp/List}-2'] },
    { name: "fmina", htmlCode: 'fMin', para: 1, option: 1, hint: ['Exp', 'Variable'], ctrlSeq: 'fmin' },
    { name: "fminb", htmlCode: 'fMin', para: 4, option: 1, hint: ['Exp', 'Variable', 'start value', 'end value', 'n'], ctrlSeq: 'fmin' },
    { name: "fmaxa", htmlCode: 'fMax', para: 1, option: 1, hint: ['Exp', 'Variable'], ctrlSeq: 'fmax' },
    { name: "fmaxb", htmlCode: 'fMax', para: 4, option: 1, hint: ['Exp', 'Variable', 'start value', 'end value', 'n'], ctrlSeq: 'fmax' },
    { name: "fRound", htmlCode: 'fRound', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "sRound", htmlCode: 'sRound', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "strtoexp", htmlCode: 'strToExp', para: 1, option: 0, hint: ['&quot;&lt;string&gt;&quot;'], ctrlSeq: 'strToExp' },
    { name: "range", htmlCode: 'rangeAppoint', para: 3, option: 0, hint: ['Exp/Eq/List', 'start value', 'end value'], ctrlSeq: 'rangeAppoint' },
    { name: "exact", htmlCode: 'toFrac', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'toFrac' },
    { name: "ffrac", htmlCode: 'frac', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'frac' },
    { name: "eliminate", htmlCode: 'eliminate', para: 3, option: 0, hint: ['Eq/Ineq/List-1', 'Variable', 'Eq-2'] },
    { name: "arrange", htmlCode: 'arrange', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List/Mat'] },
    { name: "exchange", htmlCode: 'exchange', para: 1, option: 0, hint: ['Eq/Ineq/List'] },
    { name: "replace", htmlCode: 'replace', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List/Mat'] },
    { name: "rewrite", htmlCode: 'rewrite', para: 1, option: 0, hint: ['Eq/Ineq/List'] },
    { name: "getRight", htmlCode: 'getRight', para: 1, option: 0, hint: ['Eq/Ineq/List'] },
    { name: "getLeft", htmlCode: 'getLeft', para: 1, option: 0, hint: ['Eq/Ineq/List'] },
    { name: "absExpand", htmlCode: 'absExpand', para: 1, option: 0, hint: ['Eq/Ineq'] },
    { name: "andConnect", htmlCode: 'andConnect', para: 2, option: 0, hint: ['Eq/Ineq-1', 'Eq/Ineq-2'] },
    { name: "cExpand", htmlCode: 'cExpand', para: 1, option: 0, hint: ['Exp/Eq/List/Mat'] },
    { name: "compToPol", htmlCode: 'compToPol', para: 1, option: 0, hint: ['Exp/Eq/List/Mat'], ctrlSeq: 'compToPol' },
    { name: "comptotrig", htmlCode: 'compToTrig', para: 1, option: 0, hint: ['Exp/Eq/List/Mat'] },
    { name: "qone", htmlCode: 'Q\uECF1', para: 1, option: 1, hint: ['List-1', 'List-2'], ctrlSeq: 'Q1' },
    { name: "qthree", htmlCode: 'Q\uECF3', para: 1, option: 1, hint: ['List-1', 'List-2'], ctrlSeq: 'Q3' },
    { name: "mode", htmlCode: 'mode', para: 1, option: 1, hint: ['List-1', 'List-2'] },
    { name: "arg", htmlCode: 'arg', para: 1, option: 0, hint: ['Exp/Eq/List/Mat'] },
    { name: "conjg", htmlCode: 'conjg', para: 1, option: 0, hint: ['Exp/Eq/Ineq\uEE11/List/Mat'] },
    { name: "re", htmlCode: 're', para: 1, option: 0, hint: ['Exp/Eq/Ineq\uEE11/List/Mat'] },
    { name: "im", htmlCode: 'im', para: 1, option: 0, hint: ['Exp/Eq/Ineq\uEE11/List/Mat'] },
    { name: "fint", htmlCode: 'int', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'], ctrlSeq: 'int' },
    { name: "intg", htmlCode: 'intg', para: 1, option: 0, hint: ['Exp/Eq/List/Mat/Variable'] },
    { name: "dms", htmlCode: 'dms', para: 1, option: 2, hint: ['Exp/List-1', 'Exp/List-2', 'Exp/List-3'] },
    { name: "todms", htmlCode: 'toDMS', para: 1, option: 0, hint: ['Exp/List'] },
    { name: "propfrac", htmlCode: 'propFrac', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List/Mat'] },
    { name: "toCyl", htmlCode: 'toCyl', para: 1, option: 1, hint: ['Mat', 'natural number'] },
    { name: "toPol", htmlCode: 'toPol', para: 1, option: 1, hint: ['Mat', 'natural number'] },
    { name: "torect", htmlCode: 'toRect', para: 1, option: 1, hint: ['Mat', 'natural number'] },
    { name: "toSph", htmlCode: 'toSph', para: 1, option: 1, hint: ['Mat', 'natural number'] },
    { name: "factorOut", htmlCode: 'factorOut', para: 2, option: 0, hint: ['Exp/Eq/Ineq/List/Mat', 'Exp'] },
    { name: "judge", htmlCode: 'judge', para: 1, option: 0, hint: ['Exp/Eq/Ineq'] },
    { name: "not", htmlCode: 'not', para: 1, option: 0, hint: ['Exp/Eq/Ineq/List'] },
    { name: "percentile", htmlCode: 'percentile', para: 2, option: 0, hint: ['List', 'number'] },
    { name: "ffta", htmlCode: 'FFT', para: 1, option: 0, hint: ['List'], ctrlSeq: 'fft' },
    { name: "fftb", htmlCode: 'FFT', para: 2, option: 0, hint: ['List', 'm'], ctrlSeq: 'fft' },
    { name: "iffta", htmlCode: 'IFFT', para: 1, option: 0, hint: ['List'], ctrlSeq: 'ifft' },
    { name: "ifftb", htmlCode: 'IFFT', para: 2, option: 0, hint: ['List', 'm'], ctrlSeq: 'ifft' },
    { name: "normPDf", htmlCode: 'normPDf', para: 1, option: 1, hint: ['x', '\uED78 , \uED72'] },
    { name: "normCDf", htmlCode: 'normCDf', para: 2, option: 1, hint: ['lower value', 'upper value', '\uED78 , \uED72'] },
    { name: "tPDf", htmlCode: 'tPDf', para: 2, option: 0, hint: ['x', 'df'] },
    { name: "tCDf", htmlCode: 'tCDf', para: 3, option: 0, hint: ['lower value', 'upper value', 'df'] },
    { name: "invTCDF", htmlCode: 'invTCDf', para: 2, option: 0, hint: ['prob', 'df'], ctrlSeq: 'invTCDf' },
    { name: "chiPDf", htmlCode: 'chiPDf', para: 2, option: 0, hint: ['x', 'df'] },
    { name: "chiCDf", htmlCode: 'chiCDf', para: 3, option: 0, hint: ['lower value', 'upper value', 'df'] },
    { name: "invChiCDf", htmlCode: 'invChiCDf', para: 2, option: 0, hint: ['prob', 'df'] },
    { name: "fPDf", htmlCode: 'fPDf', para: 3, option: 0, hint: ['x', 'n : df', 'd : df'] },
    { name: "fCDf", htmlCode: 'fCDf', para: 4, option: 0, hint: ['lower value', 'upper value', 'n : df', 'd : df'] },
    { name: "invfCDf", htmlCode: 'invFCDf', para: 3, option: 0, hint: ['prob', 'n : df', 'd : df'], ctrlSeq: 'invFCDf'},
    { name: "binomialPDf", htmlCode: 'binomialPDf', para: 3, option: 0, hint: ['x', 'numtrial value', 'pos'] },
    { name: "binomialCDf", htmlCode: 'binomialCDf', para: 4, option: 0, hint: ['lower value', 'upper value', 'numtrial value', 'pos'] },
    { name: "invBinomialCDf", htmlCode: 'invBinomialCDf', para: 3, option: 0, hint: ['prob', 'numtrial value', 'pos'] },
    { name: "poissonPDf", htmlCode: 'poissonPDf', para: 2, option: 0, hint: ['x', '\uED71'] },
    { name: "poissonCDf", htmlCode: 'poissonCDf', para: 3, option: 0, hint: ['lower value', 'upper value', '\uED71'] },
    { name: "invPoissonCDf", htmlCode: 'invPoissonCDf', para: 2, option: 0, hint: ['prob', '\uED71'] },
    { name: "geoPDf", htmlCode: 'geoPDf', para: 2, option: 0, hint: ['x', 'pos'] },
    { name: "geoCDf", htmlCode: 'geoCDf', para: 3, option: 0, hint: ['lower value', 'upper value', 'pos'] },
    { name: "invGeoCDf", htmlCode: 'invGeoCDf', para: 2, option: 0, hint: ['prob', 'pos'] },
    { name: "hypergeoPDf", htmlCode: 'hypergeoPDf', para: 4, option: 0, hint: ['x', 'n', 'M', 'N'] },
    { name: "hypergeoCDf", htmlCode: 'hypergeoCDf', para: 5, option: 0, hint: ['lower value', 'upper value', 'n', 'M', 'N'] },
    { name: "invHypergeoCDf", htmlCode: 'invHypergeoCDf', para: 4, option: 0, hint: ['prob', 'n', 'M', 'N'] },
    { name: "iMod", htmlCode: 'iMod', para: 2, option: 0, hint: ['Exp-1/List-1', 'Exp-2/List-2'] },
    { name: "igcda", htmlCode: 'iGcd', para: 2, option: 1, hint: ['Exp-1', 'Exp-2', 'Exp-3\uEE90Exp-10'], ctrlSeq: 'iGcd' },
    { name: "igcdb", htmlCode: 'iGcd', para: 2, option: 1, hint: ['List-1', 'List-2', 'List-3\uEE90List-10'], ctrlSeq: 'iGcd' },
    { name: "ilcma", htmlCode: 'iLcm', para: 2, option: 1, hint: ['Exp-1', 'Exp-2', 'Exp-3\uEE90Exp-10'], ctrlSeq: 'iLcm' },
    { name: "ilcmb", htmlCode: 'iLcm', para: 2, option: 1, hint: ['List-1', 'List-2', 'List-3\uEE90List-10'], ctrlSeq: 'iLcm' },
    { name: "isprime", htmlCode: 'isPrime', para: 1, option: 0, hint: ['Exp/List'], ctrlSeq: 'isPrime' },
    { name: "impDiff", htmlCode: 'impDiff', para: 3, option: 0, hint: ['Eq/Exp/List', 'independent variable', 'dependent variable'] },
    { name: "rank", htmlCode: 'rank', para: 1, option: 0, hint: ['Mat'] },
    { name: "const", htmlCode: 'const', para: 1, option: 0, hint: ['natural num'] },
    { name: "constn", htmlCode: 'constn', para: 1, option: 0, hint: ['index'] },
    { name: "dlist", htmlCode: '\uEE78list', para: 1, option: 0, hint: ['List'] },
    { name: "toRect", htmlCode: '\uEE0B', para: 1, option: 0, hint: ['Radian'] }
    //delay
    // {name:"fsolve",htmlCode:'solve',para:1, option:1, ctrlSeq: 'solve'}
];

for (var j = 0; j < functionLists.length; j++) {
    var name = functionLists[j].name;
    var ctrlSeq = null;
    if (functionLists[j].ctrlSeq == null) {
        ctrlSeq = name;
    } else {
        ctrlSeq = functionLists[j].ctrlSeq;
    }
    LatexCmds[name] = bind(AllFunction, name, functionLists[j].htmlCode, ctrlSeq, functionLists[j].para, functionLists[j].option, functionLists[j].hint);
}

var Comptorect = P(AllFunction, function(_, supper_) {
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        return optWhitespace.then(string('(')).then(function(supOrSub) {
            return block.times(self.numBlocks()).map(function(blocks) {
                self.blocks = blocks;

                for (var i = 0; i < blocks.length; i += 1) {
                    blocks[i].adopt(self, self.ends[R], 0);
                }

                return self;
            }).skip(string(')'));
        }).many().result(self);
    };
});
LatexCmds['invNormCDf'] = LatexCmds.invNormCDf = P(Comptorect, function(_, supper_) {
    _.init = function() {
        var htmlTemplate =
            '<var>invNormCDf</var>' +
            '<span class="mq-int mq-non-leaf">' +
            '<span class="mq-scaled mq-paren">(</span>' +
            '<span class="mq-non-leaf">' +
            '<span class="ph_func active" data-placeholder="[tail setting,]">&0</span>' +
            '<var> </var>' +
            '<span class="mq-scaled  ph_func active" data-placeholder="area value">&1</span>' +
            '<var> </var>' +
            // nnanh 12/13/2018 - Edit - S -  Bug#9968 localfont
            '<span class="ph_func active" data-placeholder="[,\uED72, \uED78]">&2</span>' +
            // nnanh 12/13/2018 - Edit - E -  Bug#9968 localfont
            '</span>' +
            '<span class="mq-scaled mq-paren">)</span>' +
            '</span>';
        Symbol.prototype.init.call(this, 'invNormCDf', htmlTemplate);
    };
    _.latex = function() {
        return 'invNormCDf' + '(' + this.blocks[0].latex() +
            this.blocks[1].latex() + this.blocks[2].latex() + ')';
    };
    _.sLatex = function() {
        return '\\invNormCDf' + '({' + this.blocks[0].sLatex() + "}{" +
            this.blocks[1].sLatex() + "}{" + this.blocks[2].sLatex() + '})';
    };
    _.nativeLatex = function() {
        return '\\invNormCDf' + '({' + this.blocks[0].nativeLatex() + "}{" +
            this.blocks[1].nativeLatex() + "}{" + this.blocks[2].nativeLatex() + '})';
    };
    _.sLatexToSVG = function() {
        return 'invNormCDf' + '\\left(' + this.blocks[0].sLatexToSVG() + this.blocks[1].sLatexToSVG() + this.blocks[2].sLatexToSVG() + '\\right)';
    };
    _.reflow = function() {
        supper_.reflow.apply(this);
    };
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\invNormCDf({1}{2}{3})
         //\invNormCDf\left(\right)
        return (optWhitespace.then(string('(')).then(function(supOrSub) {
            return block.times(self.numBlocks()).map(function(blocks) {
                self.blocks = blocks;

                for (var i = 0; i < blocks.length; i += 1) {
                    blocks[i].adopt(self, self.ends[R], 0);
                }

                return self;
            }).skip(string(')'));
        })).or(
            optWhitespace.then(string('\\left(')).then(function(barrierSymbol) {
                 return latexMathParser.then(function(block) {
                      return string('\\right)').map(function(end) {
                           block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                           return self;
                      });
                 });
            })
        ).many().result(self);
    };
});
LatexCmds['comptorecta'] = LatexCmds.comptorecta = P(Comptorect, function(_, supper_) {
    _.init = function() {
        var htmlTemplate =
            '<span class="mq-int mq-non-leaf">' +
            '<var>compToRect</var>' +
            '<span class="mq-scaled mq-paren">(</span>' +
            '<span class="mq-non-leaf">' +
                '<var>' + '\uEE0B' + '</var>' +
                '<span class="mq-scaled mq-paren">(</span>' +
                '<span class="mq-non-leaf">' +
                    '<span class="mq-scaled  ph_func active" data-placeholder="r">&0</span>' +
                    '<var>, </var>' +
                    '<span class="mq-scaled  ph_func active" data-placeholder="\uED6E">&1</span>' +                   
                '</span>' +
                '<span class="mq-scaled mq-paren">)</span>' +
            '</span>' +
            '<span class="mq-scaled mq-paren">)</span>' +
            '</span>';
        Symbol.prototype.init.call(this, 'comptorect', htmlTemplate);
    };
    _.latex = function() {
        return 'comptorect' + '(' + this.blocks[0].latex() + '*\\xee\\x02^((' +
            this.blocks[1].latex() + ')*\\xee\\x01))';
    };
    _.sLatex = function() {
        return '\\comptorecta({' + this.blocks[0].sLatex() + '}{' + this.blocks[1].sLatex() + '})';
    };
    _.nativeLatex = function() {
        return '\\comptorecta({' + this.blocks[0].nativeLatex() + '}{' + this.blocks[1].nativeLatex() + '})';
    };
    _.sLatexToSVG = function() {
        return 'compToRect' + '\\left(\uEE0B\\left(' + this.blocks[0].sLatexToSVG() + ',' + this.blocks[1].sLatexToSVG() + '\\right)\\right)';
    };
    _.reflow = function() {
        reflowRoundBracket(this.jQ.children(":eq(2)").children(":eq(2)"));   
        reflowRoundBracket(this.jQ.children(":eq(2)"));   
    };  
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        var firstBlock = [];
        var secondBlock = [];
        var chooseBlock = firstBlock;
        //\comptorecta\left(\angle \left({4,55}\right)\right)
        // \comptorecta({1}{2})
        return (optWhitespace.then(string('(')).then(function(supOrSub) {
            return block.times(self.numBlocks()).map(function(blocks) {
                self.blocks = blocks;

                for (var i = 0; i < blocks.length; i += 1) {
                    blocks[i].adopt(self, self.ends[R], 0);
                }

                return self;
            }).skip(string(')'));
        })).or(
            optWhitespace.then(string('\\left(')).then(function(barrierSymbol) {    
                return block.then(function(block) {                   
                    return (string('\\right)').map(function(end) {
                        var dataBlock = block.ends[L].blocks[0];
                        var startBlock = dataBlock.ends[L];
                        var endBlock = startBlock[R];
                        var leftEndBlock = endBlock[L];
                        while (endBlock !== 0 && endBlock.ctrlSeq !== ",") {
                            leftEndBlock = endBlock;
                            endBlock = endBlock[R];
                        }
                        if (endBlock !== 0 && endBlock.ctrlSeq === ",") {
                            Fragment(startBlock, endBlock[L]).adopt(blocks[0],blocks[0].ends[R], 0);
                            if (endBlock[R] !== 0) {
                                Fragment(endBlock[R], dataBlock.ends[R]).adopt(blocks[1],blocks[1].ends[R], 0);
                            }                            
                        } else if (leftEndBlock) {
                            Fragment(startBlock, leftEndBlock).adopt(blocks[0],blocks[0].ends[R], 0);
                        } else {
                            Fragment(startBlock, startBlock).adopt(blocks[0],blocks[0].ends[R], 0);
                        }
                        return self;
                    })).or(succeed(self))
                }).many().result(self)
            })
        );
    };    
});
LatexCmds['comptorectb'] = LatexCmds.comptorectb = P(Comptorect, function(_, supper_) {
    _.init = function() {
        var htmlTemplate =
            '<span class="mq-int mq-non-leaf">' +
            '<var>compToRect</var>' +
            '<span class="mq-scaled mq-paren">(</span>' +
            '<span class="mq-non-leaf">' +
            '<span class="mq-scaled  ph_func active" data-placeholder="r">&0</span>' +
            '<var>*\uEDA5</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sup">' +
            '<span class="mq-scaled  ph_func active" data-placeholder="\uED6E">&1</span>' +
            '<var>*\uEE01</var>' +
            '</span>' +
            '<span style="display:inline-block;width:0">&#8203</span>' +
            '</span>' +
            '</span>' +
            '<span class="mq-scaled mq-paren">)</span>' +
            '</span>';
        Symbol.prototype.init.call(this, 'comptorect', htmlTemplate);
    };
    _.latex = function() {
        return 'comptorect' + '(' + this.blocks[0].latex() + '*\\xee\\x02^((' +
            this.blocks[1].latex() + ')*\\xee\\x01))';
    };
    _.sLatex = function() {
        return '\\comptorectb({' + this.blocks[0].sLatex() + '}{' + this.blocks[1].sLatex() + '})';
    };
    _.nativeLatex = function() {        
       return '\\comptorectb({' + this.blocks[0].nativeLatex() + '}{' + this.blocks[1].nativeLatex() + '})';
    };
    _.sLatexToSVG = function() {
        return 'compToRect' + '\\left(' + this.blocks[0].sLatexToSVG() + '*e^{' + this.blocks[1].sLatexToSVG() + '*i}\\right)';
    };
    _.reflow = function() {
        reflowRoundBracket(this.jQ.children(":eq(2)"));  
    };
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        var firstBlock = [];
        var secondBlock = [];
        var chooseBlock = firstBlock;
        //\comptorectb\left(4*e^{5*i}\right)
        //\comptorectb({1}{2})
        return (optWhitespace.then(string('(')).then(function(supOrSub) {
            return block.times(self.numBlocks()).map(function(blocks) {
                self.blocks = blocks;

                for (var i = 0; i < blocks.length; i += 1) {
                    blocks[i].adopt(self, self.ends[R], 0);
                }

                return self;
            }).skip(string(')'));
        })).or(
            (string('\\left(').or(string('\\right)'))).then(function(barrierSymbol) {               
                if (barrierSymbol === "\\right)") {
                    for (var i = 0; i < firstBlock.length; i++) {
                        firstBlock[i].children().adopt(blocks[0],blocks[0].ends[R], 0);
                    }
                    if (secondBlock.length > 0) {
                        var dataBlock = secondBlock[0].ends[L].blocks[0];
                        var startBlock = dataBlock.ends[L];
                        var endBlock = startBlock[R];
                        var leftEndBlock = endBlock[L];
                        while (endBlock !== 0 && endBlock.ctrlSeq !== "*") {
                            leftEndBlock = endBlock;
                            endBlock = endBlock[R];
                        }
                        if (endBlock !== 0 && endBlock.ctrlSeq === "*") {
                            Fragment(startBlock, endBlock[L]).adopt(blocks[1],blocks[1].ends[R], 0);
                        } else if (leftEndBlock) {
                            Fragment(startBlock, leftEndBlock).adopt(blocks[1],blocks[1].ends[R], 0);
                        } else {
                            Fragment(startBlock, startBlock).adopt(blocks[1],blocks[1].ends[R], 0);
                        }
                    }                 
                    return succeed(self);
                } else {                   
                    return block.then(function(block) {
                        chooseBlock.push(block);
                        return (string('*\\e').map(function(end) {
                            chooseBlock = secondBlock;
                            return self;
                        }))                                           
                        .or(succeed(self))
                    }).many().result(self)
                }               
            }).atMost(2).result(self)
        );
    };   
});
// ThuyLTM write fourier and laplace funtion

LatexCmds['diffa'] =
    LatexCmds.diffa = P(MathCommand, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-fraction mq-non-leaf">' +
                '<span class="mq-numerator">' +
                '<var>d</var>' +
                '</span>' +
                '<span class="mq-denominator" >' +
                '<var>d</var>' +
                '<var>&0</var>' +
                '</span>' +
                '<span style="display:inline-block;width:0">​&#8203</span>' +
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&1</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>';
            Symbol.prototype.init.call(this, 'diff', htmlTemplate);
        };
        _.latex = function() {
            if (this.blocks[0].latex() == "") {
                return OPEN_BRACKET_SERVER + this.ctrlSeq + '(' + this.blocks[1].latex() + ')' + CLOSE_BRACKET_SERVER;
            }
            return OPEN_BRACKET_SERVER + this.ctrlSeq + '(' + this.blocks[1].latex() + "," +
                this.blocks[0].latex() + ')' + CLOSE_BRACKET_SERVER;
        };
        _.sLatex = function() {
            return '\\diffa' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}';
        };
        _.nativeLatex = function() {
            return '\\diffa' + '{' + this.blocks[0].nativeLatex() + "}{" +
                this.blocks[1].nativeLatex() + '}';
        };
        _.sLatexToSVG = function() {
            return "\\frac{d}{d" + this.blocks[0].sLatexToSVG() + "}{(" + this.blocks[1].sLatexToSVG() + ")}";
        };
        _.reflow = function() {
            reflowRoundBracket(this.blocks[1].jQ); 
        };
        _.parser = function() {
            var string = Parser.string;
            var optWhitespace = Parser.optWhitespace;
            var succeed = Parser.succeed;
            var block = latexMathParser.block;
            var regex = Parser.regex;

            var self = this;
            var blocks = self.blocks = [MathBlock(), MathBlock()];
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            //\diffa{x}\left( x\right) HandWriting
            // \diffa{x}{x}
            //NNANH 2/28/2019 - S - Edit - Bug#10472 fix parser diffa diffb
            var check = 0;
            return (
                string("\\left(").then(function() {
                        return latexMathParser.then(function(block) {
                            return string('\\right)').map(function(end) {
                                block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                                return succeed(self);
                            });
                        });
                })
            ).or(                
                 block.then(function(block) {
                    //NNANH 2/28/2019 - S - Edit - Bug#10472 fix parser diffa diffb
                    // if (blocks[0].isEmpty()) {
                    //   block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                    // } else {
                    //   block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                    // }
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                    //NNANH 2/28/2019 - E - Edit - Bug#10472 fix parser diffa diffb
                    return succeed(self);
               })
            )
            .atMost(2).result(self);
        }
    });

LatexCmds['diffb'] =
    LatexCmds.diffb = P(MathCommand, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-fraction mq-non-leaf">' +
                '<span class="mq-numerator">' +
                '<var>d</var>' +
                '<span class="mq-supsub mq-non-leaf">' +
                '<span class="mq-sup"><span class="mq-mathcall-sup-inner">&0</span></span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '</span>' +
                '<span class="mq-denominator" >' +
                '<var>d</var>' +
                '<span>&1</span>' +
                '<span class="mq-supsub mq-non-leaf">' +
                '<span class="mq-sup">' +
                '<span class="mq-non-leaf">&2</span>' +
                '</span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '</span>' +
                '<span style="display:inline-block;width:0">​&#8203</span>' +
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&3</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>';
            Symbol.prototype.init.call(this, '\\diffb', htmlTemplate);
        };
        _.latex = function() {
            if (this.blocks[2].latex() == "") {
                return OPEN_BRACKET_SERVER + 'diff' + '(' + this.blocks[3].latex() + "," +
                    this.blocks[1].latex() + ')' + CLOSE_BRACKET_SERVER;
            }
            return OPEN_BRACKET_SERVER + 'diff' + '(' + this.blocks[3].latex() + "," +
                this.blocks[1].latex() + "," + this.blocks[2].latex() + ')' + CLOSE_BRACKET_SERVER;
        };
        _.sLatex = function() {
            return '\\diffb' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + "}{" + this.blocks[2].sLatex() + '}{' + this.blocks[3].sLatex() + '}';
        };
        _.nativeLatex = function() {
            return '\\diffb' + '{' + this.blocks[0].nativeLatex() + "}{" +
                this.blocks[1].nativeLatex() + "}{" + this.blocks[2].nativeLatex() + '}{' + this.blocks[3].nativeLatex() + '}';
        };
        _.sLatexToSVG = function() {
            return "\\frac{d^{" + this.blocks[0].sLatexToSVG() + "}}{d{" + this.blocks[1].sLatexToSVG() + "}^{" +
                this.blocks[2].sLatexToSVG() + "}}{(" + this.blocks[3].sLatexToSVG() + ")}";
        };
        _.reflow = function() {
            reflowRoundBracket(this.blocks[3].jQ);   
           
            var block0 = this.blocks[0];
            var block2 = this.blocks[2];
            var cursor = this.parent.cursor;
            if (block0.jQ.is('span.mq-hasCursor') || block0.jQ.find(".mq-hasCursor").length > 0) {
                if (block2.jQ.is(".mq-empty")) {
                    block2.jQ.removeClass("mq-empty");
                }
                block2.children().remove();
                block2.jQ.children().remove();
                var block = latexMathParser.parse(block0.sLatex());
                block.children().adopt(block2, block2.ends[L], block2.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block2.jQ);
            } else if (block2.jQ.is('span.mq-hasCursor') || block2.jQ.find(".mq-hasCursor").length > 0) {
                if (block0.jQ.is(".mq-empty")) {
                    block0.jQ.removeClass("mq-empty");
                }
                block0.children().remove();
                block0.jQ.children().remove();
                var block = latexMathParser.parse(block2.sLatex());
                block.children().adopt(block0, block0.ends[L], block0.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block0.jQ);
            }
        };
        _.parser = function() {
            // \diffb{1}{x}{1}{x^2}
            // \diffb{{1}}{x}{1}\left(x^{2}\right) HandWriting
            var block = latexMathParser.block;
            var optWhitespace = Parser.optWhitespace;
            var succeed = Parser.succeed;
            var string = Parser.string;
            var self = this;
            var blocks = self.blocks = [MathBlock(),MathBlock(),MathBlock(),MathBlock()];  
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            //NNANH 2/28/2019 - S - Edit - Bug#10472 fix parser diffa diffb
            var check = 0;
            return (string("\\left(").then(function(){
                return latexMathParser.then(function(block) {
                    return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[3], blocks[3].ends[R], 0);
                            return succeed(self);
                        });
                    });
                })
            ).or(
            block.then(function(block) { 
                    // if (blocks[0].isEmpty()) {
                    //   block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                    // }
                    // else if (blocks[1].isEmpty()){     
                    //     block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                    // }
                    // else if (blocks[2].isEmpty()){
                    //     block.children().adopt(blocks[2], blocks[2].ends[R], 0);
                    // }
                    // else {
                    //     block.children().adopt(blocks[3], blocks[3].ends[R], 0);
                    // }
                    //NNANH 2/28/2019 - S - Edit - Bug#10472 fix parser diffa diffb
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                    //NNANH 2/28/2019 - E - Edit - Bug#10472 fix parser diffa diffb
                    return succeed(self);
               })
        ).atMost(4).result(self);
        }
    });
LatexCmds['diffc'] =
    LatexCmds.diffc = P(MathCommand, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =               
               '<span class="mq-matrix mq-non-leaf">' +                
                 '<div style="display: table;">' + 
                    '<div style="display: table-row;">'+   
                        '<div style="display: table-cell;vertical-align: middle;">'+
                            '<span class="mq-diffc-fraction mq-non-leaf">' +
                                '<span class="mq-numerator" style="padding: 0 0.5em"><var>d</var></span>' +
                                '<span class="mq-denominator" ><var>d</var><span>&0</span></span>' +
                                '<span style="display:inline-block;width:0">​&#8203</span>' +
                             '</span>' +
                        '</div>' +                     
                        '<div style="display: table-cell;vertical-align: middle;">'+ 
                            '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                            '<span class="mq-non-leaf">&1</span>' +
                            '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +  
                        '</div>'+
                        '<div style="display: table-cell;border-left: 1px solid;" class="mq-separate"></div>'+
                        '<div style="display: table-cell;vertical-align: bottom;padding-left: 0.15em;">'+                            
                            '<span>&2</span>' +
                            '<span class="mq-binary-operator">=</span>' +
                            '<span>&3</span>' +
                        '</div>'+
                    '</div>'+
                '</div>'+
              '</span>';
            Symbol.prototype.init.call(this, 'diffc', htmlTemplate);
        };
        _.latex = function() {
            if (this.blocks[0].latex() == "") {
                return OPEN_BRACKET_SERVER + 'diff' + '(' + this.blocks[1].latex() + 
                "," + 1 + "," + this.blocks[3].latex() +
                ')' + CLOSE_BRACKET_SERVER;
            }
            return OPEN_BRACKET_SERVER + 'diff' + '(' + this.blocks[1].latex() + "," +
                this.blocks[0].latex() + "," + 1 + "," + this.blocks[3].latex() +
                ')' + CLOSE_BRACKET_SERVER;
        };
        _.sLatex = function() {
            return '\\diffc' + '{' + this.blocks[0].sLatex() + "}{" + this.blocks[1].sLatex() + '}' + 
            '{' + this.blocks[2].sLatex() + "}" + '{' + this.blocks[3].sLatex() + "}";
        };
        _.nativeLatex = function() {
            return '\\diffc' + '{' + this.blocks[0].nativeLatex() + "}{" + this.blocks[1].nativeLatex() + '}' + 
            '{' + this.blocks[2].nativeLatex() + "}" + '{' + this.blocks[3].nativeLatex() + "}";
        };
        _.sLatexToSVG = function() {
            return "\\left.{"+"\\frac{d}{d" + this.blocks[0].sLatexToSVG() + "}{(" + this.blocks[1].sLatexToSVG() + ")}"
            + "}\\right\\vert_{" + this.blocks[2].sLatexToSVG() + "=" + this.blocks[3].sLatexToSVG() + "}";
        };
        _.reflow = function() {
            //resize the bracket ()            
            reflowRoundBracket(this.blocks[1].jQ);   
            //Resize End
            //Auto insert block 0 and block 2 Start
            var block0 = this.blocks[0];
            var block2 = this.blocks[2];
            var cursor = this.parent.cursor;
            if (block0.jQ.is('span.mq-hasCursor') || block0.jQ.find(".mq-hasCursor").length > 0) {
                if (block2.jQ.is(".mq-empty")) {
                    block2.jQ.removeClass("mq-empty");
                }
                block2.children().remove();
                block2.jQ.children().remove();
                var block = latexMathParser.parse(block0.sLatex());
                block.children().adopt(block2, block2.ends[L], block2.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block2.jQ);
            } else if (block2.jQ.is('span.mq-hasCursor') || block2.jQ.find(".mq-hasCursor").length > 0) {
                if (block0.jQ.is(".mq-empty")) {
                    block0.jQ.removeClass("mq-empty");
                }
                block0.children().remove();
                block0.jQ.children().remove();
                var block = latexMathParser.parse(block2.sLatex());
                block.children().adopt(block0, block0.ends[L], block0.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block0.jQ);
            }
            //Auto insert block 0 and block 2 End
        };
        _.parser = function() {
            // \diffc{x}{x^2}{x}{3}
            // \diffc_{x}\left(x^{2}\right)\right|x=3 HandWriting
            var block = latexMathParser.block;
            var optWhitespace = Parser.optWhitespace;
            var succeed = Parser.succeed;
            var string = Parser.string;
            var self = this;
            var blocks = self.blocks = [MathBlock(),MathBlock(),MathBlock(),MathBlock()];  
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            //NNANH - S - Add - 4/17/2019 - Bug#11387 parser diffc
            // var ownParser = block.times(4).map(function(block) { 
                // block[0].children().adopt(blocks[0], blocks[0].ends[R], 0);
                // block[1].children().adopt(blocks[1], blocks[1].ends[R], 0);
                // block[2].children().adopt(blocks[2], blocks[2].ends[R], 0);
                // block[3].children().adopt(blocks[3], blocks[3].ends[R], 0);
                // return self;
            // });
            var check = 0;
            // return (optWhitespace.then(string('_').or(string('\\left(').or(string('\\right|')).or(latexMathParser.optBlock).or(ownParser)).then(function(barrierSymbol) {
            return (optWhitespace.then(string('_').or(string('\\left(').or(string('\\right|')).or(latexMathParser.optBlock)).then(function(barrierSymbol) {
            if (barrierSymbol === "_") {
                 return block.then(function(block) {
                     block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                     block.children().adopt(blocks[2], blocks[2].ends[R], 0);
                     return succeed(self);
                 });
            } else if (barrierSymbol === "\\left(") {
                return latexMathParser.then(function(block) {
                    return string('\\right)').map(function(end) {
                        block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                        return self;
                    });
                });
            }
            else if (barrierSymbol.ctrlSeq) {
                return succeed(self);
            } else {
                 x = 0;
                return block.then(function(block) {

                     if(block.ends[R].ctrlSeq === "=")
                     {
                          x = 1;
                          return succeed(self);
                     }
                     if(x === 1){
                         block.children().adopt(blocks[3], blocks[3].ends[R], 0);
                     }
                     return succeed(self);
                 }).many().result(self);           
            }
        })))
            .or(     
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                    return succeed(self);
                })
            )
            .atMost(4).result(self);;
        }
        //NNANH - E - 4/17/2019 - Add - Bug#11387 parser diffc
    });
LatexCmds['diffd'] =
    LatexCmds.diffd = P(MathCommand, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =            
            '<span class="mq-matrix mq-non-leaf">' +                
                 '<div style="display: table;">' + 
                    '<div style="display: table-row;">'+   
                        '<div style="display: table-cell;vertical-align: middle;">'+
                            '<span class="mq-diffc-fraction mq-non-leaf">' +
                                '<span class="mq-numerator" style="padding: 0 0.5em">'+
                                    '<var>d</var>' +
                                    '<span class="mq-supsub mq-non-leaf">' +
                                        '<span class="mq-sup"><span class="mq-mathcall-sup-inner">&0</span></span>' +
                                        '<span style="display:inline-block;width:0">&#8203</span>' +
                                    '</span>' +
                                '</span>' +
                                '<span class="mq-denominator" >'+
                                    '<var>d</var>' +
                                    '<span>&1</span>' +
                                    '<span class="mq-supsub mq-non-leaf">' +
                                        '<span class="mq-sup"><span class="mq-non-leaf">&2</span></span>' +
                                        '<span style="display:inline-block;width:0">&#8203</span>' +
                                    '</span>' +
                                '</span>' +
                                '<span style="display:inline-block;width:0">​&#8203</span>' +
                             '</span>' +
                        '</div>' +                     
                        '<div style="display: table-cell;vertical-align: middle;">'+ 
                            '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                            '<span class="mq-non-leaf">&3</span>' +
                            '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +  
                        '</div>'+
                        '<div style="display: table-cell;border-left: 1px solid;" class="mq-separate"></div>'+
                        '<div style="display: table-cell;vertical-align: bottom;padding-left: 0.15em;">'+                            
                            '<span>&4</span>' +
                            '<span class="mq-binary-operator">=</span>' +
                            '<span>&5</span>' +
                        '</div>'+
                    '</div>'+
                '</div>'+
              '</span>';
            Symbol.prototype.init.call(this, '\\diffd', htmlTemplate);
        };
        _.latex = function() {
            if (this.blocks[2].latex() == "") {
                return OPEN_BRACKET_SERVER + 'diff' + '(' + this.blocks[3].latex() + "," +
                    this.blocks[1].latex() + "," + this.blocks[5].latex() +                    
                    ')' + CLOSE_BRACKET_SERVER;
            }
            return OPEN_BRACKET_SERVER + 'diff' + '(' + this.blocks[3].latex() + "," +
                this.blocks[1].latex() + "," + this.blocks[2].latex() +
                "," + this.blocks[5].latex() +
                ')' + CLOSE_BRACKET_SERVER;
        };
        _.sLatex = function() {
            return '\\diffd' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + "}{" + this.blocks[2].sLatex() + '}{' + this.blocks[3].sLatex() + '}'
                + '{' + this.blocks[4].sLatex() + '}' + '{' + this.blocks[5].sLatex() + '}';
        };
        _.nativeLatex = function() {
            return '\\diffd' + '{' + this.blocks[0].nativeLatex() + "}{" +
                this.blocks[1].nativeLatex() + "}{" + this.blocks[2].nativeLatex() + '}{' + this.blocks[3].nativeLatex() + '}'
                + '{' + this.blocks[4].nativeLatex() + '}' + '{' + this.blocks[5].nativeLatex() + '}';
        };
        _.sLatexToSVG = function() {
            return "\\left.{"+"\\frac{d^{" + this.blocks[0].sLatexToSVG() + "}}{d{" + this.blocks[1].sLatexToSVG() + "}^{" +
                this.blocks[2].sLatexToSVG() + "}}{(" + this.blocks[3].sLatexToSVG() + ")}" +
                "}\\right\\vert_{" + this.blocks[2].sLatexToSVG() + "=" + this.blocks[3].sLatexToSVG() + "}";
        };
        _.reflow = function() {
            //resize the bracket ()
            reflowRoundBracket(this.blocks[3].jQ); 
            //Resize End            
            //Auto insert block 0 and block 2 Start
            var block0 = this.blocks[0];
            var block2 = this.blocks[2];
            var cursor = this.parent.cursor;
            if (block0.jQ.is('span.mq-hasCursor') || block0.jQ.find(".mq-hasCursor").length > 0) {
                if (block2.jQ.is(".mq-empty")) {
                    block2.jQ.removeClass("mq-empty");
                }
                block2.children().remove();
                block2.jQ.children().remove();
                var block = latexMathParser.parse(block0.sLatex());
                block.children().adopt(block2, block2.ends[L], block2.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block2.jQ);
            } else if (block2.jQ.is('span.mq-hasCursor') || block2.jQ.find(".mq-hasCursor").length > 0) {
                if (block0.jQ.is(".mq-empty")) {
                    block0.jQ.removeClass("mq-empty");
                }
                block0.children().remove();
                block0.jQ.children().remove();
                var block = latexMathParser.parse(block2.sLatex());
                block.children().adopt(block0, block0.ends[L], block0.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block0.jQ);
            }
            //Auto insert block 0 and block 2 End
            //Auto insert block 1 and block 4 Start
            var block1 = this.blocks[1];
            var block4 = this.blocks[4];
            if (block1.jQ.is('span.mq-hasCursor') || block1.jQ.find(".mq-hasCursor").length > 0) {
                if (block4.jQ.is(".mq-empty")) {
                    block4.jQ.removeClass("mq-empty");
                }
                block4.children().remove();
                block4.jQ.children().remove();
                var block = latexMathParser.parse(block1.sLatex());
                block.children().adopt(block4, block4.ends[L], block4.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block4.jQ);
            } else if (block4.jQ.is('span.mq-hasCursor') || block4.jQ.find(".mq-hasCursor").length > 0) {
                if (block1.jQ.is(".mq-empty")) {
                    block1.jQ.removeClass("mq-empty");
                }
                block1.children().remove();
                block1.jQ.children().remove();
                var block = latexMathParser.parse(block4.sLatex());
                block.children().adopt(block1, block1.ends[L], block1.ends[R]);
                var jQ = block.jQize();
                jQ.appendTo(block1.jQ);
            }
            //Auto insert block 1 and block 4 End
        };
        _.parser = function() {
            // \diffd{1}{x}{1}{x^2}{x}{1}
            //\diffd_{{1}}{x}{{1}}\left(x^{2}\right)\right|x=15 HandWriting
            var block = latexMathParser.block;
            var optWhitespace = Parser.optWhitespace;
            var succeed = Parser.succeed;
            var string = Parser.string;
            var self = this;
            var blocks = self.blocks = [MathBlock(),MathBlock(),MathBlock(),MathBlock(),MathBlock(),MathBlock()];  
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            var ownParser = block.times(6).map(function(block) { 
            block[0].children().adopt(blocks[0], blocks[0].ends[R], 0);
            block[1].children().adopt(blocks[1], blocks[1].ends[R], 0);
            block[2].children().adopt(blocks[2], blocks[2].ends[R], 0);
            block[3].children().adopt(blocks[3], blocks[3].ends[R], 0);
            block[4].children().adopt(blocks[4], blocks[4].ends[R], 0);
            block[5].children().adopt(blocks[5], blocks[5].ends[R], 0);
            return self;
        });
            return (optWhitespace.then(string('_').or(string('\\left(').or(string('\\right|')).or(latexMathParser.optBlock).or(ownParser)).then(function(barrierSymbol) {
            if (barrierSymbol === "_") {
                 return block.then(function(block) {
                     if(blocks[0].isEmpty()){
                        block.children().adopt(blocks[0], blocks[0].ends[R], 0); 
                     }
                     else if(blocks[1].isEmpty()){
                         block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                         block.children().adopt(blocks[4], blocks[4].ends[R], 0);
                     }
                     else if(blocks[2].isEmpty()){
                         block.children().adopt(blocks[2], blocks[2].ends[R], 0);
                     }
                     return succeed(self);
                 }).atMost(3).result(self);
            } else if (barrierSymbol === "\\left(") {
                return latexMathParser.then(function(block) {
                    return string('\\right)').map(function(end) {
                        block.children().adopt(blocks[3], blocks[3].ends[R], 0);
                        return self;
                    });
                });
            }
            else if (barrierSymbol.ctrlSeq) {
                return succeed(self);
            } else {
                 x = 0;
                return block.then(function(block) {

                     if(block.ends[R].ctrlSeq === "=")
                     {
                          x = 1;
                          return succeed(self);
                     }
                     if(x === 1){
                         block.children().adopt(blocks[5], blocks[5].ends[R], 0);
                     }
                     return succeed(self);
                 }).many().result(self);           
            }
        }).atMost(6).result(self)));
        }
    });

var Fourier = P(MathCommand, function(_, supper_) {
    _.latex = function() {
        return OPEN_BRACKET_SERVER + this.ctrlSeq + '(' + this.blocks[1].latex() + "," +
            this.blocks[0].latex() + "," + this.blocks[2].latex() + ')' + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\' + this.ctrlSeq + '{' + this.blocks[0].sLatex() + "}{" +
            this.blocks[1].sLatex() + "}{" + this.blocks[2].sLatex() + '}';
    };
    _.nativeLatex = function() {
        return '\\' + this.ctrlSeq + '{' + this.blocks[0].nativeLatex() + "}{" +
            this.blocks[1].nativeLatex() + "}{" + this.blocks[2].nativeLatex() + '}';
    };
    _.reflow = function() {
        var scaleBracket = reflowRoundBracket(this.blocks[1].jQ); 
        if (scaleBracket > 3) {
            this.blocks[1].jQ.parent().css({"vertical-align":"middle"});
        } else {
            this.blocks[1].jQ.parent().css("vertical-align","");
        }
    };
    _.parser = function() {
        var block = latexMathParser.block;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var string = Parser.string;
        var self = this;
        var blocks = self.blocks = [MathBlock(),MathBlock(),MathBlock()];  
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }       
        //\fourier{1}{2}{3}
        //\fourier_{2}\left(5\right)[3]
        //\fourier_{2}\left(5\right)\left[3\right]
        var i = 0;
        return (optWhitespace.then(string('_').or(string('\\left(').or(string('\\left[')).or(latexMathParser.optBlock)).or(block)).then(function(barrierSymbol) {
            if (barrierSymbol === "_") {
                 i++;
                 return block.then(function(block) {
                     block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                     return succeed(self);
                 });
            } else if (barrierSymbol === "\\left(") {
                i++;
                return latexMathParser.then(function(block) {
                    return string('\\right)').map(function(end) {
                        block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                        return self;
                    });
                });
            } else if(barrierSymbol === "\\left["){
                i++;
                return latexMathParser.then(function(block) {
                    return string('\\right]').map(function(end) {
                        block.children().adopt(blocks[2], blocks[2].ends[R], 0);
                        return self;
                    });
                });
            } else {
               if (blocks[i]) {
                   barrierSymbol.children().adopt(blocks[i], blocks[i].ends[R], 0);
                   i++;
               }
               return succeed(self);
            }
        }).atMost(3).result(self));
    };
});

var invFourier = P(Fourier, function(_, supper_) {
    _.parser = function() {
        var block = latexMathParser.block;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var string = Parser.string;
        var self = this;
        var blocks = self.blocks = [MathBlock(),MathBlock(),MathBlock()];  
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }  
        var i = 0;             
        //\invfourier_{2}^{-1}\left(5\right)\left[3\right]
        return (optWhitespace.then(string('_').or(string('\\left(').or(string('\\left[')).or(latexMathParser.optBlock)).or(block)).then(function(barrierSymbol) {
            if (barrierSymbol === "_") {
                 i++;
                 return block.then(function(block) {
                     block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                     return succeed(self);
                 }).then(string('^')).then(function() {
                    i++;
                    return block.then(function(block) {
                        return succeed(self);
                    });
                 });
            } else if (barrierSymbol === "\\left(") {
                i++;
                return latexMathParser.then(function(block) {
                    return string('\\right)').map(function(end) {
                        block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                        return self;
                    });
                });
            } else if(barrierSymbol === "\\left[") {
                i++;
                return latexMathParser.then(function(block) {
                    return string('\\right]').map(function(end) {
                        block.children().adopt(blocks[2], blocks[2].ends[R], 0);
                        return self;
                    });
                });
            } else {
               if (i < 3 && blocks[i]) {
                   barrierSymbol.children().adopt(blocks[i], blocks[i].ends[R], 0);
                   i++;
               }
               return succeed(self);
            }
        }).atMost(3).result(self));
    };
});
LatexCmds['fourier'] =
    LatexCmds.fourier = P(Fourier, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-int mq-non-leaf">' +
                '<var style="font-family: \'SUDMin_WASciCalc2\';">&#xe902;</var>' +
                '<span class="mq-supsub mq-non-leaf">' +
                '<span class="mq-sub"  style="margin-top: -0.5em;">&0</span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&1</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>' +
                '<span class="mq-matrix mq-non-leaf">' +
                    '<div style="display: table;margin-bottom: 4px;">' +                    
                        '<div style="display: table-row;height: 1.2em;" class="mq-square-bracket">'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-left: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                            '<div style="display: table-cell;"><div style="margin-left: 0.15em;margin-right: 0.15em;margin-top: 0.1em;">&2</div></div>'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-right: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                        '</div>'+                   
                    '</div>'+
                '</span>';
            Symbol.prototype.init.call(this, 'fourier', htmlTemplate);
        };
        _.sLatexToSVG = function() {
            return '\ue902' + '_{' + this.blocks[0].sLatexToSVG() + "}{(" +
                this.blocks[1].sLatexToSVG() + ")}{[" + this.blocks[2].sLatexToSVG() + ']}';
        };
    });
LatexCmds['invfourier'] =
    LatexCmds.invfourier = P(invFourier, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-int mq-non-leaf">' +
                '<var style="font-family: \'SUDMin_WASciCalc2\';">&#xe902;</var>' +
                '<span class="mq-supsub mq-non-leaf">' +
                '<span class="mq-sup"><span class="mq-mathcall-sup-inner">-1</span></span>' +
                '<span class="mq-sub" style="margin-top: -0.5em;">&0</span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&1</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>' +
                '<span class="mq-matrix mq-non-leaf">' +
                    '<div style="display: table;margin-bottom: 4px;">' +                    
                        '<div style="display: table-row;height: 1.2em;" class="mq-square-bracket">'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-left: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                            '<div style="display: table-cell;"><div style="margin-left: 0.15em;margin-right: 0.15em;margin-top: 0.1em;">&2</div></div>'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-right: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                        '</div>'+                   
                    '</div>'+
                '</span>';;
            Symbol.prototype.init.call(this, 'invfourier', htmlTemplate);
        };
        _.sLatexToSVG = function() {
            return '\ue902' + '_{' + this.blocks[0].sLatexToSVG() + "}^{\uEDFC}{(" +
                this.blocks[1].sLatexToSVG() + ")}{[" + this.blocks[2].sLatexToSVG() + ']}';
        };
    });

LatexCmds['laplace'] =
    LatexCmds.laplace = P(Fourier, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-int mq-non-leaf">' +
                '<var style="font-family: \'SUDMin_WASciCalc2\';">&#xe901;</var>' +
                '<span class="mq-supsub mq-non-leaf" style="vertical-align: -0.5em;">' +
                '<span class="mq-sub" style="margin-left: 0em;">&0</span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&1</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>' +
                '<span class="mq-matrix mq-non-leaf">' +
                    '<div style="display: table;margin-bottom: 4px;">' +                    
                        '<div style="display: table-row;height: 1.2em;" class="mq-square-bracket">'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-left: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                            '<div style="display: table-cell;"><div style="margin-left: 0.15em;margin-right: 0.15em;margin-top: 0.1em;">&2</div></div>'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-right: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                        '</div>'+                   
                    '</div>'+
                '</span>';;
            Symbol.prototype.init.call(this, 'laplace', htmlTemplate);
        };
        _.sLatexToSVG = function() {
            return '\ue901' + '_{' + this.blocks[0].sLatexToSVG() + "}{(" +
                this.blocks[1].sLatexToSVG() + ")}{[" + this.blocks[2].sLatexToSVG() + ']}';
        };
    });

LatexCmds['invlaplace'] =
    LatexCmds.invlaplace = P(invFourier, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-int mq-non-leaf">' +
                '<var style="font-family: \'SUDMin_WASciCalc2\';">&#xe901;</var>' +
                '<span class="mq-supsub mq-non-leaf">' +
                '<span class="mq-sup"><span class="mq-mathcall-sup-inner" style="padding-left: 0em;">-1</span></span>' +
                '<span class="mq-sub" style="margin-left: 0em;margin-top: -0.5em;">&0</span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&1</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>' +
                '<span class="mq-matrix mq-non-leaf">' +
                    '<div style="display: table;margin-bottom: 4px;">' +                    
                        '<div style="display: table-row;height: 1.2em;" class="mq-square-bracket">'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-left: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                            '<div style="display: table-cell;"><div style="margin-left: 0.15em;margin-right: 0.15em;margin-top: 0.1em;">&2</div></div>'+
                            '<div class="mq-paren" style="display: table-cell;border-top: 2px solid;border-right: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                        '</div>'+                   
                    '</div>'+
                '</span>';;
            Symbol.prototype.init.call(this, 'invlaplace', htmlTemplate);
        };
        _.sLatexToSVG = function() {
            return '\ue901' + '_{' + this.blocks[0].sLatexToSVG() + "}^{\uEDFC}{(" +
                this.blocks[1].sLatexToSVG() + ")}{[" + this.blocks[2].sLatexToSVG() + ']}';
        };
    });

LatexCmds['lim'] =
    LatexCmds.lim = P(MathCommand, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-non-leaf mq-lim-operator">' +
                    '<span>lim</span>' +              
                    '<span class="mq-lim-condition">' +
                        '<span>&0</span>' +
                        '<span class="mq-binary-operator">\uEE65</span>' +
                        '<span>&1</span>' +
                    '</span>' +  
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&2</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>';
            Symbol.prototype.init.call(this, 'lim', htmlTemplate);
        };
        _.latex = function() {
            return OPEN_BRACKET_SERVER + this.ctrlSeq + '(' + this.blocks[2].latex() + "," +
                this.blocks[0].latex() + "," + this.blocks[1].latex() + ')' + CLOSE_BRACKET_SERVER;
        };
        _.sLatex = function() {
            return '\\' + this.ctrlSeq + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + "}{" + this.blocks[2].sLatex() + '}';
        };
        _.nativeLatex = function() {
            return '\\' + this.ctrlSeq + '_{' + this.blocks[0].nativeLatex() + " \\to " +
                this.blocks[1].nativeLatex() + "}{\\left(" + this.blocks[2].nativeLatex() + '\\right)}';
        };
        _.sLatexToSVG = function() {
            return "\\underset{" + this.blocks[0].sLatexToSVG() + "\u2192" +
                this.blocks[1].sLatexToSVG() + "}{\\lim}" + "{(" + this.blocks[2].sLatexToSVG() + ')}';
        };
        _.reflow = function() {
            reflowRoundBracket(this.blocks[2].jQ); 
        };
        _.parser = function() {
            var string = Parser.string;
            var optWhitespace = Parser.optWhitespace;
            var succeed = Parser.succeed;
            var block = latexMathParser.block;

            var self = this;
            var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            //\lim_{k \to 5} {\left(k+1\right)} Bug6621
            //\lim _{k\rightarrow 2}\left( x\right) HandWriting
            //\lim{}{}{}
           return (optWhitespace.then(string('_')).then(function(text){
                return (block.times(2).map(function(dataBlock) {
                    var firstBloclk = dataBlock[0];
                    var secondBlock = dataBlock[1];
                    var leftBlock = firstBloclk.ends[L];
                    var rightBlock = firstBloclk.ends[R];
                    var expectToBlock = leftBlock;
                    var leftExpectToBlock = expectToBlock[L];
                    while (expectToBlock !== 0 && (expectToBlock.ctrlSeq.trim() !== "\\to" && expectToBlock.ctrlSeq.trim() !== "\\rightarrow")) {
                        leftExpectToBlock = expectToBlock;
                        expectToBlock = expectToBlock[R];
                    }
                    if (expectToBlock !== 0) {
                        Fragment(leftBlock, expectToBlock[L]).adopt(blocks[0], blocks[0].ends[R], 0);
                        if (expectToBlock[R] && expectToBlock[R] !== 0) {
                             Fragment(expectToBlock[R], rightBlock).adopt(blocks[1], blocks[1].ends[R], 0);
                        }
                    } else if (leftExpectToBlock) {
                        Fragment(leftBlock, leftExpectToBlock).adopt(blocks[0], blocks[0].ends[R], 0);
                    } else {
                        Fragment(leftBlock, leftBlock).adopt(blocks[0], blocks[0].ends[R], 0);
                    }
                   
                    var bracketBlock = secondBlock.ends[L];
                    if (bracketBlock !== 0 && bracketBlock.ctrlSeq === "\\left(") {
                         bracketBlock.ends[L].children().adopt(blocks[2], blocks[2].ends[R], 0);
                    } else {
                         bracketBlock.children().adopt(blocks[2], blocks[2].ends[R], 0);
                    }                   
                    return self;
                }))
                .or(
                    latexMathParser.then(function(dataBlock){
                        var leftBlock = dataBlock.ends[L];
                        var rightBlock = dataBlock.ends[R];
                        var expectToBlock = leftBlock;
                        var leftExpectToBlock = expectToBlock[L];
                        while (expectToBlock !== 0 && (expectToBlock.ctrlSeq.trim() !== "\\to" && expectToBlock.ctrlSeq.trim() !== "\\rightarrow")) {
                            leftExpectToBlock = expectToBlock;
                            expectToBlock = expectToBlock[R];
                        }
                        if (expectToBlock !== 0) {
                            Fragment(leftBlock, expectToBlock[L]).adopt(blocks[0], blocks[0].ends[R], 0);
                            if (expectToBlock[R] && expectToBlock[R] !== 0) {
                                 Fragment(expectToBlock[R], rightBlock).adopt(blocks[1], blocks[1].ends[R], 0);
                            }
                        } else if (leftExpectToBlock) {
                            Fragment(leftBlock, leftExpectToBlock).adopt(blocks[0], blocks[0].ends[R], 0);
                        } else if (leftBlock !== 0) {
                            Fragment(leftBlock, leftBlock).adopt(blocks[0], blocks[0].ends[R], 0);
                        }
                        return succeed(self);
                    }
                ))
            })).or(block.times(self.numBlocks()).map(function(dataBlock) {
                    dataBlock[0].children().adopt(blocks[0], blocks[0].ends[R], 0);
                    dataBlock[1].children().adopt(blocks[1], blocks[1].ends[R], 0);
                    dataBlock[2].children().adopt(blocks[2], blocks[2].ends[R], 0);
                    return self;
            //Bug#9395 NNANH 11/8/2018 Fix error parser Lim loop three time
            })).result(self);
            //Bug#9395 NNANH 11/8/2018 Fix error parser Lim loop three time           
        };
    });

    LatexCmds['log'] =
    LatexCmds.log = P(MathCommand, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-non-leaf mq-log">' +
                '<span class="mq-margin-log mq-empty"></span>' +
                '<span style="display:inline-block;width:0">&#8203;</span>' +
                '</span>' +
                '<span>log</span>' +
                '<span class="mq-supsub mq-non-leaf">' +
                '<span class="mq-sub">' +
                '<span>&0</span>' +
                '</span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                '<span class="mq-non-leaf">&1</span>' +
                '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                '</span>';
            Symbol.prototype.init.call(this, '\\log', htmlTemplate);
        };
        _.latex = function() {
            return OPEN_BRACKET_SERVER + this.ctrlSeq + '(' + this.blocks[0].latex() + ',' + this.blocks[1].latex() + ')' + CLOSE_BRACKET_SERVER;
        };
        _.sLatexToSVG = function() {
            return '\\log_{' + this.blocks[0].sLatexToSVG() + '}{(' + this.blocks[1].sLatexToSVG() + ')}';
        };
        _.nativeLatex = function() {
            return this.ctrlSeq + '_{' + this.blocks[0].nativeLatex() +
                     "}\\left(" + this.blocks[1].nativeLatex() + '\\right)';
        };
        _.reflow = function() {
            reflowRoundBracket(this.blocks[1].jQ); 
        };
        _.parser = function() {
            var string = Parser.string;
            var optWhitespace = Parser.optWhitespace;
            var succeed = Parser.succeed;
            var block = latexMathParser.block;

            var self = this;
            var blocks = self.blocks = [MathBlock(), MathBlock()];
            for (var i = 0; i < blocks.length; i += 1) {
                blocks[i].adopt(self, self.ends[R], 0);
            }
            //\log{ }{ }
            //\log_{2}\left(4\right) Bug #6621
            return ((optWhitespace.then(string('_')).then(function(text){
                    return block.map(function(block){
                        block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                        return self;
                    });
                })).then(string("\\left(").then(function(text){
                    return latexMathParser.then(function(block) {
                        return (string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return self;
                        })).or(optWhitespace.then(function() {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        }));
                    });
                }))).or(string("\\left(").then(function (text) {
                  // \log\left(2\right)
                  self.htmlTemplate =
                    '<span class="mq-non-leaf mq-log">' +
                    '<span class="mq-margin-log mq-empty"></span>' +
                    '<span style="display:inline-block;width:0">&#8203;</span>' +
                    '</span>' +
                    '<span>log</span>' +
                    '<span class="mq-non-leaf">' +
                    '<span class="mq-scaled mq-paren" style="vertical-align: top;">(</span>' +
                    '<span class="mq-non-leaf">&1</span>' +
                    '<span class="mq-scaled mq-paren" style="vertical-align: top;">)</span>' +
                    '</span>';

                  self.latex = function() {
                      return OPEN_BRACKET_SERVER + this.ctrlSeq + '(10,' + this.blocks[1].latex() + ')' + CLOSE_BRACKET_SERVER;
                  };
                  self.sLatexToSVG = function() {
                      return '\\log_{10}{(' + this.blocks[1].sLatexToSVG() + ')}';
                  };
                  self.nativeLatex = function() {
                      return this.ctrlSeq + "\\left(" + this.blocks[1].nativeLatex() + '\\right)';
                  };
                  self.sLatex = function() {
                      return this.ctrlSeq + "\\left(" + this.blocks[1].sLatex() + '\\right)';
                  };

                  return latexMathParser.then(function(block) {
                      return (string('\\right)').map(function(end) {
                          block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                          return self;
                      })).or(optWhitespace.then(function() {
                          block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                          return succeed(self);
                      }));
                  });
              })).or(block.times(self.numBlocks()).then(function(dataBlock) {
                        dataBlock[0].children().adopt(blocks[0], blocks[0].ends[R], 0);
                        dataBlock[1].children().adopt(blocks[1], blocks[1].ends[R], 0);
                        return succeed(self);
                }));
        };
    });

LatexCmds['mixedFraction'] = P(MathCommand, function(_, supper_) {
    _.init = function() {
        var htmlTemplate =
            '<span class="mq-mixed-deca mq-non-leaf">' +
            '<span class="mixed-deca mq-mixedFraction" style="margin-right:-3px; padding:0.125em;">&0</span>' +
            '</span>' +
            '<span class="mq-fraction mq-non-leaf">' +
            '<span class="mq-numerator mq-mixedFraction" style="padding:0.125em;">&1</span>' +
            '<span class="mq-denominator mq-mixedFraction" style="padding:0.125em;">&2</span>' +
            '<span style="display:inline-block;width:0">&#8203;</span>' +
            '</span>';
        Symbol.prototype.init.call(this, '\\mixedFraction', htmlTemplate);
    };
    _.finalizeTree = function(options) {
            this.ends[L][R].downOutOf = this.ends[R];
            this.ends[R].upOutOf = this.ends[L][R];
    };
    _.latex = function() {
        // rm1095 if first block contains a negative, it must be split out
        var block0Tex = this.blocks[0].latex();
        var negativeClose = '';
        if (block0Tex[0] == '-') {
            var tempLatexArray = block0Tex.split(/(-)/);
            // make sure we're dealing with what we think we are
            if (tempLatexArray.shift() == '' && tempLatexArray.length == 2) {
                block0Tex = tempLatexArray.join('(');
                negativeClose = ')'
            }
        }
        return OPEN_BRACKET_SERVER + block0Tex + "+(" + this.blocks[1].sLatex() + ")/" +"(" + this.blocks[2].sLatex() + ")" + negativeClose + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return this.ctrlSeq + "{" + this.blocks[0].sLatex() + "}" + "{" + this.blocks[1].sLatex() + "}" +"{" + this.blocks[2].sLatex() + "}";
    };
    _.sLatexToSVG = function() {
        return this.blocks[0].sLatexToSVG() + "\\frac{" +   this.blocks[1].sLatexToSVG()  + "}{" +   this.blocks[2].sLatexToSVG() + "}";
    };
});  

LatexCmds['updelta'] =
    LatexCmds.updelta = P(MathCommand, function(_, supper_) {
        _.init = function() {
            var htmlTemplate =
                '<span class="mq-int mq-non-leaf">' +
                '<span>&delta;</span>' +
                '<span class="mq-supsub mq-non-leaf" style="vertical-align: 0.2em;">' +
                '<span class="mq-sup"><span>&0</span></span>' +
                '<span style="display:inline-block;width:0">&#8203</span>' +
                '</span>' +
                '</span>';
            Symbol.prototype.init.call(this, '\\delta', htmlTemplate);
        };
        _.latex = function() {
            function simplify(latex) {
                return latex.length === 1 ? latex : '{' + (latex || ' ') + '}';
            }
            return this.ctrlSeq + '^' + simplify(this.ends[L].latex());
        };
    });

var Fraction =
    LatexCmds.frac =
    LatexCmds.dfrac =
    LatexCmds.cfrac =
    LatexCmds.fraction = P(MathCommand, function(_, super_) {
  _.ctrlSeq = '\\frac';
  _.htmlTemplate =
            '<span class="mq-fraction mq-non-leaf">' +
            '<span class="mq-numerator" style="padding:0.125em;">&0</span>' +
            '<span class="mq-denominator" style="padding:0.125em;">&1</span>' +
            '<span style="display:inline-block;width:0">&#8203;</span>' +
            '</span>';
  _.textTemplate = ['(', ')/(', ')'];
  _.finalizeTree = function() {
    this.upInto = this.ends[R].upOutOf = this.ends[L];
    this.downInto = this.ends[L].downOutOf = this.ends[R];
    this.ends[L].ariaLabel = 'numerator';
    this.ends[R].ariaLabel = 'denominator';
    if(this.getFracDepth() > 1) this.mathspeakTemplate = ['StartNestedFraction,', 'NestedOver', ', EndNestedFraction'];
    else this.mathspeakTemplate = ['StartFraction,', 'Over', ', EndFraction'];
  };
  _.mathspeak = function(opts) {
    if (opts && opts.createdLeftOf) {
      var cursor = opts.createdLeftOf;
      return cursor.parent.mathspeak();
    }
    return super_.mathspeak.apply(this, arguments);
  };

  _.getFracDepth = function() {
    var level = 0;
    var walkUp = function(item, level) {
      if(item instanceof Node && item.ctrlSeq && item.ctrlSeq.toLowerCase().search('frac') >= 0) level += 1;
      if(item.parent) return walkUp(item.parent, level);
      else return level;
    };
    return walkUp(this, level);
  };
        _.createLeftOf = function(cursor) {
            //NNANH 26/9/2018 Start CR#8604 
            if (!this.replacedFragment) {
                var leftward = cursor[L];
                while (leftward &&
                    !(
                        leftward instanceof BinaryOperator ||
                        leftward instanceof(LatexCmds.text || noop) ||
                        leftward instanceof SummationNotation ||
                        leftward.ctrlSeq === '\\ ' ||
                        /^[,;:]$/.test(leftward.ctrlSeq)
                    ) //lookbehind for operator
                ) leftward = leftward[L];

                if (leftward instanceof SummationNotation && leftward[R] instanceof SupSub) {
                    leftward = leftward[R];
                    if (leftward[R] instanceof SupSub && leftward[R].ctrlSeq != leftward.ctrlSeq)
                        leftward = leftward[R];
                }

                if (leftward !== cursor[L]) {
                    this.replaces(Fragment(leftward[R] || cursor.parent.ends[L], cursor[L]));
                    cursor[L] = leftward;
                }
            }
            //NNANH 26/9/2018 End CR#8604 
            if (cursor.options.restrictFractionLevel) {
                var currentFrac = cursor.parent.parent;
                var i = 0;
                while (i < cursor.options.restrictFractionLevel) {
                    if (!(currentFrac.ctrlSeq && currentFrac.ctrlSeq === "\\frac")) {
                        break;
                    }
                    i++;
                    currentFrac = currentFrac.parent.parent;
                }
                if (i === cursor.options.restrictFractionLevel) return;
            }
            return super_.createLeftOf.apply(this, arguments);
        };
        //NNANH end CR#8610 
        _.reflow = function(){
            var lengthOfSupSup = this.jQ.closest(".mq-root-block").find(".mq-power-sup-sup").length;
            for(var i = 0; i < lengthOfSupSup; i++){
                reflowPowerSup($(this.jQ.closest(".mq-root-block").find(".mq-power-sup-sup")[i]));
            } 
        }
        //NNANH end CR#8610 
    });


var LiveFraction =
LatexCmds.over =
CharCmds['/'] = P(Fraction, function(_, super_) {
  _.createLeftOf = function(cursor) {
    if (!this.replacedFragment) {
      var leftward = cursor[L];

      if (!cursor.options.typingSlashCreatesNewFraction) {
        while (leftward &&
          !(
            leftward instanceof BinaryOperator ||
            leftward instanceof (LatexCmds.text || noop) ||
            leftward instanceof SummationNotation ||
            leftward.ctrlSeq === '\\ ' ||
            /^[,;:]$/.test(leftward.ctrlSeq)
          ) //lookbehind for operator
        ) leftward = leftward[L];
      }
      if (leftward instanceof SummationNotation && leftward[R] instanceof SupSub) {
        leftward = leftward[R];
        if (leftward[R] instanceof SupSub && leftward[R].ctrlSeq != leftward.ctrlSeq)
          leftward = leftward[R];
      }

      if (leftward !== cursor[L]) {
        this.replaces(Fragment(leftward[R] || cursor.parent.ends[L], cursor[L]));
        cursor[L] = leftward;
      }
    }
    super_.createLeftOf.call(this, cursor);
  };
});

LatexCmds.ans = P(Symbol, function(_, super_) {
  _.init = function(ch) {
    super_.init.call(this,
      '\\operatorname{ans}',
      '<span class="mq-ans">ans</span>',
      'ans'
    );
  };
});

LatexCmds.percent =
LatexCmds.percentof = P(Symbol, function (_, super_) {
  _.init = function () {
    super_.init.call(
      this,
      '\\%\\operatorname{of}',
      '<span class="mq-nonSymbola mq-operator-name">% of </span>',
      'percent of'
    )
  };
});

var SquareRoot =
    LatexCmds.sqrt =
    LatexCmds['√'] = P(MathCommand, function(_, super_) {
  _.ctrlSeq = '\\sqrt';
        _.bigSquareRoot = '<div class="mq-new-sqrt mq-symbol-font" style="float:left;margin:-0.05em -0.5em 0px 0px;display:none;">'+
                         '<div class="mq-vline">⎪</div>'+
                        '<div class="mq-vline">⎪</div>'+
                        '<div class="squareRoot">&#x23B7;</div>'+
                        '<div style="clear:both;"></div>'+
                    '</div>';
        _.normalSquareRoot = '<div class="old-sqrt" style="overflow: hidden;font-family: MJXc-TeX-main-R,MJXc-TeX-main-Rw !important;margin-top: -5px;">&radic;</div>';
        /* _.htmlTemplate =
             '<span class="mq-non-leaf">'
                +   '<span class="mq-scaled mq-sqrt-prefix" style="overflow: hidden;font-family: MJXc-TeX-main-R,MJXc-TeX-main-Rw !important;">&radic;</span>'
    +   '<span class="mq-non-leaf mq-sqrt-stem">&0</span>'
                + '</span>' */  
        _.numBlocks = function() { return 1; };
        _.html = function() {
            this.htmlTemplate =  '<span class="mq-non-leaf">' +
                            '<table class="mq-non-leaf mq-sqrt">' +
                                '<tbody class="mq-sqrt" style="padding-top: 3px;">' +
                                    '<tr>' +
                                        '<td class="squareRoot">'+
                                            this.normalSquareRoot + 
                                            this.bigSquareRoot + 
                                        '</td>' +
                                        '<td class="mq-non-leaf mq-sqrt-stem squareRoot">&0</td>' +
                                    '</tr>' +
                                '</tbody>' +
                            '</table>' +
                        '</span>';
            return super_.html.call(this);
        };
  _.textTemplate = ['sqrt(', ')'];
  _.mathspeakTemplate = ['StartRoot,', ', EndRoot'];
  _.ariaLabel = 'root';
  _.parser = function() {
    return latexMathParser.optBlock.then(function(optBlock) {
      return latexMathParser.block.map(function(block) {
        var nthroot = NthRoot();
        nthroot.blocks = [ optBlock, block ];
        optBlock.adopt(nthroot, 0, 0);
        block.adopt(nthroot, optBlock, 0);
        return nthroot;
      });
    }).or(super_.parser.call(this));
  };
        _.reflow = function() {
            var firstTD = this.jQ.find("tbody.mq-sqrt:first").find("tr").children().eq(0);
            var secondTD = this.jQ.find("tbody.mq-sqrt:first").find("tr").children().eq(0);
            var block = this.ends[R].jQ;
            var newSqrt =  firstTD.find(".mq-new-sqrt:first");
            var oldSqrt = firstTD.find('.old-sqrt:first');
            var tBody = this.jQ.find("tbody.mq-sqrt:first");
            var scaleY;
            var isiPad = navigator.userAgent.match(/iPad/i);   
            newSqrt.find('.mq-new-vline').remove();
            scaleY = (block.innerHeight()) / block.css('fontSize').slice(0, -2) + 0.1 * secondTD.find("span.mq-fraction").length;
            if(scaleY < 6){
               oldSqrt.css({"display":"block"}); 
               newSqrt.css({"display":"none"});  
               if(!tBody.find(".mq-nthroot").hasClass("mq-nthroot")){
                 tBody.find('td.squareRoot').removeAttr('style'); 
               }
               if (secondTD.find("span.mq-fraction").length > 0) {
                    scale(oldSqrt, 1, scaleY);                   
               } else {                    
                   if (isIpad() || window.innerWidth === 1024) {
                        scaleY = block.innerHeight() / +block.css('fontSize').slice(0, -2) + 0.24;
                        scale(oldSqrt, 1, scaleY);
                   } else {
                        scaleY = block.innerHeight() / +block.css('fontSize').slice(0, -2) + 0.12;
                        scale(oldSqrt, 1, scaleY);
                   }                   
               }
            }
            else {
                oldSqrt.css({"display":"none"});
                newSqrt.css({"display":"block"});
                tBody.find('td.squareRoot').css({ "margin-left": 0.15+"em" });
                var newLineHeight = newSqrt.children("div:first").height();
                // don't divide by 0
                if (newLineHeight === 0) newLineHeight = 1;
                var newSqrtHeight = newSqrt.children("div.squareRoot").height();
                var count = Math.ceil((block.innerHeight()-newSqrtHeight-2*newLineHeight)/newLineHeight);
                for (var i=0; i<count; i++) {
                    newSqrt.children("div:first").before("<div class='mq-new-vline'>⎪</div>");
                }
            }
            if (scaleY > 2) {
                tBody.addClass("mq-non-leaf");
                this.jQ.find("table:first").removeAttr('style'); 
                this.jQ.find("table:first").css({"padding-top" : 7 + "px"});
                if(scaleY > 6){
                  block.css({"margin-top" : -4 + "px"});
                } else {
                  block.css({"margin-top" : -7 + "px"});
                }
            } else {
                block.css({"margin-top" : -7 + "px"});
                tBody.removeClass("mq-non-leaf");
                this.jQ.find("table:first").css({ "vertical-align": "bottom" });
                this.jQ.find("table:first").css({"padding-top" : 7 + "px"});
            }   
        };
    });

var Hat = LatexCmds.hat = P(MathCommand, function(_, super_) {
  _.ctrlSeq = '\\hat';
  _.htmlTemplate =
      '<span class="mq-non-leaf">'
    +   '<span class="mq-hat-prefix">^</span>'
    +   '<span class="mq-hat-stem">&0</span>'
    + '</span>'
  ;
  _.textTemplate = ['hat(', ')'];
});

var NthRoot = LatexCmds.nthroot = P(SquareRoot, function(_, super_) {
       _.numBlocks = function() { return 2; };
       _.bigSquareRoot = '<div class="mq-new-nthroot mq-symbol-font" style="float:left;margin:-0.05em -0.3em 0px 0px;">'+
                        '<div class="mq-vline">⎪</div>'+
                        '<div class="mq-vline">⎪</div>'+
                        '<div class="squareRoot">&#x23B7;</div>'+
                        '<div style="clear:both;"></div>'+
                    '</div>';
       _.normalSquareRoot = '<span class="mq-sqrt-prefix mq-scaled" style="font-family: MJXc-TeX-main-R,MJXc-TeX-main-Rw !important;">&radic;</span>';
       _.html = function() {          
           this.htmlTemplate =  '<span class="mq-non-leaf mq-nthroot">' +
            '<table class="mq-non-leaf mq-nthroot" style="vertical-align:bottom; padding-top: 2px;">' +
                '<tbody class="mq-nthroot" style="padding-top: 3px;">' +
                    '<tr>' +
                        '<td class="squareRoot">' +
                            '<sup class="mq-nthroot mq-non-leaf" style="vertical-align: top;">&0</sup>' +
                            '<span class="old-nthroot" style="font-family: MJXc-TeX-main-R,MJXc-TeX-main-Rw !important;">&radic;</span>' +
                        '</td>' +
                        '<td class="squareRoot">' +
                                this.bigSquareRoot +
                        '</td>' +
                        '<td class="mq-non-leaf mq-sqrt-stem squareRoot">&1</td>' +
                    '</tr>' +
                '</tbody>' +
            '</table>' +
            '</span>';
            return MathCommand.prototype.html.call(this);
        };
  _.textTemplate = ['sqrt[', '](', ')'];
  _.latex = function() {
            if (this.ends[L].latex() == "") {
                return OPEN_BRACKET_SERVER + '\\sqrt{' + this.ends[R].latex() + '}' + CLOSE_BRACKET_SERVER;
            }
            return OPEN_BRACKET_SERVER + '\\sqrt[' + this.ends[L].latex() + ']{' + this.ends[R].latex() + '}' + CLOSE_BRACKET_SERVER;
  };
        _.sLatex = function() {
            return '\\sqrt[' + this.ends[L].sLatex() + ']{' + this.ends[R].sLatex() + '}';
        };
        _.nativeLatex = function() {
            return '\\sqrt[' + this.ends[L].nativeLatex() + ']{' + this.ends[R].nativeLatex() + '}';
        };
        _.sLatexToSVG = function() {
            return '\\sqrt[' + this.ends[L].sLatexToSVG() + ']{' + this.ends[R].sLatexToSVG() + '}';
        };
        _.reflow = function() {
            var firstTD = this.jQ.find("tbody.mq-nthroot:first").find("tr:first").children().eq(0);
            var secondTD = this.jQ.find("tbody.mq-nthroot:first").find("tr:first").children().eq(1);
            var thirdTD = this.jQ.find("tbody.mq-nthroot:first").find("tr:first").children().eq(2);
            var oldSqrt = firstTD.children().eq(1);
            var newSqrt = secondTD.find(".mq-new-nthroot");     
            var rightBlock = this.ends[R].jQ;
            var leftBlock = this.ends[L].jQ;
            var tBody = this.jQ.find("tbody.mq-nthroot:first");
            var tbNthRoot = this.jQ.find("table.mq-nthroot:first");
            tbNthRoot.css("vertical-align", "bottom");
            newSqrt.css({"display":"none"});    
            newSqrt.find('.mq-new-vline').remove();
            oldSqrt.css("margin-top", "");
            leftBlock.css("margin-top","");
            rightBlock.css("margin-top","");
            oldSqrt.css("transform-origin", "");
            oldSqrt.css("transform", "");
            var scaleY;
            scaleY = (thirdTD.innerHeight()) / thirdTD.css('fontSize').slice(0, -2) + 0.1 * thirdTD.find("span.mq-fraction").length;
            if(scaleY < 6){
                    oldSqrt.css({"display":"inline-block"});
                    oldSqrt.css("transform-origin", "left bottom");  
                    newSqrt.css({"display":"none"}); 
                    if (secondTD.find("span.mq-fraction").length > 0) {
                        scale(oldSqrt, 1, scaleY);                   
                    } else {                    
                    if (isIpad() || window.innerWidth === 1024) {
                            scaleY = thirdTD.innerHeight() / +thirdTD.css('fontSize').slice(0, -2) + 0.24;
                            scale(oldSqrt, 1, scaleY);
                    } else {
                            scaleY = thirdTD.innerHeight() / +thirdTD.css('fontSize').slice(0, -2) + 0.12;
                            scale(oldSqrt, 1, scaleY);
                    }                   
                }
                for (var i = 0; i < 3; i++) {
                    if (oldSqrt[0].getBoundingClientRect().top < rightBlock[0].getBoundingClientRect().top) {
                        var oldMarginTop = +oldSqrt.css('margin-top').slice(0, -2);
                        var newMarginTop = oldMarginTop + (rightBlock[0].getBoundingClientRect().top - oldSqrt[0].getBoundingClientRect().top);
                        if (newMarginTop > 0) {
                            oldSqrt.css("margin-top", newMarginTop + "px");
                        }
                    } else {
                        break;
                    }
                }
                var middleBottomTbody = oldSqrt[0].getBoundingClientRect().bottom - (oldSqrt[0].getBoundingClientRect().height/2);
                for (var i = 0; i < 3; i++) {
                    if (leftBlock[0].getBoundingClientRect().bottom < middleBottomTbody) {
                        var oldMarginTop = +leftBlock.css('margin-top').slice(0, -2);
                        var newMarginTop = oldMarginTop + (middleBottomTbody - leftBlock[0].getBoundingClientRect().bottom);
                        if (newMarginTop > 0) {
                            leftBlock.css("margin-top", newMarginTop + "px");
                        }
                    } else {
                        break;
                    }
                }
                //align sqrtBlock to the left block
                for (var i = 0; i < 3; i++) {
                    if (leftBlock[0].getBoundingClientRect().bottom >= oldSqrt[0].getBoundingClientRect().bottom) {
                        var oldMarginTop = +oldSqrt.css('margin-top').slice(0, -2);
                        var newMarginTop = oldMarginTop + (leftBlock[0].getBoundingClientRect().height - oldSqrt[0].getBoundingClientRect().height);
                        if (newMarginTop > 0) {
                            oldSqrt.css("margin-top", newMarginTop + "px");
                        }
                    } else {
                        break;
                    }
                } 
                for (var i = 0; i < 3; i++) {
                    if (rightBlock[0].getBoundingClientRect().top < oldSqrt[0].getBoundingClientRect().top) {
                        var oldMarginTop = +rightBlock.css('margin-top').slice(0, -2);
                        var newMarginTop = oldMarginTop + (oldSqrt[0].getBoundingClientRect().top - rightBlock[0].getBoundingClientRect().top);
                        if (newMarginTop > 0) {
                            rightBlock.css("margin-top", newMarginTop + "px");
                        }
                    } else {
                        break;
                    }
                }                
            } else {
                    oldSqrt.css({"display":"none"});
                    newSqrt.css({"display":"inline-block"});
                    var newLineHeight = newSqrt.children("div:first").height();
                    // don't divide by 0
                    if (newLineHeight === 0) newLineHeight = 1;
                    var newSqrtHeight = newSqrt.children("div.squareRoot").height();
                    var count = Math.ceil((tBody.innerHeight()-newSqrtHeight-2*newLineHeight)/newLineHeight);
                    for (var i=0; i<count; i++) {
                        newSqrt.children("div:first").before("<div class='mq-new-vline'>⎪</div>");
                    } 
                    for (var i = 0; i < 3; i++) {
                        if ((newSqrt.find(".squareRoot")[0].getBoundingClientRect().top - leftBlock[0].getBoundingClientRect().bottom) > newLineHeight) {
                            var oldMarginTop = +leftBlock.css('margin-top').slice(0, -2);
                            var newMarginTop = oldMarginTop + (newSqrt.find(".squareRoot")[0].getBoundingClientRect().top - leftBlock[0].getBoundingClientRect().bottom);
                            if (newMarginTop > 0) {
                                leftBlock.css("margin-top", newMarginTop + "px");
                            }
                        } else {
                            break;
                        }
                    }
            }
            if (scaleY >= 2) {
                tbNthRoot.css("vertical-align", "middle");
            }
            if (this.jQ.parent("sup.mq-nthroot").length === 0 && leftBlock.find("table.mq-nthroot").length > 0) {
                scale(leftBlock, 0.95, 0.95);
            }
        };
  _.mathspeak = function() {
    var indexMathspeak = this.ends[L].mathspeak();
    var radicandMathspeak = this.ends[R].mathspeak();
    this.ends[L].ariaLabel = 'Index';
    this.ends[R].ariaLabel = 'Radicand';
    if (indexMathspeak === '3') { // cube root
      return 'Start Cube Root, '+radicandMathspeak+', End Cube Root';
    } else {
      return 'Root Index '+indexMathspeak+', Start Root, '+radicandMathspeak+', End Root';
    }
  };
});

var CubeRoot =
LatexCmds.cbrt = P(NthRoot, function(_, super_) {
  _.createLeftOf = function(cursor) {
    super_.createLeftOf.apply(this, arguments);
    Digit('3').createLeftOf(cursor);
    cursor.controller.moveRight();
  };
});

var DiacriticAbove = P(MathCommand, function(_, super_) {
  _.init = function(ctrlSeq, symbol, textTemplate) {
    var htmlTemplate =
            '<span class="mq-non-leaf">' +
            '<span class="mq-diacritic-above">' + symbol + '</span>' +
            '<span class="mq-diacritic-stem">&0</span>' +
            '</span>';

    super_.init.call(this, ctrlSeq, htmlTemplate, textTemplate);
  };
});
// NNANH 12/11/2018 Add - S - Bug#9383 Change font arrow to casio font
LatexCmds.vec = bind(DiacriticAbove, '\\vec', '\uEE0F', ['vec(', ')']);
// NNANH 12/11/2018 Add - S - Bug#9383 Change font arrow to casio font
LatexCmds.tilde = bind(DiacriticAbove, '\\tilde', '~', ['tilde(', ')']);

function DelimsMixin(_, super_) {
     _.bigLeftCurlBracket = '<div style="float:left;display:inline-block;font-weight:bold;" class="mq-bracket-solve mq-symbol-font">'+
                '<div class="mq-curl-head">⎧</div>'+
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div style="text-align: center;" class="mq-curl">⎨</div>'+
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div class="mq-curl-tail">⎩</div>'+
                '<div style="clear:both;"></div></div>';
    _.bigRightCurlBracket = '<div style="float:left;display:inline-block;font-weight:bold;" class="mq-bracket-solve mq-symbol-font">'+
                '<div class="mq-curl-head">⎫</div>'+ 
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div style="text-align: center;" class="mq-curl">⎬</div>'+
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div class="mq-curl-tail">⎭</div>'+
                '<div style="clear:both;"></div></div>';
    _.bigLeftRoundBracket = '<div style="float:left;display:inline-block;font-weight:bold;" class="mq-bracket-round mq-round-left mq-symbol-font">'+
                '<div class="mq-round-head">⎛</div>'+
                '<div style="text-align: center;" class="mq-vline">&#x239C;</div>'+
                '<div class="mq-round-tail">⎝</div>'+
                '<div style="clear:both;"></div></div>';
    _.bigRightRoundBracket = '<div style="float:left;display:inline-block;font-weight:bold;"  class="mq-bracket-round mq-round-right mq-symbol-font">'+
                '<div class="mq-round-head">⎞</div>'+
                '<div style="text-align: center;" class="mq-vline">&#x239F;</div>'+
                '<div class="mq-round-tail">⎠</div>'+
                '<div style="clear:both;"></div></div>';
  _.jQadd = function() {
    super_.jQadd.apply(this, arguments);
        if (this.ctrlSeq === "\\left[" || this.ctrlSeq === "\\left|") {
             var squareBracketHtml = this.jQ.find(".mq-square-bracket:first");
             this.delimjQs = squareBracketHtml.children(':first').add(squareBracketHtml.children(':last'));
             this.contentjQ = squareBracketHtml.children(':eq(1)');
        } else {
    this.delimjQs = this.jQ.children(':first').add(this.jQ.children(':last'));
    this.contentjQ = this.jQ.children(':eq(1)');
        }       
    };
    _.reflow = function() {
        var firstTD = this.jQ.children(':first');
        var secondTD = this.jQ.children(':last'); 
        // NNANH start 10/10/2018 padding top with bracket table. BUG#8715
        if (this.ctrlSeq === "\\left[" || this.ctrlSeq === "\\left|") {
            if(firstTD[0].getBoundingClientRect().height > 120){
                this.contentjQ.css({"padding-top": 5 + "px"});
            } else if(firstTD[0].getBoundingClientRect().height > 220){
                this.contentjQ.css({"padding-top": 10 + "px"});
            } else if(firstTD[0].getBoundingClientRect().height > 300){
                this.contentjQ.css({"padding-top": 15 + "px"});
            } else this.contentjQ.css({"padding-top": 0 + "px"}); 
            // NNANH end 10/10/2018 padding top with bracket table. BUG#8715
            return;
        }
        var scaleY = this.contentjQ.outerHeight() / parseFloat(this.contentjQ.css('fontSize'));
        if (this.ctrlSeq !== "\\left\\{" && this.ctrlSeq !== "\\left(") {
            scale(this.delimjQs, min(1 + .2 * (scaleY - 1), 1.2), 1.2 * scaleY);
            return;
        }         
        if (scaleY < 6) {
            firstTD.html(this.sides[L].ch);
            secondTD.html(this.sides[R].ch);
            if(scaleY < 1.2){
                scale(this.delimjQs, min(1 + .2 * (scaleY - 1), 1.2), max(scaleY, 1.2));
            } else {
                if(this.contentjQ.find(".mq-nthroot").hasClass("mq-nthroot")){
                    scale(this.delimjQs, min(1 + .2 * (scaleY - 1), 1.2), max(scaleY, 1.3) + 0.5);
                } else {
                    scale(this.delimjQs, min(1 + .2 * (scaleY - 1), 1.2), max(scaleY, 1.3) + 0.25);
                }
            }
            this.delimjQs.css({"transform-origin": "top left"});
            this.contentjQ.css("margin-top", "");
            if (isIpad() || window.innerWidth === 1024){
                var oldMarginTop = +firstTD.css("margin-top").slice(0,-2);
                var newMarginTop = oldMarginTop + (this.contentjQ[0].getBoundingClientRect().bottom - firstTD[0].getBoundingClientRect().bottom);
                this.contentjQ.css({"margin-top":  - newMarginTop + "px"});
            } else {
                for (var i = 0; i<2; i++) {
                if (firstTD[0].getBoundingClientRect().bottom < this.contentjQ[0].getBoundingClientRect().bottom) {
                    var oldMarginTop = +firstTD.css("margin-top").slice(0,-2);
                    var newMarginTop = oldMarginTop + (this.contentjQ[0].getBoundingClientRect().bottom - firstTD[0].getBoundingClientRect().bottom);
                    this.contentjQ.css({"margin-top": - newMarginTop + 4 + "px"});
                } else {
                    if(scaleY < 2){
                        break;
                    }
                    var oldMarginTop = +firstTD.css("margin-top").slice(0,-2);
                    var newMarginTop = oldMarginTop + (this.contentjQ[0].getBoundingClientRect().bottom - firstTD[0].getBoundingClientRect().bottom);
                    this.contentjQ.css({"margin-top":  - newMarginTop + 2.5 + "px"});
                    break;
                }
            }  
            }        
            return;
        }     
        if (this.ctrlSeq === "\\left\\{") {
            firstTD.html(this.bigLeftCurlBracket);
            secondTD.html(this.bigRightCurlBracket);
            this.delimjQs.removeAttr("style");           
        } else if (this.ctrlSeq === "\\left(") {
            firstTD.html(this.bigLeftRoundBracket);
            secondTD.html(this.bigRightRoundBracket);
            this.delimjQs.removeAttr("style");
        }  
        var blockHeight = this.contentjQ[0].getBoundingClientRect().height;
        var currentHeightBigBracket = firstTD[0].getBoundingClientRect().height;
        var lineBracketHeight = firstTD.find(".mq-vline").height(); 
        if (Math.round(firstTD[0].getBoundingClientRect().bottom) < Math.round(this.contentjQ[0].getBoundingClientRect().bottom)) {            
            if (this.ctrlSeq === "\\left\\{") {
                var adjustLineBracket = Math.ceil(Math.abs(blockHeight - currentHeightBigBracket)/(2*lineBracketHeight)) + 1;
                for (var i = 0; i < adjustLineBracket + 2; i++) {
                    $('<div style="text-align: center;" class="mq-vline">⎪</div>').insertAfter(firstTD.find(".mq-curl"));
                    $('<div style="text-align: center;" class="mq-vline">⎪</div>').insertAfter(secondTD.find(".mq-curl"));
                    $('<div style="text-align: center;" class="mq-vline">⎪</div>').insertBefore(firstTD.find(".mq-curl"));
                    $('<div style="text-align: center;" class="mq-vline">⎪</div>').insertBefore(secondTD.find(".mq-curl"));                 
                    if (Math.round(firstTD[0].getBoundingClientRect().bottom) >= Math.round(this.contentjQ[0].getBoundingClientRect().bottom)) {
                        this.contentjQ.css("margin-top", "");
                        var oldMarginTop = +firstTD.css("margin-top").slice(0,-2);
                        var newMarginTop = oldMarginTop + (firstTD[0].getBoundingClientRect().bottom - this.contentjQ[0].getBoundingClientRect().bottom);
                        this.contentjQ.css({"margin-top":  newMarginTop/2 + 3 + "px"});
                        break;
                    }
                }                
            } else if (this.ctrlSeq === "\\left(") {
                var adjustLineBracket = Math.ceil(Math.abs(blockHeight - currentHeightBigBracket)/(lineBracketHeight));
                for (var i = 0; i < adjustLineBracket + 2; i++) {
                    $('<div style="text-align: center;" class="mq-vline">&#x239C;</div>').insertBefore(firstTD.find(".mq-round-tail"));  
                    $('<div style="text-align: center;" class="mq-vline">&#x239F;</div>').insertBefore(secondTD.find(".mq-round-tail"));
                    if (Math.round(firstTD[0].getBoundingClientRect().bottom) >= Math.round(this.contentjQ[0].getBoundingClientRect().bottom)) {
                        this.contentjQ.css("margin-top", "");
                        var oldMarginTop = +firstTD.css("margin-top").slice(0,-2);
                        var newMarginTop = oldMarginTop + (firstTD[0].getBoundingClientRect().bottom - this.contentjQ[0].getBoundingClientRect().bottom);
                        this.contentjQ.css({"margin-top":   newMarginTop/2 + 3 + "px"});
                        break;
                    }
                }               
            }
        } 
        if (this.ctrlSeq === "\\left\\{") {
            firstTD.find(".mq-curl").prev().addClass("mq-vline-pre");
            secondTD.find(".mq-curl").prev().addClass("mq-vline-pre");  
        } else if (this.ctrlSeq === "\\left("){
            firstTD.find(".mq-vline:last").addClass("mq-vline-pre");
            secondTD.find(".mq-vline:last").addClass("mq-vline-pre");
        }
  };
}

// Round/Square/Curly/Angle Brackets (aka Parens/Brackets/Braces)
//   first typed as one-sided bracket with matching "ghost" bracket at
//   far end of current block, until you type an opposing one
var Bracket = P(P(MathCommand, DelimsMixin), function(_, super_) {
  _.init = function(side, open, close, ctrlSeq, end) {
        super_.init.call(this, '\\left' + ctrlSeq, undefined, [open, close]);
    this.side = side;
    this.sides = {};
    this.sides[L] = { ch: open, ctrlSeq: ctrlSeq };
    this.sides[R] = { ch: close, ctrlSeq: end };
  };
  _.numBlocks = function() { return 1; };
    _.html = function() { // wait until now so that .side may
        if (this.ctrlSeq === "\\left[") {
            this.htmlTemplate =  '<span class="mq-matrix mq-non-leaf">' +
                 '<div style="display: table;">' +                    
                    '<div style="display: table-row;" class="mq-square-bracket">'+
                        '<div class="mq-paren' + (this.side === R ? ' mq-ghost' : '') + '" style="display: table-cell;border-top: 2px solid;border-left: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                        '<div style="display: table-cell;"><div style="margin-left: 0.15em;margin-right: 0.15em;">&0</div></div>'+
                        '<div class="mq-paren' + (this.side === L ? ' mq-ghost' : '') + '" style="display: table-cell;border-top: 2px solid;border-right: 2px solid;border-bottom: 2px solid;width: 0.3em;"></div>'+
                    '</div>'+                   
                '</div>'+
                '</span>';
        } else if (this.ctrlSeq === "\\left|") {
            this.htmlTemplate =  '<span class="mq-matrix mq-non-leaf">' +
                 '<div style="display: table;">' +                    
                    '<div style="display: table-row;" class="mq-square-bracket">'+
                        '<div class="mq-paren' + (this.side === R ? ' mq-ghost' : '') + '" style="display: table-cell;border-left: 2px solid;width: 0.3em;"></div>'+
                        '<div style="display: table-cell;">&0</div>'+
                        '<div class="mq-paren' + (this.side === L ? ' mq-ghost' : '') + '" style="display: table-cell;border-right: 2px solid;width: 0.3em;"></div>'+
                    '</div>'+                   
                '</div>'+
                '</span>';
        } else {
            this.htmlTemplate = // be set by createLeftOf or parser
                '<span class="mq-non-leaf">' +
                '<span class="mq-scaled mq-paren' + (this.side === R ? ' mq-ghost' : '') + '">' +
                this.sides[L].ch +
                '</span>' +
                '<span class="mq-non-leaf">&0</span>' +
                '<span class="mq-scaled mq-paren' + (this.side === L ? ' mq-ghost' : '') + '">' +
                this.sides[R].ch +
                '</span>' +
                '</span>';
        }
        return super_.html.call(this);
    };
    _.latex = function() {
        if (this.ctrlSeq === '\\left"') {
            return '"' + this.ends[L].latex() + '"';
        }
        return '\\left' + this.sides[L].ctrlSeq + this.ends[L].latex() + '\\right' + this.sides[R].ctrlSeq;

  };
    _.sLatex = function() {        
        return '\\left' + this.sides[L].ctrlSeq + '{' + this.ends[L].sLatex() + '}\\right' + this.sides[R].ctrlSeq;
  };
    _.nativeLatex = function() {
        if (this.ctrlSeq === '\\left"') {
            return '"{' + this.ends[L].nativeLatex() + '}"';
        }
        return '\\left' + this.sides[L].ctrlSeq + '{' + this.ends[L].nativeLatex() + '}\\right' + this.sides[R].ctrlSeq;
  };
    _.sLatexToSVG = function() {
        if (this.ctrlSeq === '\\left"') {
            return '"{' + this.ends[L].sLatexToSVG() + '}"';
        }
        return '\\left' + this.sides[L].ctrlSeq + '{' + this.ends[L].sLatexToSVG() + '}\\right' + this.sides[R].ctrlSeq;
    };
  _.mathspeak = function(opts) {
    var open = this.sides[L].ch, close = this.sides[R].ch;
    if (open === '|' && close === '|') {
      this.mathspeakTemplate = ['StartAbsoluteValue,', ', EndAbsoluteValue'];
      this.ariaLabel = 'absolute value';
    }
    else if (opts && opts.createdLeftOf && this.side) {
      var ch = '';
      if (this.side === L) ch = this.textTemplate[0];
      else if (this.side === R) ch = this.textTemplate[1];
      return (this.side === L ? 'left ' : 'right ') + BRACKET_NAMES[ch];
    }
    else {
      this.mathspeakTemplate = ['left ' + BRACKET_NAMES[open]+',', ', right ' + BRACKET_NAMES[close]];
      this.ariaLabel = BRACKET_NAMES[open]+' block';
    }
    return super_.mathspeak.call(this);
  };
  _.matchBrack = function(opts, expectedSide, node) {
    // return node iff it's a matching 1-sided bracket of expected side (if any)
        return node instanceof Bracket &&
            node.side &&
            node.side !== -expectedSide &&
            (!opts.restrictMismatchedBrackets ||
                OPP_BRACKS[this.sides[this.side].ch] === node.sides[node.side].ch || { '(': ']', '[': ')' }[this.sides[L].ch] === node.sides[R].ch) && node;
  };
  _.closeOpposing = function(brack) {
    brack.side = 0;
    brack.sides[this.side] = this.sides[this.side]; // copy over my info (may be
        if (this.ctrlSeq === "\\left[" || this.ctrlSeq === "\\left\\{" || this.ctrlSeq === "\\left(" || this.ctrlSeq === "\\left|") {
           brack.delimjQs.eq(this.side === L ? 0 : 1).removeClass('mq-ghost');
        } else {
           brack.delimjQs.eq(this.side === L ? 0 : 1).removeClass('mq-ghost').html(this.sides[this.side].ch);
        }
        
        /**
         * @author ThuyLTM
         * Remove bold letter when closing )
         */
        if (this.side === 1) {
            var lastObj = brack.ends[L].ends[1];
            if ((lastObj.digit || lastObj instanceof Letter || lastObj.variable) && lastObj.jQ.is(".bold")) {
                lastObj._clearLeftBold(true);
            }
            if (lastObj instanceof Subscript && !lastObj.isEmpty() && lastObj.getFirstRightSubscript().jQ.is(".bold")) {
                lastObj.getFirstRightSubscript()._clearLeftBold(true);
            }
        }
  };
  _.createLeftOf = function(cursor) {
    if (!this.replacedFragment) { // unless wrapping seln in brackets,
        // check if next to or inside an opposing one-sided bracket
      var opts = cursor.options;
            if (this.sides[L].ch === '|' || this.sides[L].ch === '"') { // check both sides if I'm a pipe
                var brack = this.matchBrack(opts, R, cursor[R]) ||
                    this.matchBrack(opts, L, cursor[L]) ||
                    this.matchBrack(opts, 0, cursor.parent.parent);
            } else {
                var brack = this.matchBrack(opts, -this.side, cursor[-this.side]) ||
                    this.matchBrack(opts, -this.side, cursor.parent.parent);
      }
    }
    if (brack) {
      var side = this.side = -brack.side; // may be pipe with .side not yet set
      this.closeOpposing(brack);
      if (brack === cursor.parent.parent && cursor[side]) { // move the stuff between
        Fragment(cursor[side], cursor.parent.ends[side], -side) // me and ghost outside
          .disown().withDirAdopt(-side, brack.parent, brack, brack[side])
          .jQ.insDirOf(side, brack.jQ);
      }
      if (side === L) cursor.insAtLeftEnd(brack.ends[L]);
      else cursor.insRightOf(brack);      
      brack.bubble(function (node) { node.reflow(); });
      return;
    }
    else {
      brack = this, side = brack.side;
      if (brack.replacedFragment) brack.side = 0; // wrapping seln, don't be one-sided
      else if (cursor[-side]) { // elsewise, auto-expand so ghost is at far end
        brack.replaces(Fragment(cursor[-side], cursor.parent.ends[-side], side));
        cursor[-side] = 0;
      }
      super_.createLeftOf.call(brack, cursor);
    }
    if (side === L) cursor.insAtLeftEnd(brack.ends[L]);
    else cursor.insRightOf(brack);
        if (this instanceof Bracket && this.ctrlSeq === "\\left(" && this.side === 1) {
            var lastObj = this.ends[L].ends[1];
            if ((lastObj.digit || lastObj instanceof Letter || lastObj.variable) && lastObj.jQ.is(".bold")) {
                lastObj._clearLeftBold(true);
            }
            if (lastObj instanceof Subscript && !lastObj.isEmpty() && lastObj.getFirstRightSubscript().jQ.is(".bold")) {
                lastObj.getFirstRightSubscript()._clearLeftBold(true);
            }
        }
  };
  _.placeCursor = noop;
  _.unwrap = function() {
    this.ends[L].children().disown().adopt(this.parent, this, this[R])
      .jQ.insertAfter(this.jQ);
    this.remove();
  };
  _.deleteSide = function(side, outward, cursor) {
        var parent = this.parent,
            sib = this[side],
            farEnd = parent.ends[side];

    if (side === this.side) { // deleting non-ghost of one-sided bracket, unwrap
      this.unwrap();
      sib ? cursor.insDirOf(-side, sib) : cursor.insAtDirEnd(side, parent);
      return;
    }

        var opts = cursor.options,
            wasSolid = !this.side;
    this.side = -side;
    // if deleting like, outer close-brace of [(1+2)+3} where inner open-paren
    if (this.matchBrack(opts, side, this.ends[L].ends[this.side])) { // is ghost,
      this.closeOpposing(this.ends[L].ends[this.side]); // then become [1+2)+3
      var origEnd = this.ends[L].ends[side];
      this.unwrap();
      if (origEnd.siblingCreated) origEnd.siblingCreated(cursor.options, side);
      sib ? cursor.insDirOf(-side, sib) : cursor.insAtDirEnd(side, parent);
        } else { // if deleting like, inner close-brace of ([1+2}+3) where outer
      if (this.matchBrack(opts, side, this.parent.parent)) { // open-paren is
        this.parent.parent.closeOpposing(this); // ghost, then become [1+2+3)
        this.parent.parent.unwrap();
      } // else if deleting outward from a solid pair, unwrap
      else if (outward && wasSolid) {
        this.unwrap();
        sib ? cursor.insDirOf(-side, sib) : cursor.insAtDirEnd(side, parent);
        return;
            } else { // else deleting just one of a pair of brackets, become one-sided
                this.sides[side] = {
                    ch: OPP_BRACKS[this.sides[this.side].ch],
                    ctrlSeq: OPP_BRACKS[this.sides[this.side].ctrlSeq]
                };
                if (this.ctrlSeq === "\\left[" || this.ctrlSeq === "\\left\\{" || this.ctrlSeq === "\\left(" || this.ctrlSeq === "\\left|") {
                   this.delimjQs.removeClass('mq-ghost').eq(side === L ? 0 : 1).addClass('mq-ghost');
                } else {
                   this.delimjQs.removeClass('mq-ghost').eq(side === L ? 0 : 1).addClass('mq-ghost').html(this.sides[side].ch);
      }
                
      }
      if (sib) { // auto-expand so ghost is at far end
        var origEnd = this.ends[L].ends[side];
        Fragment(sib, farEnd, -side).disown()
          .withDirAdopt(-side, this.ends[L], origEnd, 0)
          .jQ.insAtDirEnd(side, this.ends[L].jQ.removeClass('mq-empty'));
        if (origEnd.siblingCreated) origEnd.siblingCreated(cursor.options, side);
        cursor.insDirOf(-side, sib);
      } // didn't auto-expand, cursor goes just outside or just inside parens
      else (outward ? cursor.insDirOf(side, this)
                    : cursor.insAtDirEnd(side, this.ends[L]));
    }
  };
  _.replaceBracket = function ($brack, side) {
    var symbol = this.getSymbol(side);
    $brack.html(symbol.html).css('width', symbol.width);

    if (side === L) {
      $brack.next().css('margin-left', symbol.width);
    } else {
      $brack.prev().css('margin-right', symbol.width);
    }
  };
  _.deleteTowards = function(dir, cursor) {
    this.deleteSide(-dir, false, cursor);
  };
  _.finalizeTree = function() {
    this.ends[L].deleteOutOf = function(dir, cursor) {
      this.parent.deleteSide(dir, true, cursor);
    };
    // FIXME HACK: after initial creation/insertion, finalizeTree would only be
    // called if the paren is selected and replaced, e.g. by LiveFraction
    this.finalizeTree = this.intentionalBlur = function() {
      this.delimjQs.eq(this.side === L ? 1 : 0).removeClass('mq-ghost');
      this.side = 0;
    };
  };
  _.siblingCreated = function(opts, dir) { // if something typed between ghost and far
    if (dir === -this.side) this.finalizeTree(); // end of its block, solidify
  };
});

var OPP_BRACKS = {
  '(': ')',
  ')': '(',
  '[': ']',
  ']': '[',
  '{': '}',
  '}': '{',
  '\\{': '\\}',
  '\\}': '\\{',
  '&lang;': '&rang;',
  '&rang;': '&lang;',
  '\\langle ': '\\rangle ',
  '\\rangle ': '\\langle ',
  '|': '|',
  '\\lVert ' : '\\rVert ',
  '\\rVert ' : '\\lVert ',
};

var BRACKET_NAMES = {
  '&lang;': 'angle-bracket',
  '&rang;': 'angle-bracket',
  '|': 'pipe'
};

function bindCharBracketPair(open, ctrlSeq, name) {
  var ctrlSeq = ctrlSeq || open, close = OPP_BRACKS[open], end = OPP_BRACKS[ctrlSeq];
  CharCmds[open] = bind(Bracket, L, open, close, ctrlSeq, end);
  CharCmds[close] = bind(Bracket, R, open, close, ctrlSeq, end);
  BRACKET_NAMES[open] = BRACKET_NAMES[close] = name;
}
bindCharBracketPair('(', null, 'parenthesis');
bindCharBracketPair('[', null, 'bracket');
bindCharBracketPair('{', '\\{', 'brace');
LatexCmds.langle = bind(Bracket, L, '&lang;', '&rang;', '\\langle ', '\\rangle ');
LatexCmds.rangle = bind(Bracket, R, '&lang;', '&rang;', '\\langle ', '\\rangle ');
/* comment the mathquill default
CharCmds['|'] = bind(Bracket, L, '|', '|', '|', '|');
| represents command 'or' */
LatexCmds.vert = bind(Bracket, L, '|', '|', '|', '|');
LatexCmds.lVert = bind(Bracket, L, '&#8741;', '&#8741;', '\\lVert ', '\\rVert ');
LatexCmds.rVert = bind(Bracket, R, '&#8741;', '&#8741;', '\\lVert ', '\\rVert ');


LatexCmds.left = P(MathCommand, function(_) {
  _.parser = function() {
    var regex = Parser.regex;
    var string = Parser.string;
    var succeed = Parser.succeed;
    var optWhitespace = Parser.optWhitespace;

        return optWhitespace.then(regex(/^(?:["([|]|\\\{|\\langle\b|\\lVert\b)/))
      .then(function(ctrlSeq) {
                var open = (ctrlSeq.charAt(0) === '\\' ? ctrlSeq.slice(1) : ctrlSeq);
                if (ctrlSeq == "\\langle") {
                    open = '&lang;';
                    ctrlSeq = ctrlSeq + ' ';
                }
                if (ctrlSeq == "\\lVert") {
                    open = '&#8741;';
                    ctrlSeq = ctrlSeq + ' ';
                }
                return optWhitespace.then((latexMathParser.then(function(block) {
          return string('\\right').skip(optWhitespace)
            .then(regex(/^(?:[\])|]|\\\}|\\rangle\b|\\rVert\b)/)).map(function(end) {
              var close = end.replace(/^\\/, '');
	      if (end=="\\rangle") { close = '&rang;'; end = end + ' '; }
	      if (end=="\\rVert") { close = '&#8741;'; end = end + ' '; }
              var cmd = Bracket(0, open, close, ctrlSeq, end);
              cmd.blocks = [ block ];
              block.adopt(cmd, 0, 0);
              return cmd;
        });
                })).or((latexMathParser.then(function(block){
                    var close = OPP_BRACKS[open];
                    var cmd = Bracket(0, open, close, ctrlSeq, close);
                    cmd.blocks = [block];
                    block.adopt(cmd, 0, 0);
                    return succeed(cmd);
                }))))
            });
  };
});

LatexCmds.right = P(MathCommand, function(_) {
  _.parser = function() {
    return Parser.fail('unmatched \\right');
  };
});

var Binomial =
    LatexCmds.binom =
    LatexCmds.binomial = P(P(MathCommand, DelimsMixin), function(_, super_) {
  _.ctrlSeq = '\\binom';
  _.htmlTemplate =
            '<span class="mq-non-leaf">' +
            '<span class="mq-paren mq-scaled">(</span>' +
            '<span class="mq-non-leaf">' +
            '<span class="mq-array mq-non-leaf">' +
            '<span>&0</span>' +
            '<span>&1</span>' +
            '</span>' +
            '</span>' +
            '<span class="mq-paren mq-scaled">)</span>' +
            '</span>';
        _.textTemplate = ['choose(', ',', ')'];
  _.mathspeakTemplate = ['StartBinomial,', 'Choose', ', EndBinomial'];
  _.ariaLabel = 'binomial';
});

var Choose =
LatexCmds.choose = P(Binomial, function(_) {
  _.createLeftOf = LiveFraction.prototype.createLeftOf;
});

LatexCmds.editable = // backcompat with before cfd3620 on #233
LatexCmds.MathQuillMathField = P(MathCommand, function(_, super_) {
  _.ctrlSeq = '\\MathQuillMathField';
  _.htmlTemplate =
      '<span class="mq-editable-field">'
    +   '<span class="mq-root-block">&0</span>'
    + '</span>'
  ;
  _.parser = function() {
    var self = this,
      string = Parser.string, regex = Parser.regex, succeed = Parser.succeed;
    return string('[').then(regex(/^[a-z][a-z0-9]*/i)).skip(string(']'))
      .map(function(name) { self.name = name; }).or(succeed())
      .then(super_.parser.call(self));
  };
  _.finalizeTree = function(options) {
    var ctrlr = Controller(this.ends[L], this.jQ, options);
    ctrlr.KIND_OF_MQ = 'MathField';
    ctrlr.editable = true;
    ctrlr.createTextarea();
    ctrlr.editablesTextareaEvents();
    ctrlr.cursor.insAtRightEnd(ctrlr.root);
    RootBlockMixin(ctrlr.root);
  };
  _.registerInnerField = function(innerFields, MathField) {
    innerFields.push(innerFields[this.name] = MathField(this.ends[L].controller));
  };
  _.latex = function(){ return this.ends[L].latex(); };
  _.text = function(){ return this.ends[L].text(); };
});

// Embed arbitrary things
// Probably the closest DOM analogue would be an iframe?
// From MathQuill's perspective, it's a Symbol, it can be
// anywhere and the cursor can go around it but never in it.
// Create by calling public API method .dropEmbedded(),
// or by calling the global public API method .registerEmbed()
// and rendering LaTeX like \embed{registeredName} (see test).
var Embed = LatexCmds.embed = P(Symbol, function(_, super_) {
  _.setOptions = function(options) {
    function noop () { return ""; }
    this.text = options.text || noop;
    this.htmlTemplate = options.htmlString || "";
    this.latex = options.latex || noop;
    return this;
  };
  _.parser = function() {
    var self = this,
            string = Parser.string,
            regex = Parser.regex,
            succeed = Parser.succeed;
    return string('{').then(regex(/^[a-z][a-z0-9]*/i)).skip(string('}'))
      .then(function(name) {
        // the chars allowed in the optional data block are arbitrary other than
        // excluding curly braces and square brackets (which'd be too confusing)
        return string('[').then(regex(/^[-\w\s]*/)).skip(string(']'))
          .or(succeed()).map(function(data) {
            return self.setOptions(EMBEDS[name](data));
                    });
            });
    };
});

// LaTeX environments
// Environments are delimited by an opening \begin{} and a closing
// \end{}. Everything inside those tags will be formatted in a
// special manner depending on the environment type.
var Environments = {};

LatexCmds.begin = P(MathCommand, function(_, super_) {
    _.parser = function() {
        var string = Parser.string;
        var regex = Parser.regex;
        return string('{')
            .then(regex(/^[a-z]+/i))
            .skip(string('}'))
            .then(function(env) {
                return (Environments[env] ?
                    Environments[env]().parser() :
                    Parser.fail('unknown environment type: ' + env)
                ).skip(string('\\end{' + env + '}'));
            });
    };
});

var Environment = P(MathCommand, function(_, super_) {
    _.template = [
        ['\\begin{', '}'],
        ['\\end{', '}']
    ];
    _.wrappers = function() {
        return [
            _.template[0].join(this.environment),
            _.template[1].join(this.environment)
        ];
    };
});

var Matrix =
    Environments.matrix = P(Environment, function(_, super_) {

        var delimiters = {
            column: '&',
            row: '\\\\'
        };
        _.parentheses = {
            left: null,
            right: null
        };
        _.environment = 'matrix';
       
        _.latex = function() {
            var latex = '';
            var row;

            this.eachChild(function(cell) {
                if (typeof row !== 'undefined') {
                    latex += (row !== cell.row) ?
                        delimiters.row :
                        delimiters.column;
                }
                row = cell.row;
                latex += cell.latex();
            });

            return this.wrappers().join(latex);
        };
        _.sLatex = function() {
            var latex = '';
            var row;

            this.eachChild(function(cell) {
                if (typeof row !== 'undefined') {
                    latex += (row !== cell.row) ?
                        delimiters.row :
                        delimiters.column;
                }
                row = cell.row;
                latex += cell.sLatex();
            });

            return this.wrappers().join(latex);
        };
        _.nativeLatex = function() {
            var latex = '';
            var row;

            this.eachChild(function(cell) {
                if (typeof row !== 'undefined') {
                    latex += (row !== cell.row) ?
                        delimiters.row :
                        delimiters.column;
                }
                row = cell.row;
                latex += cell.nativeLatex();
            });

            return this.wrappers().join(latex);
        };
        _.sLatexToSVG = function() {
            var latex = '';
            var row;

            this.eachChild(function(cell) {
                if (typeof row !== 'undefined') {
                    latex += (row !== cell.row) ?
                        delimiters.row :
                        delimiters.column;
                }
                row = cell.row;
                latex += cell.sLatexToSVG();
            });

            return this.wrappers().join(latex);
        };
        _.html = function() {
            var cells = [],
                trs = '',
                i = 0,
                row;

            function parenHtml(paren) {
                return (paren) ?
                    '<span class="mq-scaled mq-paren">' +
                    paren +
                    '</span>' : '';
            }

            // Build <tr><td>.. structure from cells
            this.eachChild(function(cell) {
                if (row !== cell.row) {
                    row = cell.row;
                    trs += '<tr>$tds</tr>';
                    cells[row] = [];
                }
                cells[row].push('<td class="mq-td">&' + (i++) + '</td>');
            });
           
            //this.htmlTemplate =
            var dataMatrix = 
                '<span class="mq-matrix mq-non-leaf">' +
                '<table class="mq-non-leaf mq-matrix-content">' +
                trs.replace(/\$tds/g, function() {
                    return cells.shift().join('');
                }) +
                '</table>' +
                '</span>';
            this.htmlTemplate = 
                 '<span class="mq-matrix mq-non-leaf">' +
                 '<div style="display: table;">' +                    
                    '<div style="display: table-row;">'+
                        '<div style="display: table-cell;border-top: 2px solid;border-left: 2px solid;border-bottom: 2px solid;width: 0.3em;padding: 0.1em;"></div>'+
                        '<div style="display: table-cell;">'+dataMatrix+'</div>'+
                        '<div style="display: table-cell;border-top: 2px solid;border-right: 2px solid;border-bottom: 2px solid;width: 0.3em;padding: 0.1em;"></div>'+
                    '</div>'+                   
                '</div>'+
                '</span>';
            return super_.html.call(this);
        };
        // Create default 1-cell matrix
        _.createBlocks = function() {
            this.blocks = [
                MatrixCell(0, this)
            ];
        };
        _.parser = function() {
            var self = this;
            var optWhitespace = Parser.optWhitespace;
            var string = Parser.string;

            return optWhitespace
                .then(string(delimiters.column)
                    .or(string(delimiters.row))
                    .or(latexMathParser.block))
                .many()
                .skip(optWhitespace)
                .then(function(items) {
                    var blocks = [];
                    var row = 0;
                    self.blocks = [];

                    function addCell() {
                        self.blocks.push(MatrixCell(row, self, blocks));
                        blocks = [];
                    }

                    for (var i = 0; i < items.length; i += 1) {
                        if (items[i] instanceof MathBlock) {
                            blocks.push(items[i]);
                        } else {
                            addCell();
                            if (items[i] === delimiters.row) row += 1;
                        }
                    }
                    addCell();
                    self.autocorrect();
                    return Parser.succeed(self);
                });
        };
        // Relink all the cells after parsing
        _.finalizeTree = function() {
            var table = this.jQ.find('table');
            table.toggleClass('mq-rows-1', table.find('tr').length === 1);
            this.relink();
        };
        // Enter the matrix at the top or bottom row if updown is configured.
        _.getEntryPoint = function(dir, cursor, updown) {
            if (updown === 'up') {
                if (dir === L) {
                    return this.blocks[this.rowSize - 1];
                } else {
                    return this.blocks[0];
                }
            } else { // updown === 'down'
                if (dir === L) {
                    return this.blocks[this.blocks.length - 1];
                } else {
                    return this.blocks[this.blocks.length - this.rowSize];
                }
            }
        };
        // Exit the matrix at the first and last columns if updown is configured.
        _.atExitPoint = function(dir, cursor) {
            // Which block are we in?
            var i = this.blocks.indexOf(cursor.parent);
            if (dir === L) {
                // If we're on the left edge and moving left, we should exit.
                return i % this.rowSize === 0;
            } else {
                // If we're on the right edge and moving right, we should exit.
                return (i + 1) % this.rowSize === 0;
            }
        };
        _.moveTowards = function(dir, cursor, updown) {
            var entryPoint = updown && this.getEntryPoint(dir, cursor, updown);
            cursor.insAtDirEnd(-dir, entryPoint || this.ends[-dir]);
        };

        // Set up directional pointers between cells
        _.relink = function() {
            var blocks = this.blocks;
            var rows = [];
            var row, column, cell;

            // The row size will be used by other functions down the track.
            // Begin by assuming we're a one-row matrix, and we'll overwrite this if we find another row.
            this.rowSize = blocks.length;

            // Use a for loop rather than eachChild
            // as we're still making sure children()
            // is set up properly
            for (var i = 0; i < blocks.length; i += 1) {
                cell = blocks[i];
                if (row !== cell.row) {
                    if (cell.row === 1) {
                        // We've just finished iterating the first row.
                        this.rowSize = column;
                    }
                    row = cell.row;
                    rows[row] = [];
                    column = 0;
                }
                rows[row][column] = cell;

                // Set up horizontal linkage
                cell[R] = blocks[i + 1];
                cell[L] = blocks[i - 1];

                // Set up vertical linkage
                if (rows[row - 1] && rows[row - 1][column]) {
                    cell.upOutOf = rows[row - 1][column];
                    rows[row - 1][column].downOutOf = cell;
                }

                column += 1;
            }

            // set start and end blocks of matrix
            this.ends[L] = blocks[0];
            this.ends[R] = blocks[blocks.length - 1];
        };
        // Ensure consistent row lengths
        _.autocorrect = function(rows) {
            var lengths = [],
                rows = [];
            var blocks = this.blocks;
            var maxLength, shortfall, position, row, i;

            for (i = 0; i < blocks.length; i += 1) {
                row = blocks[i].row;
                rows[row] = rows[row] || [];
                rows[row].push(blocks[i]);
                lengths[row] = rows[row].length;
            }

            maxLength = Math.max.apply(null, lengths);
            if (maxLength !== Math.min.apply(null, lengths)) {
                // Pad shorter rows to correct length
                for (i = 0; i < rows.length; i += 1) {
                    shortfall = maxLength - rows[i].length;
                    while (shortfall) {
                        position = maxLength * i + rows[i].length;
                        blocks.splice(position, 0, MatrixCell(i, this));
                        shortfall -= 1;
                    }
                }
                this.relink();
            }
        };
        // Deleting a cell will also delete the current row and
        // column if they are empty, and relink the matrix.
        _.deleteCell = function(currentCell) {
            var rows = [],
                columns = [],
                myRow = [],
                myColumn = [];
            var blocks = this.blocks,
                row, column;

            // Create arrays for cells in the current row / column
            this.eachChild(function(cell) {
                if (row !== cell.row) {
                    row = cell.row;
                    rows[row] = [];
                    column = 0;
                }
                columns[column] = columns[column] || [];
                columns[column].push(cell);
                rows[row].push(cell);

                if (cell === currentCell) {
                    myRow = rows[row];
                    myColumn = columns[column];
                }

                column += 1;
            });

            function isEmpty(cells) {
                var empties = [];
                for (var i = 0; i < cells.length; i += 1) {
                    if (cells[i].isEmpty()) empties.push(cells[i]);
                }
                return empties.length === cells.length;
            }

            function remove(cells) {
                for (var i = 0; i < cells.length; i += 1) {
                    if (blocks.indexOf(cells[i]) > -1) {
                        cells[i].remove();
                        blocks.splice(blocks.indexOf(cells[i]), 1);
                    }
                }
            }

            if (isEmpty(myRow) && myColumn.length > 1) {
                row = rows.indexOf(myRow);
                // Decrease all following row numbers
                this.eachChild(function(cell) {
                    if (cell.row > row) cell.row -= 1;
                });
                // Dispose of cells and remove <tr>
                remove(myRow);
                this.jQ.find(".mq-matrix-content:first").children("tbody").children('tr').eq(row).remove();
            }
            if (isEmpty(myColumn) && myRow.length > 1) {
                remove(myColumn);
            }
            this.finalizeTree();
        };
        _.addRow = function(afterCell) {
            var previous = [],
                newCells = [],
                next = [];
            var newRow = $('<tr></tr>'),
                row = afterCell.row;
            var columns = 0,
                block, column;

            this.eachChild(function(cell) {
                // Cache previous rows
                if (cell.row <= row) {
                    previous.push(cell);
                }
                // Work out how many columns
                if (cell.row === row) {
                    if (cell === afterCell) column = columns;
                    columns += 1;
                }
                // Cache cells after new row
                if (cell.row > row) {
                    cell.row += 1;
                    next.push(cell);
                }
            });

            // Add new cells, one for each column
            for (var i = 0; i < columns; i += 1) {
                block = MatrixCell(row + 1);
                block.parent = this;
                newCells.push(block);

                // Create cell <td>s and add to new row
                block.jQize($('<td mathquill-block-id="' + block.id + '" class="mq-empty mq-td">'))
                    .appendTo(newRow);
            }

            // Insert the new row
            this.jQ.find(".mq-matrix-content:first").children("tbody").children('tr').eq(row).after(newRow);
            this.blocks = previous.concat(newCells, next);
            return newCells[column];
        };
        _.addColumn = function(afterCell) {
            var rows = [],
                newCells = [];
            var column, block;

            // Build rows array and find new column index
            this.eachChild(function(cell) {
                rows[cell.row] = rows[cell.row] || [];
                rows[cell.row].push(cell);
                if (cell === afterCell) column = rows[cell.row].length;
            });

            // Add new cells, one for each row
            for (var i = 0; i < rows.length; i += 1) {
                block = MatrixCell(i);
                block.parent = this;
                newCells.push(block);
                rows[i].splice(column, 0, block);

                block.jQize($('<td mathquill-block-id="' + block.id + '" class="mq-empty mq-td">'))
            }

            // Add cell <td> elements in correct positions
            this.jQ.find(".mq-matrix-content:first").children("tbody").children('tr').each(function(i) {
                $(this).find('td').eq(column - 1).after(rows[i][column].jQ);
            });

            // Flatten the rows array-of-arrays
            this.blocks = [].concat.apply([], rows);
            return newCells[afterCell.row];
        };
        _.insert = function(method, afterCell) {
            var cellToFocus = this[method](afterCell);
            this.cursor = this.cursor || this.parent.cursor;
            //THUYLTM fix bug when matrix is inside the function or others
            if (this.cursor === undefined) {
                var rootParent = this.parent.parent;
                while (rootParent.cursor === undefined) {
                    rootParent = rootParent.parent;
                }
                this.cursor = rootParent.cursor;
            }
            this.finalizeTree();
            this.bubble(function(node) { node.reflow(); }).cursor.insAtRightEnd(cellToFocus);
        };
        _.backspace = function(cell, dir, cursor, finalDeleteCallback) {
            var dirwards = cell[dir];
            if (cell.isEmpty()) {
                this.deleteCell(cell);
                while (dirwards &&
                    dirwards[dir] &&
                    this.blocks.indexOf(dirwards) === -1) {
                    dirwards = dirwards[dir];
                }
                if (dirwards) {
                    cursor.insAtDirEnd(-dir, dirwards);
                }
                if (this.blocks.length === 1 && this.blocks[0].isEmpty()) {
                    finalDeleteCallback();
                    this.finalizeTree();
                }
                this.bubble(function(node) { node.controller && node.controller.handle('edited'); });
            }
        };
        _.reflow = function(){
            var lengthOfSupSup = this.jQ.closest(".mq-root-block").find(".mq-power-sup-sup").length;
            for(var i = 0; i < lengthOfSupSup; i++){
                reflowPowerSup($(this.jQ.closest(".mq-root-block").find(".mq-power-sup-sup")[i]));
            } 
        }
    });


Environments.pmatrix = P(Matrix, function(_, super_) {
    _.environment = 'pmatrix';
    _.parentheses = {
        left: '(',
        right: ')'
    };
});

Environments.bmatrix = P(Matrix, function(_, super_) {
    _.environment = 'bmatrix';
    _.parentheses = {
        left: '[',
        right: ']'
    };
});

Environments.Bmatrix = P(Matrix, function(_, super_) {
    _.environment = 'Bmatrix';
    _.parentheses = {
        left: '{',
        right: '}'
    };
});

Environments.vmatrix = P(Matrix, function(_, super_) {
    _.environment = 'vmatrix';
    _.parentheses = {
        left: '|',
        right: '|'
    };
});

Environments.Vmatrix = P(Matrix, function(_, super_) {
    _.environment = 'Vmatrix';
    _.parentheses = {
        left: '&#8214;',
        right: '&#8214;'
    };
});



// Replacement for mathblocks inside matrix cells
// Adds matrix-specific keyboard commands
var MatrixCell = P(MathBlock, function(_, super_) {
    _.init = function(row, parent, replaces) {
        super_.init.call(this);
        this.row = row;
        if (parent) {
            this.adopt(parent, parent.ends[R], 0);
        }
        if (replaces) {
            for (var i = 0; i < replaces.length; i++) {
                replaces[i].children().adopt(this, this.ends[R], 0);
            }
        }
    };
    _.keystroke = function(key, e, ctrlr) {
        switch (key) {
            case 'Shift-Spacebar':
                e.preventDefault();
                return this.parent.insert('addColumn', this);
            case 'Shift-Enter':
                return this.parent.insert('addRow', this);
        }
        return super_.keystroke.apply(this, arguments);
    };
    _.deleteOutOf = function(dir, cursor) {
        var self = this,
            args = arguments;
        this.parent.backspace(this, dir, cursor, function() {
            // called when last cell gets deleted
            return super_.deleteOutOf.apply(self, args);
        });
    };
    _.moveOutOf = function(dir, cursor, updown) {
        var atExitPoint = updown && this.parent.atExitPoint(dir, cursor);
        // Step out of the matrix if we've moved past an edge column
        if (!atExitPoint && this[dir]) cursor.insAtDirEnd(-dir, this[dir]);
        else cursor.insDirOf(dir, this.parent);
        cursor.parent.bubble(function(node) { node.moveOutOfBlock && node.moveOutOfBlock(); });
    };
});

// ThuyLTM add piecewise and solve
var Cases =
    Environments.cases = P(Environment, function(_, super_) {
        _.delimiters = {
            column: '&',
            row: '\\\\'
        };
        _.environment = 'cases';
        _.reflow = function() {
            var blockData = this.jQ.find(".mq-cases-block:first");
            var bigBracket = blockData.children(":eq(0)").find(".mq-bracket-solve:first")[0];            
            var bigBracketQ = blockData.find('div.mq-bracket-solve:first');
            var curlBracket = bigBracketQ.find(".mq-curl:first");
            var normalBracket = blockData.children(":eq(0)").find('.mq-solve-paren:first');
            bigBracketQ.css({"display":"none"});
            normalBracket.css({"display":"block"});
            normalBracket.empty();
            normalBracket.removeAttr("style");
            normalBracket.append("<div>{</div>"); 

            var blockHeight = this.jQ[0].getBoundingClientRect().height;
            var bracketHeight = normalBracket.children(":first").height();
            var height = +Number(blockHeight /+bracketHeight).toFixed(2);
            
            if (height < 8) {
                parens = normalBracket;     
                var scaleX = min(1 + .2 * (height - 1), 1.2);      
                var scaleY = undefined;
                if (parens.length) {               
                    scaleY = height;          
                    scale(parens, scaleX, scaleY);
                    parens.css({"transform-origin": "top left"});
                }
            } else {
                bigBracketQ.css({"display":"block"});
                normalBracket.css({"display":"none"});
                normalBracket.empty();
                normalBracket.removeAttr("style"); 
                bigBracketQ.find('.mq-new-vline').remove();
                curlBracket.prev().addClass("mq-vline-pre");  
                var currentHeightBigBracket = bigBracket.getBoundingClientRect().height;
                var lineBracketHeight = bigBracketQ.children("div:nth-child(2)").height();
                if (blockHeight > currentHeightBigBracket) {
                    var adjustLineBracket = Math.ceil(Math.abs(blockHeight - currentHeightBigBracket)/(lineBracketHeight)); 
                    var curlTail = bigBracketQ.find(".mq-curl-tail:first");
                    var contentBracketList = blockData.children(".mq-bracket-content");
                    var lowContentBracket = Math.round(contentBracketList[0].getBoundingClientRect().bottom);
                    for (var j = 1; j < contentBracketList.length; j++) {
                        if (lowContentBracket < Math.round(contentBracketList[j].getBoundingClientRect().bottom)) {                                
                            lowContentBracket = Math.round(contentBracketList[j].getBoundingClientRect().bottom);
                        } 
                    }
                    for (var i = 0; i < adjustLineBracket + 2; i++) {                             
                        var lineBefore = curlBracket.prevAll(".mq-new-vline").length;
                        var lineAfter = curlBracket.nextAll(".mq-new-vline").length;
                        curlBracket.prev().removeClass("mq-vline-pre");     
                        if (lineBefore >= lineAfter) {
                            $('<div style="text-align: center;" class="mq-new-vline">⎪</div>').insertAfter(curlBracket);   
                        } else {                                
                            $('<div style="text-align: center;" class="mq-new-vline">⎪</div>').insertBefore(curlBracket);                               
                        }  
                        curlBracket.prev().addClass("mq-vline-pre");                         
                        if (Math.round(curlTail[0].getBoundingClientRect().bottom) - curlTail[0].getBoundingClientRect().height/2 + 3 >= lowContentBracket) {
                            return;
                        }                           
                    }     
                }
            }
        };
        _.latex = function() {
            var latex = '';
            var row;
            var self = this;

            this.eachChild(function(cell) {
                if (typeof row !== 'undefined') {
                    latex += (row !== cell.row) ?
                        self.delimiters.row :
                        self.delimiters.column;
                }
                row = cell.row;
                latex += cell.latex();
            });

            return this.wrappers().join(latex);
        };
        _.html = function() {
            return super_.html.call(this);
        };
        _.parser = function() {
            var self = this;
            var optWhitespace = Parser.optWhitespace;
            var string = Parser.string;

            return optWhitespace
                .then(string(self.delimiters.column)
                    .or(string(self.delimiters.row))
                    .or(latexMathParser.block))
                .many()
                .skip(optWhitespace)
                .then(function(items) {
                    var blocks = [];
                    var row = 0;
                    var column = 0;
                    self.blocks = [];

                    function addCell() {
                        self.blocks.push(FunctionCell(row, column, self, blocks));
                        blocks = [];
                    }

                    for (var i = 0; i < items.length; i += 1) {
                        if (items[i] instanceof MathBlock) {
                            blocks.push(items[i]);
                        } else {
                            addCell();
                            if (items[i] === self.delimiters.row) {
                                row += 1;
                                column = 0;
                            }
                            if (items[i] === self.delimiters.column) {
                                column += 1;
                            }
                        }
                    }
                    addCell();
                    self.autocorrect();
                    return Parser.succeed(self);
                });
        };
        // Relink all the cells after parsing
        _.finalizeTree = function() {
            var table = this.jQ.find('table');
            table.toggleClass('mq-rows-1', table.find('tr').length === 1);
            this.relink();
        };
        // Enter the matrix at the top or bottom row if updown is configured.
        _.getEntryPoint = function(dir, cursor, updown) {
            if (updown === 'up') {
                if (dir === L) {
                    return this.blocks[this.blocks.length - 1];
                } else {
                    return this.blocks[0];
                }
            } else { // updown === 'down'
                if (dir === L) {
                    return this.blocks[this.blocks.length - 1];
                } else {
                    return this.blocks[this.blocks.length - this.rowSize];
                }
            }
        };
        // Exit the matrix at the first and last columns if updown is configured.
        _.atExitPoint = function(dir, cursor) {
            // Which block are we in?
            var i = this.blocks.indexOf(cursor.parent);
            if (dir === L) {
                // If we're on the left edge and moving left, we should exit.
                return i === 0;
            } else {
                // If we're on the right edge and moving right, we should exit.
                return i === this.blocks.length - 1;
            }
        };
        _.moveTowards = function(dir, cursor, updown) {
            var entryPoint = updown && this.getEntryPoint(dir, cursor, updown);
            cursor.insAtDirEnd(-dir, entryPoint || this.ends[-dir]);
        };

        // Set up directional pointers between cells
        _.relink = function() {
            var blocks = this.blocks;
            var rows = [];
            var row, column, cell;

            // Use a for loop rather than eachChild
            // as we're still making sure children()
            // is set up properly
            for (var i = 0; i < blocks.length; i += 1) {
                cell = blocks[i];
                if (row !== cell.row) {
                    row = cell.row;
                    rows[row] = [];
                    column = 0;
                }
                rows[row][column] = cell;

                // Set up horizontal linkage
                cell[R] = blocks[i + 1];
                cell[L] = blocks[i - 1];

                // Set up vertical linkage
                if (rows[row - 1] && rows[row - 1][column]) {
                    cell.upOutOf = rows[row - 1][column];
                    rows[row - 1][column].downOutOf = cell;
                }

                column += 1;
            }

            // set start and end blocks of matrix
            this.ends[L] = blocks[0];
            this.ends[R] = blocks[blocks.length - 1];
            this.rowSize = rows.length;
        };
        // Ensure consistent row lengths
        _.autocorrect = function(rows) {
            var lengths = [],
                rows = [];
            var blocks = this.blocks;
            var maxLength, shortfall, position, row, i;

            for (i = 0; i < blocks.length; i += 1) {
                row = blocks[i].row;
                rows[row] = rows[row] || [];
                rows[row].push(blocks[i]);
                lengths[row] = rows[row].length;
            }

            maxLength = Math.max.apply(null, lengths);
            if (maxLength !== Math.min.apply(null, lengths)) {
                // Pad shorter rows to correct length
                for (i = 0; i < rows.length; i += 1) {
                    shortfall = maxLength - rows[i].length;
                    while (shortfall) {
                        position = maxLength * i + rows[i].length;
                        blocks.splice(position, 0);
                        shortfall -= 1;
                    }
                }
                this.relink();
            }
        };
        // Deleting a cell will also delete the current row and
        // column if they are empty, and relink the matrix.
        _.deleteRow = function(currentRow) {
            var blocks = this.blocks;
            var maxColumn = this.maxColumn;

            function remove(currentRow) {
                blocks.splice(currentRow * maxColumn, maxColumn);
            }
            // Re-index all following row numbers
            this.eachChild(function(cell) {
                if (cell.row > currentRow) cell.row -= 1;
            });
            // Dispose of cells and remove <tr>
            remove(currentRow);
            this.jQ.find('tr').eq(currentRow).remove();
            this.rowSize -= 1;
            this.finalizeTree();
        };
        _.addRow = function(afterCell) {
            var previous = [],
                newCells = [],
                next = [];
            var newRow = $('<tr></tr>'),
                row = afterCell.row;
            var columns = 0,
                block, column;

            this.eachChild(function(cell) {
                // Cache previous rows
                if (cell.row <= row) {
                    previous.push(cell);
                }
                // Work out how many columns
                if (cell.row === row) {
                    if (cell === afterCell) column = columns;
                    columns += 1;
                }
                // Cache cells after new row
                if (cell.row > row) {
                    cell.row += 1;
                    next.push(cell);
                }
            });

            // Add new cells, one for each column
            for (var i = 0; i < this.maxColumn; i += 1) {
                block = FunctionCell(row + 1, i);
                block.parent = this;
                newCells.push(block);

                // Create cell <td>s and add to new row
                block.jQize($('<td mathquill-block-id="' + block.id + '" class="mq-empty mq-td">'))
                    .appendTo(newRow);
            }

            // Insert the new row
            this.jQ.find('tr').eq(row).after(newRow);
            this.blocks = previous.concat(newCells, next);
            return newCells[column];
        };
        _.insert = function(method, afterCell) {
            var cellToFocus = this[method](afterCell);
            this.cursor = this.cursor || this.parent.cursor;
            //THUYLTM fix bug when matrix is inside the function or others
            if (this.cursor === undefined) {
                var rootParent = this.parent.parent;
                while (rootParent.cursor === undefined) {
                    rootParent = rootParent.parent;
                }
                this.cursor = rootParent.cursor;
            }
            this.finalizeTree();
            this.bubble(function(node) { node.reflow(); }).cursor.insAtRightEnd(cellToFocus);
        };
        _.backspace = function(cell, dir, cursor, finalDeleteCallback) {
            var dirwards = cell[dir];
            if (cell.isEmpty()) {
                this.deleteRow(cell.row);
                while (dirwards &&
                    dirwards[dir] &&
                    this.blocks.indexOf(dirwards) === -1) {
                    dirwards = dirwards[dir];
                }
                if (dirwards) {
                    cursor.insAtDirEnd(-dir, dirwards);
                }
                if (this.rowSize < 2) {
                    finalDeleteCallback();
                    this.finalizeTree();
                }
                this.bubble(function(node) { node.controller && node.controller.handle('edited'); });
            }
        };
    });


Environments.solve = P(Cases, function(_, super_) {
    _.bigCurlBracket = '<div style="float:left;display:none;font-weight:bold;" class="mq-bracket-solve mq-symbol-font">'+
                '<div class="mq-curl-head">⎧</div>'+
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div style="text-align: center;" class="mq-curl">⎨</div>'+
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div class="mq-curl-tail">⎩</div>'+
                '<div style="clear:both;"></div></div>';
    _.delimiters = {
        column: '&',
        row: '\\\\'
    };
    _.maxColumn = 1;
    _.environment = 'solve'; 
    _.html = function() {
        var cells = [],
            trs = '',
            i = 0,
            row;
        // Build <tr><td>.. structure from cells
        this.eachChild(function(cell) {
            if (row !== cell.row) {
                row = cell.row;
                trs += '<tr>$tds</tr>';
                cells[row] = [];
            }
            if (cell.column == 0) {
                cells[row].push('<td class="mq-td">&' + (i++) + '</td>');
            }
        });
        var dataSolve = '<table class="mq-non-leaf mq-solve" style="border-right: solid 1px;">' +
            trs.replace(/\$tds/g, function() {
                return cells.shift().join('');
            }) +
            '</table>'; 
        if (isSafariV9()) {
             this.htmlTemplate = '<span class="mq-matrix mq-non-leaf">' +                
                 '<div style="display: table;">' + 
                    '<div style="display: table-row;" class="mq-cases-block">'+
                        '<div style="display: table-cell;" class="mq-paren">'+
                            this.bigCurlBracket+
                            '<div class="mq-solve-paren"></div>'+
                        '</div>'+
                        '<div style="display: table-cell;" class="mq-bracket-content">'+
                           dataSolve +        
                           '<span class="mq-scaled mq-paren mq-seperate" style="width:0px;"></span>' +
                           '<table class="mq-non-leaf">' +
                           '<tr><td>&' + i + '</td></tr></table>' +                   
                        '</div>'+                       
                    '</div>'+
                '</div>'+
                '</span>';
        } else {
            this.htmlTemplate = '<span class="mq-matrix mq-non-leaf">' +                
                 '<div style="display: table;">' + 
                    '<div style="display: table-row;" class="mq-cases-block">'+
                        '<div style="display: table-cell;" class="mq-paren">'+
                            this.bigCurlBracket+
                            '<div class="mq-solve-paren"></div>'+
                        '</div>'+
                        '<div style="display: table-cell;" class="mq-bracket-content">'+
                           dataSolve +                           
                        '</div>'+
                        '<div style="display: table-cell;" class="mq-separate"></div>'+
                        '<div style="display: table-cell;" class="mq-bracket-content mq-solve-condition"><div style="margin-left: 0.25em;padding: 0.1em 0.3em;">&'+i+'</div></div>'+
                    '</div>'+
                '</div>'+
                '</span>';
        }  
        return super_.html.call(this);
    };
    _.reflow = function() {
        if (isSafariV9()) {
            var blockData = this.jQ.find(".mq-cases-block:first");
            var blockContent = blockData.children(":eq(1)");
            var separateLine = blockContent.children(":eq(1)");
            separateLine.css("height","");
        }
        super_.reflow.call(this);   
        if (isSafariV9()) {
            var blockData = this.jQ.find(".mq-cases-block:first");
            var blockContent = blockData.children(":eq(1)");
            var separateLine = blockContent.children(":eq(1)");
            separateLine.css({"height":blockContent.innerHeight()+"px"});
        }     
    };
    _.createBlocks = function() {
        this.blocks = [
            FunctionCell(0, 0, this),
            FunctionCell(1, 0, this),
            FunctionCell(1, 1, this),
        ];
    };
    _.latex = function() {
        var blocks = this.blocks;
        var latex = "";
        var loopCount = blocks.length - 2;
        for (var i = 0; i <= loopCount; i++) {
            latex += blocks[i].latex();
            if (i < loopCount) {
                latex += ",";
            }
        }
        return OPEN_BRACKET_SERVER + "solve" + '(' + '\\{' + latex + "\\}" + "," + "\\{" + blocks[i].latex() + "\\}" + ")" + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        var blocks = this.blocks;
        var latex = "";
        var loopCount = blocks.length - 2;
        for (var i = 0; i <= loopCount; i++) {
            latex += blocks[i].sLatex();
            if (i < loopCount) {
                latex += this.delimiters.row;
            }
        }
        latex += this.delimiters.column + blocks[blocks.length - 1].sLatex();
        return this.wrappers().join(latex);
    };
    _.nativeLatex = function() {
        var blocks = this.blocks;
        var latex = "";
        var loopCount = blocks.length - 2;
        for (var i = 0; i <= loopCount; i++) {
            latex += blocks[i].nativeLatex();
            if (i < loopCount) {
                latex += this.delimiters.row;
            }
        }
        latex += this.delimiters.column + blocks[blocks.length - 1].nativeLatex();
        return this.wrappers().join(latex);
    };
    _.sLatexToSVG = function() {
        var blocks = this.blocks;
        var latex = "";
        var loopCount = blocks.length - 2;
        for (var i = 0; i <= loopCount; i++) {
            latex += blocks[i].sLatexToSVG();
            if (i < loopCount) {
                latex += this.delimiters.row;
            }
        }
        latex = "\\left.{" + this.wrappers().join(latex) + "}\\right\\vert_{" + blocks[blocks.length - 1].sLatexToSVG()+"}";
        return latex;
    };
    _.parser = function() {
        var self = this;
        var optWhitespace = Parser.optWhitespace;
        var string = Parser.string;
        return optWhitespace
            .then(string(this.delimiters.row)
                .or(string(this.delimiters.column))
                .or(latexMathParser.block)
            )
            .many()
            .skip(optWhitespace)
            .then(function(items) {
                var blocks = [];
                var row = 0;
                var column = 0;
                self.blocks = [];

                function addCell() {
                    self.blocks.push(FunctionCell(row, column, self, blocks));
                    blocks = [];
                }

                for (var i = 0; i < items.length; i += 1) {
                    if (items[i] instanceof MathBlock) {
                        blocks.push(items[i]);
                    } else {
                        addCell();
                        if (items[i] === self.delimiters.row) {
                            row += 1;
                        } else if (items[i] === self.delimiters.column) {
                            column += 1;
                        }
                    }
                }
                addCell();
                self.relink();
                return Parser.succeed(self);
            });
    };
    _.relink = function() {
        var blocks = this.blocks;
        var rows = [];
        var row, cell;

        // Use a for loop rather than eachChild
        // as we're still making sure children()
        // is set up properly
        for (var i = 0; i < blocks.length; i += 1) {
            cell = blocks[i];
            if (row !== cell.row && cell.column === 0) {
                row = cell.row;
                rows[row] = [];
                rows[row][0] = cell;
                // Set up vertical linkage
                if (rows[row - 1] && rows[row - 1][0]) {
                    cell.upOutOf = rows[row - 1][0];
                    rows[row - 1][0].downOutOf = cell;
                }
            }
            // Set up horizontal linkage
            if (blocks[i + 1]) {
                cell[R] = blocks[i + 1];
            } else {
                cell[R] = 0;
            }
            if (blocks[i - 1]) {
                cell[L] = blocks[i - 1];
            } else {
                cell[L] = 0;
            }
        }

        // set start and end blocks of matrix
        this.ends[L] = blocks[0];
        this.ends[R] = blocks[blocks.length - 1];
        this.rowSize = rows.length;
    };
    _.backspace = function(cell, dir, cursor, finalDeleteCallback) {
        var dirwards = cell[dir];
        if (cell.isEmpty()) {
            if (cell.column == 0) {
                this.deleteRow(cell.row);
                if (dirwards === 0) {
                    dirwards = cell[-dir];
                }
            }

            while (dirwards &&
                dirwards[dir] &&
                this.blocks.indexOf(dirwards) === -1) {
                dirwards = dirwards[dir];
            }
            if (dirwards) {
                cursor.insAtDirEnd(-dir, dirwards);
            }
            if (this.rowSize < 2) {
                finalDeleteCallback();
                this.finalizeTree();
            }
            this.bubble(function(node) { node.controller && node.controller.handle('edited'); });
        }
    };
    _.deleteRow = function(currentRow) {
        var blocks = this.blocks;
        var lastRow = blocks[blocks.length - 1].row;

        function remove(currentRow) {
            blocks.splice(currentRow, 1);
        }
        // Re-index all following row numbers
        this.eachChild(function(cell) {
            if (cell.row > currentRow) cell.row -= 1;
        });
        if (currentRow === lastRow) {
            blocks[blocks.length - 1].row -= 1;
        }
        // Dispose of cells and remove <tr>
        remove(currentRow);
        this.jQ.find('table.mq-solve:first').children("tbody").children('tr').eq(currentRow).remove();
        this.rowSize -= 1;
        this.finalizeTree();
    };
    _.addRow = function(afterCell) {
        if (afterCell.column === 1) return afterCell;
        var previous = [],
            newCells = [],
            next = [];
        var newRow = $('<tr></tr>'),
            row = afterCell.row;
        var block;
        var blocks = this.blocks;

        if (row === blocks[blocks.length - 1].row) {
            blocks[blocks.length - 1].row += 1;
        }

        this.eachChild(function(cell) {
            // Cache previous rows
            if (cell.row <= row) {
                previous.push(cell);
            }
            // Cache cells after new row
            if (cell.row > row) {
                cell.row += 1;
                next.push(cell);
            }
        });


        // Add new cells, one for each column
        for (var i = 0; i < this.maxColumn; i += 1) {
            block = FunctionCell(row + 1, i);
            block.parent = this;
            newCells.push(block);

            // Create cell <td>s and add to new row
            block.jQize($('<td mathquill-block-id="' + block.id + '" class="mq-empty mq-td">'))
                .appendTo(newRow);
        }

        // Insert the new row
        this.jQ.find('table.mq-solve:first').children("tbody").children('tr').eq(row).after(newRow);
        this.blocks = previous.concat(newCells, next);
        return newCells[0];
    };
});


Environments.piecewise = P(Cases, function(_, super_) {
   _.bigCurlBracket = '<div style="float:left;display:none;font-weight:bold;" class="mq-bracket-solve mq-symbol-font">'+
                '<div class="mq-curl-head">⎧</div>'+
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div style="text-align: center;" class="mq-curl">⎨</div>'+
                '<div style="text-align: center;" class="mq-vline">⎪</div>'+
                '<div class="mq-curl-tail">⎩</div>'+
                '<div style="clear:both;"></div></div>';
    _.delimiters = {
        column: '&',
        row: '\\\\'
    };
    _.maxColumn = 2;
    _.environment = 'piecewise';    
    _.html = function() {
        var cells = [],
            trs = '',
            i = 0,
            row;
        // Build <tr><td>.. structure from cells
        this.eachChild(function(cell) {
            if (row !== cell.row) {
                row = cell.row;
                trs += '<tr>$tds</tr>';
                cells[row] = [];
            }
            if (cell.column == 1) {
                cells[row].push('<td style="border-right:1px solid black;" class="mq-solve mq-solve-separate"></td>');
                cells[row].push('<td style="padding: 0.1em 0.3em;" class="mq-solve">&' + (i++) + '</td>');
            } else {
                cells[row].push('<td style="padding: 0.1em 0.3em;" class="mq-solve">&' + (i++) + '</td>');
            }
        });
        var dataPiecewise =  '<table class="mq-non-leaf mq-piecew">' +
                            trs.replace(/\$tds/g, function() {
                                return cells.shift().join('');
                            }) +
                            '</table>';       
        this.htmlTemplate = '<span class="mq-matrix mq-non-leaf">' +                
                 '<div style="display: table;">' + 
                    '<div style="display: table-row;" class="mq-cases-block">'+
                        '<div style="display: table-cell;" class="mq-paren">'+
                            this.bigCurlBracket+
                            '<div class="mq-solve-paren"></div>'+
                        '</div>'+
                        '<div style="display: table-cell;" class="mq-bracket-content">'+dataPiecewise+'</div>'+
                    '</div>'+
                '</div>'+
                '</span>';
        return super_.html.call(this);
    };
    _.createBlocks = function() {
        this.blocks = [
            FunctionCell(0, 0, this),
            FunctionCell(0, 1, this),
            FunctionCell(1, 0, this),
            FunctionCell(1, 1, this)
        ];
    };
    _.backspace = function(cell, dir, cursor, finalDeleteCallback) {
        var blocks = this.blocks;
        var dirwards = cell[dir];
        var maxColumn = this.maxColumn;
        if (cell.isEmpty()) {
            function isEmptyRow(currentRow) {
                for (var i = 0; i < maxColumn; i++) {
                    if (!blocks[currentRow * maxColumn + i].isEmpty()) {
                        return false;
                    }
                }
                if (cell.column === 0) {
                    return true;
                }
                return false;
            }
            if (isEmptyRow(cell.row)) {
                this.deleteRow(cell.row);
                if (dirwards === undefined) {
                    dirwards = blocks[maxColumn - 1];
                }
            } else if (dirwards === undefined) {
                dirwards = blocks[(cell.row + 1)* maxColumn + maxColumn - 1];
            }
            while (dirwards &&
                dirwards[dir] &&
                this.blocks.indexOf(dirwards) === -1) {
                dirwards = dirwards[dir];
            }
            if (dirwards) {
                cursor.insAtDirEnd(-dir, dirwards);
            }
            if (this.rowSize < 2) {
                finalDeleteCallback();
                this.finalizeTree();
            }
            this.bubble(function(node) { node.controller && node.controller.handle('edited'); });
        }
    };
    _.addRow = function(afterCell) {
        var previous = [],
            newCells = [],
            next = [];
        var newRow = $('<tr></tr>'),
            row = afterCell.row;
        var columns = 0,
            block, column;

        this.eachChild(function(cell) {
            // Cache previous rows
            if (cell.row <= row) {
                previous.push(cell);
            }
            // Work out how many columns
            if (cell.row === row) {
                if (cell === afterCell) column = columns;
                columns += 1;
            }
            // Cache cells after new row
            if (cell.row > row) {
                cell.row += 1;
                next.push(cell);
            }
        });

        // Add new cells, one for each column
        for (var i = 0; i < this.maxColumn; i += 1) {
            block = FunctionCell(row + 1, i);
            block.parent = this;
            newCells.push(block);

            // Create cell <td>s and add to new row
            block.jQize($('<td mathquill-block-id="' + block.id + '" class="mq-empty mq-td">'))
                .appendTo(newRow);
            if (i == 0) {
                $('<td style="border-right:1px solid black;" class="mq-solve mq-solve-separate"></td>').appendTo(newRow);
            }
        }

        // Insert the new row
        this.jQ.find(".mq-piecew:first").children("tbody").children('tr').eq(row).after(newRow);
        this.blocks = previous.concat(newCells, next);
        return newCells[column];
    };
     _.deleteRow = function(currentRow) {
        var blocks = this.blocks;
        var maxColumn = this.maxColumn;

        function remove(currentRow) {
            blocks.splice(currentRow * maxColumn, maxColumn);
        }
        // Re-index all following row numbers
        this.eachChild(function(cell) {
            if (cell.row > currentRow) cell.row -= 1;
        });
        // Dispose of cells and remove <tr>
        remove(currentRow);
        this.jQ.find(".mq-piecew:first").children("tbody").children('tr').eq(currentRow).remove();
        this.rowSize -= 1;
        this.finalizeTree();
    };
    _.latex = function() {
        var blocks = this.blocks;
        var length = blocks.length;
        var totalGroup = [];

        var groupsBlock = [];
        var lastGroup = [];
        var currentGroup = [];

        lastGroup[0] = blocks[length - 3];
        lastGroup[1] = blocks[length - 4];
        lastGroup[2] = blocks[length - 2];
        lastGroup[3] = blocks[length - 1];
        var latex = "";
        if (lastGroup[3].isEmpty()) {
            latex = "piecew(";
            lastGroup.splice(3, 1);
            lastGroup.forEach(function(cBlock, index) {
                if (index === 0 || index === 1) {
                    latex += cBlock.latex() + ",";
                }
                if (index === 2) {
                    latex += cBlock.latex();
                }
            });
            latex += ")";
        } else {
            latex = "piecew(" + lastGroup[3].latex() + "," + lastGroup[2].latex() + ",(1/0)" + ")";
            latex = "piecew(" + lastGroup[0].latex() + "," + lastGroup[1].latex() + "," + latex + ")";
        }
        for (var i = 0; i <= length - 5; i++) {
            if (i % 2 == 0) {
                currentGroup[1] = blocks[i];
            }
            if (i % 2 == 1) {
                currentGroup[0] = blocks[i];
                groupsBlock.push(currentGroup);
                currentGroup = [undefined, undefined];
            }
        }

        for (var i = groupsBlock.length - 1; i >= 0; i--) {
            currentGroup = groupsBlock[i];
            latex = "piecew(" + currentGroup[0].latex() + "," + currentGroup[1].latex() + "," + latex + ")";
        }
        return OPEN_BRACKET_SERVER + latex + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        var blocks = this.blocks;
        var maxRowIndex = blocks.length / 2 - 1;
        var latex = "";
        var loopCount = blocks.length;
        for (var i = 0; i < loopCount; i++) {
            if (i % 2 == 0) {
                latex += blocks[i].sLatex();
            } else {
                latex += this.delimiters.column + blocks[i].sLatex();
                if (blocks[i].row < maxRowIndex) {
                    latex += this.delimiters.row;
                }
            }
        }
        return this.wrappers().join(latex);
    };
    _.nativeLatex = function() {
        var blocks = this.blocks;
        var maxRowIndex = blocks.length / 2 - 1;
        var latex = "";
        var loopCount = blocks.length;
        for (var i = 0; i < loopCount; i++) {
            if (i % 2 == 0) {
                latex += blocks[i].nativeLatex();
            } else {
                latex += this.delimiters.column + blocks[i].nativeLatex();
                if (blocks[i].row < maxRowIndex) {
                    latex += this.delimiters.row;
                }
            }
        }
        return this.wrappers().join(latex);
    };
    _.sLatexToSVG = function() {
        var blocks = this.blocks;
        var maxRowIndex = blocks.length / 2 - 1;
        var latex = "";
        var loopCount = blocks.length;
        for (var i = 0; i < loopCount; i++) {
            if (i % 2 == 0) {
                latex += "\\left.{" + blocks[i].sLatexToSVG() + "}";
            } else {
                latex += "\\right|{" + blocks[i].sLatexToSVG() + "}";
                if (blocks[i].row < maxRowIndex) {
                    latex += this.delimiters.row;
                }
            }
        }
        return [
            ['\\begin{cases}'],
            ['\\end{cases}']
        ].join(latex);
    };
    _.parser = function() {
        var self = this;
        var optWhitespace = Parser.optWhitespace;
        var string = Parser.string;
        return optWhitespace
            .then(string(this.delimiters.row)
                .or(string(this.delimiters.column))
                .or(latexMathParser.block)
            )
            .many()
            .skip(optWhitespace)
            .then(function(items) {
                var blocks = [];
                var row = 0;
                var column = 0;
                self.blocks = [];

                function addCell() {
                    self.blocks.push(FunctionCell(row, column, self, blocks));
                    blocks = [];
                }

                for (var i = 0; i < items.length; i += 1) {
                    if (items[i] instanceof MathBlock) {
                        blocks.push(items[i]);
                    } else {
                        addCell();
                        if (items[i] === self.delimiters.row) {
                            row += 1;
                            column = 0;
                        } else if (items[i] === self.delimiters.column) {
                            column += 1;
                        }
                    }
                }
                addCell();
                self.autocorrect();
                return Parser.succeed(self);
            });
    };
});

var FunctionCell = P(MathBlock, function(_, super_) {
    _.init = function(row, column, parent, replaces) {
        super_.init.call(this);
        this.row = row;
        this.column = column;
        if (parent) {
            this.adopt(parent, parent.ends[R], 0);
        }
        if (replaces) {
            for (var i = 0; i < replaces.length; i++) {
                replaces[i].children().adopt(this, this.ends[R], 0);
            }
        }
    };
    _.keystroke = function(key, e, ctrlr) {
        switch (key) {
            case 'Shift-Enter':
                return this.parent.insert('addRow', this);
        }
        return super_.keystroke.apply(this, arguments);
    };
    _.deleteOutOf = function(dir, cursor) {
        var self = this,
            args = arguments;
        this.parent.backspace(this, dir, cursor, function() {
            // called when last cell gets deleted
            return super_.deleteOutOf.apply(self, args);
        });
    };
    _.moveOutOf = function(dir, cursor, updown) {
        var atExitPoint = updown && this.parent.atExitPoint(dir, cursor);
        // Step out of the matrix if we've moved past an edge column
        if (!atExitPoint && this[dir]) cursor.insAtDirEnd(-dir, this[dir]);
        else cursor.insDirOf(dir, this.parent);
        cursor.parent.bubble(function(node) { node.moveOutOfBlock && node.moveOutOfBlock(); });
    };
});

//ThuyLTM
Environments.parametric = P(Environment, function(_, super_) {
    _.delimiters = {
        column: '&',
        row: '\\\\'
    };
    _.maxColumn = 1;
    _.environment = 'parametric';
    _.ctrlSeq = 'parametric';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-parametric mq-matrix mq-non-leaf">' +
            '<div style="margin-top:1rem;">' +
            '<span class="mq-scaled mq-paren">{</span>' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td">' +
            '<var>x</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>t</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td">' +
            '<var>y</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>t</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td">&1</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '<div>' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="accept-only-digit mq-td">&2</td>' +
            // NNANH 12/20/2018 - S - Edit Bug#9903 Change font parametric
            '<td class="mq-td">	&#60946;</td>' +
            // NNANH 12/20/2018 - E- Edit Bug#9903 Change font parametric
            '<td class="mq-td">' +
            '<var>t</var>' +
            '</td>' +
            // NNANH 12/20/2018 - S - Edit Bug#9903 Change font parametric
            '<td class="mq-td">	&#60946;</td>' +
            // NNANH 12/20/2018 - E - Edit Bug#9903 Change font parametric
            '<td class="accept-only-digit mq-td">&3</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
        var blockjQ = this.jQ.find("table:first");

        var height = blockjQ.outerHeight() / +blockjQ.css('fontSize').slice(0, -2);

        var parens = this.jQ.find('.mq-paren:first');
        if (parens.length) {
            scale(parens, min(1 + .2 * (height - 1), 1.2), 1.05 * height);
        }
        //NNANH 26/9/2018 Start Bug #8631 
        parens.css("margin-top", 0);
        var topOfBlockJQ = blockjQ[0].getBoundingClientRect().top;
        var heightOfBlockJQ = blockjQ[0].getBoundingClientRect().height;
        var topOfParens = parens[0].getBoundingClientRect().top;
        var heightOfParens = parens[0].getBoundingClientRect().height;
        valueOfMargin = topOfBlockJQ - topOfParens;
        valueSupMargin = (heightOfParens - heightOfBlockJQ)/2;
        parens.css("margin-top", valueOfMargin - valueSupMargin);
        //NNANH 26/9/2018 Start Bug #8631 
    };
    _.createBlocks = function() {
        this.blocks = [
            ParametricCell(0, 0, this),
            ParametricCell(1, 0, this),
            ParametricCell(2, 0, this),
            ParametricCell(2, 1, this),
        ];
    };
    // Exit the matrix at the first and last columns if updown is configured.
    _.atExitPoint = function(dir, cursor) {
        // Which block are we in?
        var i = this.blocks.indexOf(cursor.parent);
        if (dir === L) {
            // If we're on the left edge and moving left, we should exit.
            return i === 0;
        } else {
            // If we're on the right edge and moving right, we should exit.
            return i === this.blocks.length - 1;
        }
    };
    _.latex = function() {
        var blocks = this.blocks;
        return "parametric(\\{" + blocks[0].latex() + "\\},\\{" + blocks[1].latex() +
            "\\},\\{" + blocks[2].latex() + "\\},\\{" + blocks[3].latex() + "\\})";
    };
    _.sLatex = function() {
        var blocks = this.blocks;
        var latex = "{ x_t= }" + this.delimiters.column + "{" + blocks[0].sLatex() + "}" + this.delimiters.row +
            "{ y_t= }" + this.delimiters.column + "{" + blocks[1].sLatex() + "}" + this.delimiters.row +
            "{" + blocks[2].sLatex() + "}" + this.delimiters.column + "{ \\leq t \\leq }" + this.delimiters.column +
            "{" + blocks[3].sLatex() + "}";
        return this.wrappers().join(latex);
    };
    _.nativeLatex = function() {
        var blocks = this.blocks;
        var latex = "{ x_t= }" + this.delimiters.column + "{" + blocks[0].nativeLatex() + "}" + this.delimiters.row +
            "{ y_t= }" + this.delimiters.column + "{" + blocks[1].nativeLatex() + "}" + this.delimiters.row +
            "{" + blocks[2].nativeLatex() + "}" + this.delimiters.column + "{ \\leq t \\leq }" + this.delimiters.column +
            "{" + blocks[3].nativeLatex() + "}";
        return this.wrappers().join(latex);
    };
    _.sLatexToSVG = function() {
        var blocks = this.blocks;
        var latex = "{ x_t= }" + this.delimiters.column + "{" + blocks[0].sLatexToSVG() + "}" + this.delimiters.row +
            "{ y_t= }" + this.delimiters.column + "{" + blocks[1].sLatexToSVG() + "}" + this.delimiters.row +
            "{" + blocks[2].sLatexToSVG() + "}" + this.delimiters.column + "{ \\leq t \\leq }" + this.delimiters.column +
            "{" + blocks[3].sLatexToSVG() + "}";
        return this.wrappers().join(latex);
    };
    _.finalizeTree = function() {
        this.downInto = this.ends[R];
        this.upInto = this.ends[L];
        this.relink();
    };
    // Set up directional pointers between cells
    _.relink = function() {
        var blocks = this.blocks;
        var rows = [];
        var row, column, cell;

        // Use a for loop rather than eachChild
        // as we're still making sure children()
        // is set up properly
        for (var i = 0; i < blocks.length; i += 1) {
            cell = blocks[i];
            if (i === 0) {
                // Set up horizontal linkage
                cell[L] = 0;
                cell[R] = blocks[i + 1];
                // Set up vertical linkage
                cell.downOutOf = blocks[i + 1];
            } else if (i === blocks.length - 1) {
                // Set up horizontal linkage
                cell[L] = blocks[i - 1];
                cell[R] = 0;
                // Set up vertical linkage
                cell.upOutOf = blocks[0];
            } else {
                // Set up horizontal linkage
                cell[R] = blocks[i + 1];
                cell[L] = blocks[i - 1];
                // Set up vertical linkage
                cell.downOutOf = blocks[i + 1];
                cell.upOutOf = blocks[i - 1];
            }
        }
        // set start and end blocks of matrix
        this.ends[L] = blocks[0];
        this.ends[R] = blocks[blocks.length - 1];
    };
    _.isEmpty = function() { return true; };
    _.parser = function() {
        var self = this;
        var optWhitespace = Parser.optWhitespace;
        var string = Parser.string;
        return optWhitespace
            .then(string(this.delimiters.row)
                .or(string(this.delimiters.column))
                .or(latexMathParser.block)
            )
            .many()
            .skip(optWhitespace)
            .then(function(items) {
                self.blocks = [];
                self.blocks.push(ParametricCell(0, 0, self, [items[2]]));
                self.blocks.push(ParametricCell(1, 0, self, [items[6]]));
                self.blocks.push(ParametricCell(2, 0, self, [items[8]]));
                self.blocks.push(ParametricCell(2, 1, self, [items[12]]));
                self.relink();
                return Parser.succeed(self);
            });
    };
    _.backspace = function(cell, dir, cursor, finalDeleteCallback) {
        finalDeleteCallback();
        this.bubble(function(node) { node.controller && node.controller.handle('edited'); });
    };
});

var ParametricCell = P(MathBlock, function(_, super_) {
    _.init = function(row, column, parent, replaces) {
        super_.init.call(this);
        this.row = row;
        this.column = column;
        if (parent) {
            this.adopt(parent, parent.ends[R], 0);
        }
        if (replaces) {
            for (var i = 0; i < replaces.length; i++) {
                replaces[i].children().adopt(this, this.ends[R], 0);
            }
        }
    };
    _.keystroke = function(key, e, ctrlr) {
        switch (key) {
            case 'Shift-Enter':
                return this.parent.insert('addRow', this);
                break;
        }
        return super_.keystroke.apply(this, arguments);
    };
    _.deleteOutOf = function(dir, cursor) {
        var i = this.parent.blocks.indexOf(cursor.parent);
        if (i === 0) {
            var blocks = this.parent.blocks;
            var isNotEmpty;
            for (var j = 0; j < blocks.length; j++) {
                if (!blocks[j].isEmpty()) {
                    isNotEmpty = true;
                    break;
                }
            }
            if (!isNotEmpty) {
                var self = this,
                    args = arguments;
                this.parent.backspace(this, dir, cursor, function() {
                    // called when last cell gets deleted
                    return super_.deleteOutOf.apply(self, args);
                });
            } else {
                cursor.insDirOf(L, this.parent);
            }
        } else {
            cursor.insAtRightEnd(this.parent.blocks[i - 1]);
        }
    };
    _.moveOutOf = function(dir, cursor, updown) {
        var atExitPoint = this.parent.atExitPoint(dir, cursor);
        // Step out of the matrix if we've moved past an edge column
        if (!atExitPoint && this[dir]) {
            cursor.insAtDirEnd(-dir, this[dir]);
        } else {
            var i = this.parent.blocks.indexOf(cursor.parent);
            /*  if (i === 0) {
               cursor.insAtRightEnd(this.parent.blocks[1]);
             } else { */
            cursor.insDirOf(dir, this.parent);
            //}
        }
        cursor.parent.bubble(function(node) { node.moveOutOfBlock && node.moveOutOfBlock(); });
    };
});
//NNANH add
var SequenceCell = P(MathBlock, function(_, super_) {
    _.init = function(row, column, parent, replaces) {
        super_.init.call(this);
        this.row = row;
        this.column = column;
        if (parent) {
            this.adopt(parent, parent.ends[R], 0);
        }
        if (replaces) {
            for (var i = 0; i < replaces.length; i++) {
                replaces[i].children().adopt(this, this.ends[R], 0);
            }
        }
    };
    _.keystroke = function(key, e, ctrlr) {
        switch (key) {
            case 'Shift-Enter':
                return this.parent.insert('addRow', this);
                break;
        }
        return super_.keystroke.apply(this, arguments);
    };
    _.deleteOutOf = function(dir, cursor) {
        var i = this.parent.blocks.indexOf(cursor.parent);
        if (i === 0) {
            var blocks = this.parent.blocks;
            var isNotEmpty;
            for (var j = 0; j < blocks.length; j++) {
                if (!blocks[j].isEmpty()) {
                    isNotEmpty = true;
                    break;
                }
            }
            if (!isNotEmpty) {
                var self = this,
                    args = arguments;
                this.parent.backspace(this, dir, cursor, function() {
                    // called when last cell gets deleted
                    return super_.deleteOutOf.apply(self, args);
                });
            } else {
                cursor.insDirOf(L, this.parent);
            }
        } else {
            cursor.insAtRightEnd(this.parent.blocks[i - 1]);
        }
    };
    _.moveOutOf = function(dir, cursor, updown) {
        var atExitPoint = this.parent.atExitPoint(dir, cursor);
        // Step out of the matrix if we've moved past an edge column
        if (!atExitPoint && this[dir]) {
            cursor.insAtDirEnd(-dir, this[dir]);
        } 
        // else {
        //     var i = this.parent.blocks.indexOf(cursor.parent);
        //     /*  if (i === 0) {
        //        cursor.insAtRightEnd(this.parent.blocks[1]);
        //      } else { */
        //     cursor.insDirOf(dir, this.parent);
        //     //}
        // }
        cursor.parent.bubble(function(node) { node.moveOutOfBlock && node.moveOutOfBlock(); });
    };
});

//Typea0 and  Type a
var RecursiveType = LatexCmds.recursiveType = P(MathCommand, function(_, super_) {
    _.reflow = function() {
    };
    _.atExitPoint = function(dir, cursor) {
        // Which block are we in?
        var i = this.blocks.indexOf(cursor.parent);
        if (dir === L) {
            // If we're on the left edge and moving left, we should exit.
            return i === 0;
        } else {
            // If we're on the right edge and moving right, we should exit.
            return i === this.blocks.length - 1;
        }
    };
    // Exit the matrix at the first and last columns if updown is configured.
    _.finalizeTree = function() {
        this.downInto = this.ends[R];
        this.upInto = this.ends[L];
        this.relink();
    };
    // Set up directional pointers between cells
    _.relink = function() {
        var blocks = this.blocks;
        var rows = [];
        var row, column, cell;

        // Use a for loop rather than eachChild
        // as we're still making sure children()
        // is set up properly
        for (var i = 0; i < blocks.length; i += 1) {
            cell = blocks[i];
            if (i === 0) {
                // Set up horizontal linkage
                cell[L] = 0;
                cell[R] = blocks[i + 1];
                // Set up vertical linkage
                cell.downOutOf = blocks[i + 1];
            } else if (i === blocks.length - 1) {
                // Set up horizontal linkage
                cell[L] = blocks[i - 1];
                cell[R] = 0;
                // Set up vertical linkage
                cell.upOutOf = blocks[0];
            } else {
                // Set up horizontal linkage
                cell[R] = blocks[i + 1];
                cell[L] = blocks[i - 1];
                // Set up vertical linkage
                cell.downOutOf = blocks[i + 1];
                cell.upOutOf = blocks[i - 1];
            }
        }
        // set start and end blocks of matrix
        this.ends[L] = blocks[0];
        this.ends[R] = blocks[blocks.length - 1];
    };
    _.isEmpty = function() { return true; };
    _.backspace = function(cell, dir, cursor, finalDeleteCallback) {
        return;
    };
    _.nativeLatex = function() {
       return " ";
    };
});

// rm1097
LatexCmds.typeax = P(RecursiveType, function(_, super_) {
    _.environment = 'typeax';
    _.ctrlSeq = 'typeax';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypebx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typea aN_1 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>0</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typea a_0 block-sequence">&1</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';
        return super_.html.call(this);
    };

    _.reflow = function() {
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'a_n+1' + '=' + this.blocks[0].latex() + '\\,' + 'a_0' + '=' + this.blocks[1].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typeax' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}';
    };
    // _.nativeLatex = function() {
    //     return '\\typeax' + '{' + this.blocks[0].nativeLatex() + "}{" +
    //             this.blocks[1].nativeLatex() + '}';
    // };
    _.sLatexToSVG = function() {
       return "a_{n+1}=" + this.blocks[0].sLatexToSVG() + ",a_0=" + this.blocks[1].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
          })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(2).result(self);
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
        ];
    };
});

//Typea0 and  Type b
LatexCmds.typebx = P(RecursiveType, function(_, super_) {
    _.environment = 'typebx';
    _.ctrlSeq = 'typebx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypebx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typea aN_1 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>0</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typea a_0 block-sequence">&1</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };

    _.reflow = function() {
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'b_n+1' + '=' + this.blocks[0].latex() + '\\,' + 'b_0' + '=' + this.blocks[1].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typebx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}';
    };
    // _.nativeLatex = function() {
        // return '\\typebx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].nativeLatex() + '}';
    // };
    _.sLatexToSVG = function() {
        return "b_{n+1}=" + this.blocks[0].sLatexToSVG() + ",b_0=" + this.blocks[1].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(2).result(self);
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
        ];
    };
});

//Typea0 and  Type c
LatexCmds.typecx = P(RecursiveType, function(_, super_) {
    _.environment = 'typecx';
    _.ctrlSeq = 'typecx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypecx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typea aN_1 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>0</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typea a_0 block-sequence">&1</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };

    _.reflow = function() {
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'c_n+1' + '=' + this.blocks[0].latex() + '\\,' + 'c_0' + '=' + this.blocks[1].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typecx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}';
    };
    // _.nativeLatex = function() {
        // return '\\typecx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].nativeLatex() + '}';
    // };
    _.sLatexToSVG = function() {
        return "c_{n+1}=" + this.blocks[0].sLatexToSVG() + ",c_0=" + this.blocks[1].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(2).result(self);
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
        ];
    };
});

//a(n+1) and a1 Type A1
LatexCmds.typeaxx = P(RecursiveType, function(_, super_) {
    _.environment = 'typeaxx';
    _.ctrlSeq = 'typeaxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypeaxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typea aN_1 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typea a_1 block-sequence">&1</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'a_n+1' + '=' + this.blocks[0].latex() + '\\,' + 'a_1' + '=' + this.blocks[1].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typeaxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}';
    };
    // _.nativeLatex = function() {
        // return '\\typeaxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].nativeLatex() + '}';
    // };
    _.sLatexToSVG = function() {
        return "a_{n+1}=" + this.blocks[0].sLatexToSVG() + ",a_1=" + this.blocks[1].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
      })
        )
        .atMost(2).result(self);
    };
});

//Typea1 and  Type b
LatexCmds.typebxx = P(RecursiveType, function(_, super_) {
    _.environment = 'typebxx';
    _.ctrlSeq = 'typebxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypebxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typea aN_1 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typea a_1 block-sequence">&1</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'b_n+1' + '=' + this.blocks[0].latex() + '\\,' + 'b_1' + '=' + this.blocks[1].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typebxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}';
    };
    // _.nativeLatex = function() {
        // return '\\typebxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].nativeLatex() + '}';
    // };
    _.sLatexToSVG = function() {
        return "b_{n+1}=" + this.blocks[0].sLatexToSVG() + ",b_1=" + this.blocks[1].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(2).result(self);
    };
});

//Typea1 and  Type c
LatexCmds.typecxx = P(RecursiveType, function(_, super_) {
    _.environment = 'typecxx';
    _.ctrlSeq = 'typecxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypecxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typea aN_1 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typea a_1 block-sequence">&1</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'c_n+1' + '=' + this.blocks[0].latex() + '\\,' + 'c_1' + '=' + this.blocks[1].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typecxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}';
    };
    // _.nativeLatex = function() {
        // return '\\typecxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].nativeLatex() + '}';
    // };
    _.sLatexToSVG = function() {
        return "c_{n+1}=" + this.blocks[0].sLatexToSVG() + ",c_1=" + this.blocks[1].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(2).result(self);
    };
});

//a(n+2) and a0,a1 Type A0,A1
LatexCmds.typeaxaxx = P(RecursiveType, function(_, super_) {
    _.environment = 'typeaxaxx';
    _.ctrlSeq = 'typeaxaxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypeaxaxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aN_2 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>0</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec1 a_0 block-sequence">&1</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec2 a_1 block-sequence">&2</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
            SequenceCell(2, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'a_n+2' + '=' + this.blocks[0].latex() + '\\,' + 'a_0' + '=' + this.blocks[1].latex() + '\\,' + 'a_1' + '=' + this.blocks[2].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typeaxaxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\typeaxaxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "a_{n+2}=" + this.blocks[0].sLatexToSVG() + ",a_0=" + this.blocks[1].sLatexToSVG() + ",a_1=" + this.blocks[2].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(3).result(self);
    };
});

//a(n+2) and  Type B0,B1
LatexCmds.typebxbxx = P(RecursiveType, function(_, super_) {
    _.delimiters = {
        column: '&',
        row: '\\\\'
    };
    _.maxColumn = 1;
    _.environment = 'typebxbxx';
    _.ctrlSeq = 'typebxbxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypebxbxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aN_2 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>0</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec1 a_0 block-sequence">&1</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec2 a_2 block-sequence">&2</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
            SequenceCell(2, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'b_n+2' + '=' + this.blocks[0].latex() + '\\,' + 'b_0' + '=' + this.blocks[1].latex() + '\\,' + 'b_1' + '=' + this.blocks[2].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typebxbxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\typebxbxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "b_{n+2}=" + this.blocks[0].sLatexToSVG() + ",b_0=" + this.blocks[1].sLatexToSVG() + ",b_1=" + this.blocks[2].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(3).result(self);
    };
});

//a(n+2) and  Type c0,c1
LatexCmds.typecxcxx = P(RecursiveType, function(_, super_) {
    _.delimiters = {
        column: '&',
        row: '\\\\'
    };
    _.maxColumn = 1;
    _.environment = 'typecxcxx';
    _.ctrlSeq = 'typecxcxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypecxcxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aN_2 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>0</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec1 a_0 block-sequence">&1</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec2 a_1 block-sequence">&2</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
            SequenceCell(2, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'c_n+2' + '=' + this.blocks[0].latex() + '\\,' + 'c_0' + '=' + this.blocks[1].latex() + '\\,' + 'c_1' + '=' + this.blocks[2].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typecxcxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\typecxcxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "c_{n+2}=" + this.blocks[0].sLatexToSVG() + ",c_0=" + this.blocks[1].sLatexToSVG() + ",c_1=" + this.blocks[2].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(3).result(self);
    };
});

//A(n+2) Type a1, A2
LatexCmds.typeaxxaxxx = P(RecursiveType, function(_, super_) {
    _.environment = 'typeaxxaxxx';
    _.ctrlSeq = 'typeaxxaxxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypeaxxaxxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aN_2 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec1 a_1 block-sequence">&1</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec2 a_2 block-sequence">&2</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
            SequenceCell(2, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'a_n+2' + '=' + this.blocks[0].latex() + '\\,' + 'a_1' + '=' + this.blocks[1].latex() + '\\,' + 'a_2' + '=' + this.blocks[2].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typeaxxaxxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\typeaxxaxxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "a_{n+2}=" + this.blocks[0].sLatexToSVG() + ",a_1=" + this.blocks[1].sLatexToSVG() + ",a_2=" + this.blocks[2].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(3).result(self);
  };
});

//A(n+2) Type b1, b2
LatexCmds.typebxxbxxx = P(RecursiveType, function(_, super_) {
    _.environment = 'typebxxbxxx';
    _.ctrlSeq = 'typebxxbxxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypebxxbxxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aN_2 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec1 a_1 block-sequence">&1</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec2 a_2 block-sequence">&2</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
            SequenceCell(2, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'b_n+2' + '=' + this.blocks[0].latex() + '\\,' + 'b_1' + '=' + this.blocks[1].latex() + '\\,' + 'b_2' + '=' + this.blocks[2].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typebxxbxxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\typebxxbxxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "b_{n+2}=" + this.blocks[0].sLatexToSVG() + ",b_1=" + this.blocks[1].sLatexToSVG() + ",b_2=" + this.blocks[2].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(3).result(self);
    };
});

//A(n+2) Type c1, c2
LatexCmds.typecxxcxxx = P(RecursiveType, function(_, super_) {
    _.environment = 'typecxxcxxx';
    _.ctrlSeq = 'typecxxcxxx';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursivetypecxxcxxx mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n+2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aN_2 block-sequence">&0</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>1</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec1 a_1 block-sequence">&1</td>' +
            '</tr>' +
            '<tr>' +
            '<td class="mq-td mq-left-side" style=" text-align: right;">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>2</var></span>' +
            '</span>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-bottom-typec2 a_2 block-sequence">&2</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this),
            SequenceCell(1, 0, this),
            SequenceCell(2, 0, this),
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'c_n+2' + '=' + this.blocks[0].latex() + '\\,' + 'c_1' + '=' + this.blocks[1].latex() + '\\,' + 'c_2' + '=' + this.blocks[2].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\typecxxcxxx' + '{' + this.blocks[0].sLatex() + "}{" +
                this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\typecxxcxxx' + '{' + this.blocks[0].nativeLatex() + "}{" +
        //         this.blocks[1].sLatex() + '}{' + this.blocks[2].sLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "c_{n+2}=" + this.blocks[0].sLatexToSVG() + ",c_1=" + this.blocks[1].sLatexToSVG() + ",c_2=" + this.blocks[2].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock(), MathBlock(), MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        var check = 0;
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                    block.children().adopt(blocks[check], blocks[check].ends[R], 0);
                    check++;
                return succeed(self);
            })
        )
        .atMost(3).result(self);
    };
});

//explicit a
LatexCmds.explicita = P(RecursiveType, function(_, super_) {
    _.environment = 'explicita';
    _.ctrlSeq = '   ';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursive-explicita mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>a</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n</var></span>' +
            '</span>' +
            '<var>E</var>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aNE block-sequence">&0</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this)
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'explicita' + '=' + this.blocks[0].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\explicita' + '{' + this.blocks[0].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\explicita' + '{' + this.blocks[0].nativeLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "a_nE=" + this.blocks[0].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        //\diffa{x}\left( x\right) HandWriting
        // \diffa{x}{x}
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                if (blocks[0].isEmpty()) {
                    block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                }
                return succeed(self);
            })
        )
        .atMost(1).result(self);
    };
});

//explicit b
LatexCmds.explicitb = P(RecursiveType, function(_, super_) {
    _.environment = 'explicitb';
    _.ctrlSeq = 'explicitb';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursive-explicit mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>b</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n</var></span>' +
            '</span>' +
            '<var>E</var>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aNE block-sequence">&0</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this)
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'explicitb' + '=' + this.blocks[0].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\explicitb' + '{' + this.blocks[0].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\explicitb' + '{' + this.blocks[0].nativeLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "b_nE=" + this.blocks[0].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                if (blocks[0].isEmpty()) {
                    block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                }
                return succeed(self);
            })
        )
        .atMost(1).result(self);
    };
});

//explicit c
LatexCmds.explicitc = P(RecursiveType, function(_, super_) {
    _.environment = 'explicitc';
    _.ctrlSeq = 'explicitc';
    _.html = function() {
        this.htmlTemplate =
            '<span class="mq-recursiveType mq-recursive-explicitc mq-matrix mq-non-leaf">' +
            '<div style="margin-top:0.125rem;">' +
            '<table class="mq-non-leaf">' +
            '<tbody>' +
            '<tr>' +
            '<td class="mq-td mq-left-side">' +
            '<var>c</var>' +
            '<span class="mq-supsub mq-non-leaf">' +
            '<span class="mq-sub"><var>n</var></span>' +
            '</span>' +
            '<var>E</var>' +
            '<span class="mq-binary-operator">=</span>' +
            '</td>' +
            '<td class="mq-td block-top-typec aNE block-sequence">&0</td>' +
            '</tr>' +
            '</tbody>' +
            '</table>' +
            '</div>' +
            '</span>';

        return super_.html.call(this);
    };
    _.reflow = function() {
    };
    _.createBlocks = function() {
        this.blocks = [
            SequenceCell(0, 0, this)
        ];
    };
    _.latex = function() {
        return OPEN_BRACKET_SERVER + 'explicitc' + '=' + this.blocks[0].latex() + CLOSE_BRACKET_SERVER;
    };
    _.sLatex = function() {
        return '\\explicitc' + '{' + this.blocks[0].sLatex() + "}";
    };
    // _.nativeLatex = function() {
        // return '\\explicitc' + '{' + this.blocks[0].nativeLatex() + "}";
    // };
    _.sLatexToSVG = function() {
        return "c_nE=" + this.blocks[0].sLatexToSVG();
    };
    //typeax{12}{123}
    _.parser = function() {
        var string = Parser.string;
        var optWhitespace = Parser.optWhitespace;
        var succeed = Parser.succeed;
        var block = latexMathParser.block;
        var regex = Parser.regex;

        var self = this;
        var blocks = self.blocks = [MathBlock()];
        for (var i = 0; i < blocks.length; i += 1) {
            blocks[i].adopt(self, self.ends[R], 0);
        }
        return (
            string("\\left(").then(function() {
                    return latexMathParser.then(function(block) {
                        return string('\\right)').map(function(end) {
                            block.children().adopt(blocks[1], blocks[1].ends[R], 0);
                            return succeed(self);
                        });
                    });
            })
        ).or(                
                block.then(function(block) {  
                if (blocks[0].isEmpty()) {
                    block.children().adopt(blocks[0], blocks[0].ends[R], 0);
                }
                return succeed(self);
            })
        )
        .atMost(1).result(self);
    };
});
